unit mxBNMathUtils;

// Binuar 1.2.1
// Big integer number arithmetic
// ---
// Copyright  2000-3000, Andrey A. Meshkov (AL-CHEMIST)
// All rights reserved
//
// http://maalchemist.ru
// maa@maalchemist.ru
// maalchemist@yandex.ru
// maalchemist@gmail.com
//
// Free for any use.
// No warranty of any kind.
// ---

{$I version.inc}
{$I mxBNMath.inc}

interface

uses
  mxBNMathTypes;

function  bsf (a: UIntX): IntX;
function  bsr (a: UIntX): IntX;

function  bit_s_test (a, b: UIntX): Boolean;
function  bit_c_test (a, b: UIntX): Boolean;

function  cdiv (a, b: UIntX): UIntX; overload;
function  cdiv (a, b: IntX): IntX; overload;
function  cdiv_mod (a, b: UIntX; var a_mod: UIntX): UIntX; overload;
function  cdiv_mod (a, b: IntX; var a_mod: IntX): IntX; overload;
function  div_mod (a, b: UIntX; var a_mod: UIntX): UIntX; overload;
function  div_mod (a, b: IntX; var a_mod: IntX): IntX; overload;

function  BN_bsf (const a: T_BNX): IntX;
function  BN_bsr (const a: T_BNX): IntX;

function  BN_Bytes (AValue: UIntX): UIntX;
function  BN_Nibbles (AValue: UIntX): UIntX;
function  BN_Bits (AValue: UIntX): UIntX; overload;
function  BN_Bits (const a: T_BNX): UIntX; overload;

function  BN_Size (ACount: IntX): UIntX; overload;
function  BN_Size (ACount: UIntX): UIntX; overload;

function  BN_IncPointer (P: Pointer): Pointer;
function  BN_IncPointer_(var P): Pointer;
function  BN_DecPointer (P: Pointer): Pointer;
function  BN_DecPointer_(var P): Pointer;

procedure BN_ZeroData (DST: Pointer; Count: IntX);
procedure BN_ZeroDataDesc (DST: Pointer; Count: IntX);
procedure BN_FillData (DST: Pointer; Count: IntX; V: UIntX);
procedure BN_FillDataDesc (DST: Pointer; Count: IntX; V: UIntX);
procedure BN_CopyData (DST, SRC: Pointer; Count: IntX);
procedure BN_CopyDataDesc (DST, SRC: Pointer; Count: IntX);

function  Break_(ABreak: PBoolean): Boolean;

implementation

uses
  SysUtils, mxBNMathProcs;

// [+] 1.2.1
{
  BSF - Bit Scan Forward
}
function bsf (a: UIntX): IntX;
{$IFDEF CPUX64}
// rcx | a |
asm
       .NOFRAME

        or      rax, -1
        bsf     rax, rcx
end;
{$ELSE}
// eax | a | ecx
asm
        mov     ecx, eax
        or      eax, -1
        bsf     eax, ecx
end;
{$ENDIF}

// [+] 1.2.1
{
  BSR - Bit Scan Reverse
}
function bsr (a: UIntX): IntX;
{$IFDEF CPUX64}
// rcx | a |
asm
       .NOFRAME

        or      rax, -1
        bsr     rax, rcx
end;
{$ELSE}
// eax | a | ecx
asm
        mov     ecx, eax
        or      eax, -1
        bsr     eax, ecx
end;
{$ENDIF}

// [+] 1.2.1
{
  BT - Bit Test
  Returns TRUE if bit is set
}
function bit_s_test (a, b: UIntX): Boolean;
{$IFDEF CPUX64}
// rcx | a |
// rdx | b |
asm
       .NOFRAME

        xor     rax, rax
        bt      rcx, rdx
        adc     rax, 0
end;
(*
asm
       .NOFRAME

        bt      rcx, rdx
        lahf                    // AH = EFLAGS(SF:ZF:0:AF:0:PF:1:CF)
        shr     rax, 8          // AL = EFLAGS(SF:ZF:0:AF:0:PF:1:CF)
        and     rax, 1          // AL = CF
end;
// *)
{$ELSE}
// eax | a | ecx ?
// edx | b |
asm
        mov     ecx, eax
        xor     eax, eax
        bt      ecx, edx
        adc     eax, 0
end;
(*
asm
        bt      eax, edx
        lahf                    // AH = EFLAGS(SF:ZF:0:AF:0:PF:1:CF)
        shr     eax, 8          // AL = EFLAGS(SF:ZF:0:AF:0:PF:1:CF)
        and     eax, 1          // AL = CF
end;
// *)
{$ENDIF}

// [+] 1.2.1
{
  BT - Bit Test
  Returns TRUE if bit is clear
}
function bit_c_test (a, b: UIntX): Boolean;
{$IFDEF CPUX64}
// rcx | a |
// rdx | b |
asm
       .NOFRAME

        bt      rcx, rdx
        lahf                    // AH = EFLAGS(SF:ZF:0:AF:0:PF:1:CF)
        shr     rax, 8          // AL = EFLAGS(SF:ZF:0:AF:0:PF:1:CF)
        and     rax, 1          // AL = CF
        xor     rax, 1          // 0->1 | 1->0
end;
(*
asm
       .NOFRAME

        xor     rax, rax
        bt      rcx, rdx
        jc      @R
        inc     rax
  @R:
end;
// *)
{$ELSE}
// eax | a | ecx ?
// edx | b |
asm
        bt      eax, edx
        lahf                    // AH = EFLAGS(SF:ZF:0:AF:0:PF:1:CF)
        shr     eax, 8          // AL = EFLAGS(SF:ZF:0:AF:0:PF:1:CF)
        and     eax, 1          // AL = CF
        xor     eax, 1          // 0->1 | 1->0
end;
(*
asm
        mov     ecx, eax
        xor     eax, eax
        bt      ecx, edx
        jc      @R
        inc     eax
  @R:
end;
// *)
{$ENDIF}

{
  a mod b == 0 : result = a div b
  a mod b != 0 : result = a div b + 1
}
function cdiv (a, b: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | a |
// rdx | b | r10
asm
       .NOFRAME

        mov     r10, rdx
        mov     rax, rcx
        xor     rdx, rdx
        div     r10      // rdx = a mod b
        test    rdx, rdx // ? a mod b == 0
        jz      @r
        inc     rax      // a mod b != 0
  @r:
end;
{$ELSE}
// eax | a |
// edx | b | ecx
asm
        mov     ecx, edx
        xor     edx, edx
        div     ecx      // edx = a mod b
        test    edx, edx // ? a mod b == 0
        jz      @r
        inc     eax      // a mod b != 0
  @r:
end;
{$ENDIF}

{
  a mod b == 0 : result = a div b
  a mod b != 0 : result = a div b + 1
}
function cdiv (a, b: IntX): IntX;
{$IFDEF CPUX64}
// rcx | a |
// rdx | b | r10
asm
       .NOFRAME

        mov     r10, rdx
        mov     rax, rcx
        cqo
        idiv    r10      // rdx = a mod b
        test    rdx, rdx // ? a mod b == 0
        jz      @r
        inc     rax      // a mod b != 0
  @r:
end;
{$ELSE}
// eax | a |
// edx | b | ecx
asm
        mov     ecx, edx
        cdq
        idiv    ecx      // edx = a mod b
        test    edx, edx // ? a mod b == 0
        jz      @r
        inc     eax      // a mod b != 0
  @r:
end;
{$ENDIF}

{
  a mod b == 0 : result = a div b
  a mod b != 0 : result = a div b + 1
                 a_mod  = a mod b
}
function cdiv_mod (a, b: UIntX; var a_mod: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | a      |
// rdx | b      | r10
// r8  | @a_mod |
asm
       .NOFRAME

        mov     r10, rdx
        mov     rax, rcx
        xor     rdx, rdx
        div     r10      // rdx = a mod b
        test    rdx, rdx // ? a mod b == 0
        jz      @r
        inc     rax      // a mod b != 0
  @r:
        mov     [r8], rdx
end;
{$ELSE}
// eax | a      |
// edx | b      | ebx
// ecx | @a_mod |
asm
        push    ebx

        mov     ebx, edx
        xor     edx, edx
        div     ebx      // edx = a mod b
        test    edx, edx // ? a mod b == 0
        jz      @r
        inc     eax      // a mod b != 0
  @r:
        mov     [ecx], edx

        pop     ebx
end;
{$ENDIF}

{
  a mod b == 0 : result = a div b
  a mod b != 0 : result = a div b + 1
                 a_mod  = a mod b
}
function cdiv_mod (a, b: IntX; var a_mod: IntX): IntX;
{$IFDEF CPUX64}
// rcx | a      |
// rdx | b      | r10
// r8  | @a_mod |
asm
       .NOFRAME

        mov     r10, rdx
        mov     rax, rcx
        cqo
        idiv    r10      // rdx = a mod b
        test    rdx, rdx // ? a mod b == 0
        jz      @r
        inc     rax      // a mod b != 0
  @r:
        mov     [r8], rdx
end;
{$ELSE}
// eax | a      |
// edx | b      | ebx
// ecx | @a_mod |
asm
        push    ebx

        mov     ebx, edx
        cdq
        idiv    ebx      // edx = a mod b
        test    edx, edx // ? a mod b == 0
        jz      @r
        inc     eax      // a mod b != 0
  @r:
        mov     [ecx], edx

        pop     ebx
end;
{$ENDIF}

{
  result = a div b
  a_mod  = a mod b
}
function div_mod (a, b: UIntX; var a_mod: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | a      |
// rdx | b      | r10
// r8  | @a_mod |
asm
       .NOFRAME

        mov     r10, rdx
        mov     rax, rcx
        xor     rdx, rdx
        div     r10
        mov     [r8], rdx // rdx = a mod b
  @r:
end;
{$ELSE}
// eax | a      |
// edx | b      | ebx
// ecx | @a_mod |
asm
        push    ebx

        mov     ebx, edx
        xor     edx, edx
        div     ebx
        mov     [ecx], edx // edx = a mod b
  @r:
        pop     ebx
end;
{$ENDIF}

{
  result = a div b
  a_mod  = a mod b
}
function div_mod (a, b: IntX; var a_mod: IntX): IntX;
{$IFDEF CPUX64}
// rcx | a      |
// rdx | b      | r10
// r8  | @a_mod |
asm
       .NOFRAME

        mov     r10, rdx
        mov     rax, rcx
        cqo
        idiv    r10
        mov     [r8], rdx // rdx = a mod b
  @r:
end;
{$ELSE}
// eax | a      |
// edx | b      | ebx
// ecx | @a_mod |
asm
        push    ebx

        mov     ebx, edx
        cdq
        idiv    ebx
        mov     [ecx], edx // edx = a mod b
  @r:
        pop     ebx
end;
{$ENDIF}

// [+] 1.2.1
{
  Bit Scan Forward
  Scans a.Data for first bit set
  BSF scans in (0-n)

  Returns -1 if no bits found
}
function BN_bsf (const a: T_BNX): IntX;
{$IFDEF CPUX64}
// rcx | @a |
asm
       .NOFRAME

        or      rax, -1         // Result = -1
        xor     r10, r10        // r10 = 0

        mov     rdx, [rcx].T_BNX.Count
        cmp     rdx, 0          // rdx = a.Count
        jle     @R

        mov     r8, [rcx].T_BNX.PLData
  @A:                           // for I := 0 to HIndex ...
        mov     r9, [r8]        // r9 = a.Data[I]
        test    r9, r9
        jz      @C
  @B:
//      xor     rax, rax
        bsf     rax, r9
        add     rax, r10
        jmp     @R
  @C:
        add     r10, CXBits
  @D:
        add     r8, CXBytes
        dec     rdx
        jnz     @A

  @R:
end;
{$ELSE}
// eax | @a | ecx
asm
        push    ebx
        push    esi
        push    edi

        mov     ecx, eax        // ecx = @a
        or      eax, -1         // Result = -1
        xor     ebx, ebx        // ebx = 0

        mov     edx, [ecx].T_BNX.Count
        cmp     edx, 0          // edx = a.Count
        jle     @R

        mov     esi, [ecx].T_BNX.PLData
  @A:                           // for I := 0 to HIndex ...
        mov     edi, [esi]      // edi = a.Data[I]
        test    edi, edi
        jz      @C
  @B:
//      xor     eax, eax
        bsf     eax, edi
        add     eax, ebx
        jmp     @R
  @C:
        add     ebx, CXBits
  @D:
        add     esi, CXBytes
        dec     edx
        jnz     @A

  @R:
        pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

// [+] 1.2.1
{
  Bit Scan Reverse
  Scans a.Data for first bit set
  BSR scans in (n-0)

  Returns -1 if no bits found
}
function BN_bsr (const a: T_BNX): IntX;
{$IFDEF CPUX64}
// rcx | @a |
asm
       .NOFRAME

        or      rax, -1         // Result = -1

        mov     rdx, [rcx].T_BNX.Count
        cmp     rdx, 0          // rdx = a.Count
        jle     @R

        mov     r8, [rcx].T_BNX.PHData
  @A:                           // for I := HIndex downto 0 ...
        mov     r9, [r8]        // r9 = a.Data[I]
        test    r9, r9
        jz      @C
  @B:
//      xor     rax, rax
        bsr     rax, r9
        dec     rdx
        shl     rdx, CXBiSh     // rdx = c * CXBits
        add     rax, rdx
        jmp     @R
  @C:
  @D:
        sub     r8, CXBytes
        dec     rdx
        jnz     @A

  @R:
end;
{$ELSE}
// eax | @a | ecx
asm
        push    esi
        push    edi

        mov     ecx, eax        // ecx = @a
        or      eax, -1         // Result = -1

        mov     edx, [ecx].T_BNX.Count
        cmp     edx, 0          // edx = a.Count
        jle     @R

        mov     esi, [ecx].T_BNX.PHData
  @A:                           // for I := HIndex downto 0 ...
        mov     edi, [esi]      // edi = a.Data[I]
        test    edi, edi
        jz      @C
  @B:
//      xor     eax, eax
        bsr     eax, edi
        dec     edx
        shl     edx, CXBiSh     // edx = c * CXBits
        add     eax, edx
        jmp     @R
  @C:
  @D:
        sub     esi, CXBytes
        dec     edx
        jnz     @A

  @R:
        pop     edi
        pop     esi
end;
{$ENDIF}

{
       
  The number of bytes excluding the upper zeros
}
function BN_Bytes (AValue: UIntX): UIntX;
var
  VBits : UIntX;
begin
  VBits := BN_Bits (AValue);
  Result := cdiv (VBits, 8);
end;

{
       
  The number of half-bytes excluding the upper zeros
}
function BN_Nibbles (AValue: UIntX): UIntX;
var
  VBits : UIntX;
begin
  VBits := BN_Bits (AValue);
  Result := cdiv (VBits, 4);
end;

{
       
  The number of bits excluding the upper zeros
}
function BN_Bits (AValue: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | a |
asm
       .NOFRAME

        or      rax, -1
        bsr     rax, rcx
        inc     rax
end;
{$ELSE}
// eax | a | ecx
asm
        mov     ecx, eax
        or      eax, -1
        bsr     eax, ecx
        inc     eax
end;
{$ENDIF}

// [+] 1.2.1
{
       
  The number of bits excluding the upper zeros
}
function BN_Bits (const a: T_BNX): UIntX;
var
  VCount : IntX;
begin
  Result := 0;
  if BN_IsZero (a, VCount) then Exit;

  Result := BN_Bits (a.Data[VCount - 1]);

  Dec (VCount);
  if VCount = 0 then Exit;

  Result := Result + UIntX(VCount) shl CXBiSh; // ... Result + VCount * CXBits
end;

{
}
function BN_Size (ACount: IntX): UIntX;
begin
  if ACount <= 0 then begin
    Result := 0;
    Exit;
  end;

  Result := UIntX(ACount) shl CXBySh; // ACount * CXBytes
end;

{
}
function BN_Size (ACount: UIntX): UIntX;
begin
  if ACount = 0 then begin
    Result := 0;
    Exit;
  end;

  Result := ACount shl CXBySh; // ACount * CXBytes
end;

{
}
function BN_IncPointer (P: Pointer): Pointer;
{$IFDEF CPUX64}
// rcx | P
asm
       .NOFRAME

        lea     rax, [rcx + CXBytes]
end;
{$ELSE}
// eax | P
asm
        lea     eax, [eax + CXBytes]
end;
{$ENDIF}

{
}
function BN_IncPointer_(var P): Pointer;
{$IFDEF CPUX64}
// rcx | @P
asm
       .NOFRAME

        mov     rax, [rcx]
        lea     rax, [rax + CXBytes]
        mov     [rcx], rax
end;
{$ELSE}
// eax | @P | ecx
asm
        mov     ecx, eax
        mov     eax, [ecx]
        lea     eax, [eax + CXBytes]
        mov     [ecx], eax
end;
{$ENDIF}

{
}
function BN_DecPointer (P: Pointer): Pointer;
{$IFDEF CPUX64}
// rcx | P
asm
       .NOFRAME

        lea     rax, [rcx - CXBytes]
end;
{$ELSE}
// eax | P
asm
        lea     eax, [eax - CXBytes]
end;
{$ENDIF}

{
}
function BN_DecPointer_(var P): Pointer;
{$IFDEF CPUX64}
// rcx | @P
asm
       .NOFRAME

        mov     rax, [rcx]
        lea     rax, [rax - CXBytes]
        mov     [rcx], rax
end;
{$ELSE}
// eax | @P | ecx
asm
        mov     ecx, eax
        mov     eax, [ecx]
        lea     eax, [eax - CXBytes]
        mov     [ecx], eax
end;
{$ENDIF}

{
  ascending address order
}
procedure BN_ZeroData (DST: Pointer; Count: IntX);
{$IFDEF CPUX64}
// rcx | DST   |
// rdx | Count |
//     | 0     | rax
asm
       .NOFRAME

        dec     rdx
        jl      @r // Count <= 0
        xor     rax, rax
  @loop:
        mov     [rcx], rax // DST = 0
        lea     rcx, [rcx + CXBytes]
        dec     rdx
        jge     @loop
  @r:
end;
{$ELSE}
// eax | DST   |
// edx | Count |
//     | 0     | ecx
asm
        dec     edx
        jl      @r // Count <= 0
        xor     ecx, ecx
  @loop:
        mov     [eax], ecx // DST = 0
        lea     eax, [eax + CXBytes]
        dec     edx
        jge     @loop
  @r:
end;
{$ENDIF}

{
  descending address order
}
procedure BN_ZeroDataDesc (DST: Pointer; Count: IntX);
{$IFDEF CPUX64}
// rcx | DST   |
// rdx | Count |
//     | 0     | rax
asm
       .NOFRAME

        dec     rdx
        jl      @r // Count <= 0
        xor     rax, rax
  @loop:
        mov     [rcx], rax // DST = 0
        lea     rcx, [rcx - CXBytes]
        dec     rdx
        jge     @loop
  @r:
end;
{$ELSE}
// eax | DST   |
// edx | Count |
//     | 0     | ecx
asm
        dec     edx
        jl      @r // Count <= 0
        xor     ecx, ecx
  @loop:
        mov     [eax], ecx // DST = 0
        lea     eax, [eax - CXBytes]
        dec     edx
        jge     @loop
  @r:
end;
{$ENDIF}

{
  ascending address order
}
procedure BN_FillData (DST: Pointer; Count: IntX; V: UIntX);
{$IFDEF CPUX64}
// rcx | DST
// rdx | Count
// r8  | V
asm
       .NOFRAME

        dec     rdx
        jl      @r // Count <= 0
  @loop:
        mov     [rcx], r8 // DST = V
        lea     rcx, [rcx + CXBytes]
        dec     rdx
        jge     @loop
  @r:
end;
{$ELSE}
// eax | DST
// edx | Count
// ecx | V
asm
        dec     edx
        jl      @r // Count <= 0
  @loop:
        mov     [eax], ecx // DST = V
        lea     eax, [eax + CXBytes]
        dec     edx
        jge     @loop
  @r:
end;
{$ENDIF}

{
  descending address order
}
procedure BN_FillDataDesc (DST: Pointer; Count: IntX; V: UIntX);
{$IFDEF CPUX64}
// rcx | DST
// rdx | Count
// r8  | V
asm
       .NOFRAME

        dec     rdx
        jl      @r // Count <= 0
  @loop:
        mov     [rcx], r8 // DST = V
        lea     rcx, [rcx - CXBytes]
        dec     rdx
        jge     @loop
  @r:
end;
{$ELSE}
// eax | DST
// edx | Count
// ecx | V
asm
        dec     edx
        jl      @r // Count <= 0
  @loop:
        mov     [eax], ecx // DST = V
        lea     eax, [eax - CXBytes]
        dec     edx
        jge     @loop
  @r:
end;
{$ENDIF}

{
}
procedure BN_CopyData_(DST, SRC: Pointer; Count: IntX);
var
  VSize : IntX;
begin
  if Count <= 0 then Exit;
  VSize := Count shl CXBySh;
  move (SRC^, DST^, VSize);
end;

{
  ascending address order
}
procedure BN_CopyData (DST, SRC: Pointer; Count: IntX);
{$IFDEF CPUX64}
// rcx | DST
// rdx | SRC
// r8  | Count
asm
       .NOFRAME

        dec     r8
        jl      @r // Count <= 0
  @loop:
        mov     rax, [rdx] // rax = SRC
        mov     [rcx], rax // DST = SRC
        lea     rcx, [rcx + CXBytes]
        lea     rdx, [rdx + CXBytes]
        dec     r8
        jge     @loop
  @r:
end;
{$ELSE}
// eax | DST
// edx | SRC
// ecx | Count
asm
        push    ebx

        dec     ecx
        jl      @r // Count <= 0
  @loop:
        mov     ebx, [edx] // ebx = SRC
        mov     [eax], ebx // DST = SRC
        lea     eax, [eax + CXBytes]
        lea     edx, [edx + CXBytes]
        dec     ecx
        jge     @loop
  @r:
        pop     ebx
end;
{$ENDIF}

{
  descending address order
}
procedure BN_CopyDataDesc (DST, SRC: Pointer; Count: IntX);
{$IFDEF CPUX64}
// rcx | DST
// rdx | SRC
// r8  | Count
asm
       .NOFRAME

        dec     r8
        jl      @r // Count <= 0
  @loop:
        mov     rax, [rdx] // rax = SRC
        mov     [rcx], rax // DST = SRC
        lea     rcx, [rcx - CXBytes]
        lea     rdx, [rdx - CXBytes]
        dec     r8
        jge     @loop
  @r:
end;
{$ELSE}
// eax | DST
// edx | SRC
// ecx | Count
asm
        push    ebx

        dec     ecx
        jl      @r // Count <= 0
  @loop:
        mov     ebx, [edx] // ebx = SRC
        mov     [eax], ebx // DST = SRC
        lea     eax, [eax - CXBytes]
        lea     edx, [edx - CXBytes]
        dec     ecx
        jge     @loop
  @r:
        pop     ebx
end;
{$ENDIF}

// [+] 1.2.1
{
}
function Break_(ABreak: PBoolean): Boolean;
begin
  Result := false;
  if ABreak = nil then Exit;
  Result := ABreak^;
end;

end.

