unit mxBNMath;

// Binuar 1.2.1
// Big integer number arithmetic
// ---
// Copyright  2000-3000, Andrey A. Meshkov (AL-CHEMIST)
// All rights reserved
//
// http://maalchemist.ru
// maa@maalchemist.ru
// maalchemist@yandex.ru
// maalchemist@gmail.com
//
// Free for any use.
// No warranty of any kind.
// ---

{$I version.inc}
{$I mxBNMath.inc}

interface

uses
  mxBNMathTypes;

type
  T_BN_copy  = function (var V: T_BNX; const A: T_BNX): IntX;
  T_BN_scopy = function (var V: T_BNX; ASign: ShortInt; const A: T_BNX): IntX;
  T_BN_acopy = function (var V: T_BNX; const A: T_BNX): IntX;
  T_BN_ncopy = function (var V: T_BNX; const A: T_BNX): IntX;
  T_BN_ccopy = function (var V: T_BNX; const A: T_BNX; ACount: IntX): IntX;
  T_BN_hcopy = function (var V: T_BNX; const A: T_BNX; AIndex: IntX): IntX;

function  BN_copy (var V: T_BNX; const A: T_BNX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_move (var V, A: T_BNX): IntX; // {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_swap (var A, B: T_BNX): IntX; overload; // {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_swap (var A, B: T_BNX; var imv: T_BNX): IntX; overload; // {$IFDEF BN_INLINE} inline; {$ENDIF}

function  BN_scopy (var V: T_BNX; ASign: ShortInt; const A: T_BNX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_acopy (var V: T_BNX; const A: T_BNX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_ncopy (var V: T_BNX; const A: T_BNX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}

function  BN_ccopy (var V: T_BNX; const A: T_BNX; ACount: IntX): IntX; // {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_hcopy (var V: T_BNX; const A: T_BNX; AIndex: IntX): IntX; overload; // {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_hcopy (var V: T_BNX; const A: T_BNX; AIndex: IntX; ACount: IntX): IntX; overload; // {$IFDEF BN_INLINE} inline; {$ENDIF}

procedure BN_split (var A, B: T_BNX; const C: T_BNX; AIndex: IntX); overload; // {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_split (var A: UIntX; var B: T_BNX; const C: T_BNX); overload; // {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_split (var A: T_BNX; var B: UIntX; const C: T_BNX); overload; // {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_split_n2 (var AL, AH: T_BNX; const A: T_BNX); // {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_split_n4 (var A1, A2, A3, A4: T_BNX; const A: T_BNX); // {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_merge (var V: T_BNX; const A, B: T_BNX): IntX; overload; // {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_merge (var V: T_BNX; const A: T_BNX; B: UIntX): IntX; overload; // {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_merge (var V: T_BNX; A: UIntX; const B: T_BNX): IntX; overload; // {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_merge (var V: T_BNX; const A: T_BNX_Array): IntX; overload; // {$IFDEF BN_INLINE} inline; {$ENDIF}

procedure BN_abs (var V: T_BNX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_abs (var V: T_BNX; const A: T_BNX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_nabs (var V: T_BNX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_nabs (var V: T_BNX; const A: T_BNX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_chs (var V: T_BNX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_chs (var V: T_BNX; const A: T_BNX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}

procedure BN_not (var V: T_BNX; const A: T_BNX); overload;
procedure BN_not (var A: T_BNX); overload;
procedure BN_or (var V: T_BNX; const A, B: T_BNX); overload;
procedure BN_or (var A: T_BNX; const B: T_BNX); overload;
procedure BN_xor (var V: T_BNX; const A, B: T_BNX); overload;
procedure BN_xor (var A: T_BNX; const B: T_BNX); overload;
procedure BN_and (var V: T_BNX; const A, B: T_BNX); overload;
procedure BN_and (var A: T_BNX; const B: T_BNX); overload;

procedure BN_pshld (var V: T_BNX; ABits: IntX);
procedure BN_shld (var V: T_BNX; ABits: IntX);
procedure BN_pshrd (var V: T_BNX; ABits: IntX);
procedure BN_shrd (var V: T_BNX; ABits: IntX);

function  BN_shl_1 (var V: T_BNX; const A: T_BNX): IntX; overload;
function  BN_shl_1 (var V: T_BNX): IntX; overload;
function  BN_shr_1 (var V: T_BNX; const A: T_BNX): IntX; overload;
function  BN_shr_1 (var V: T_BNX): IntX; overload;

function  BN_shl_2 (var V: T_BNX; const A: T_BNX): IntX; overload;
function  BN_shl_2 (var V: T_BNX): IntX; overload;
function  BN_shr_2 (var V: T_BNX; const A: T_BNX): IntX; overload;
function  BN_shr_2 (var V: T_BNX): IntX; overload;

procedure BN_pshl (var b: T_BNX; const a: T_BNX; ABits: IntX); overload;
procedure BN_pshl (var a: T_BNX; ABits: IntX); overload;
procedure BN_shl (var b: T_BNX; const a: T_BNX; ABits: IntX); overload;
procedure BN_shl (var a: T_BNX; ABits: IntX); overload;
procedure BN_pshr (var b: T_BNX; const a: T_BNX; ABits: IntX); overload;
procedure BN_pshr (var a: T_BNX; ABits: IntX); overload;
procedure BN_shr (var b: T_BNX; const a: T_BNX; ABits: IntX); overload;
procedure BN_shr (var a: T_BNX; ABits: IntX); overload;

procedure BN_pcshl (var b: T_BNX; const a: T_BNX; C: IntX); overload;
procedure BN_pcshl (var a: T_BNX; C: IntX); overload;
procedure BN_cshl (var b: T_BNX; const a: T_BNX; C: IntX); overload;
procedure BN_cshl (var a: T_BNX; C: IntX); overload;
procedure BN_pcshr (var b: T_BNX; const a: T_BNX; C: IntX); overload;
procedure BN_pcshr (var a: T_BNX; C: IntX); overload;
procedure BN_cshr (var b: T_BNX; const a: T_BNX; C: IntX); overload;
procedure BN_cshr (var a: T_BNX; C: IntX); overload;

procedure BN_add (var a3: T_BNX; const a1, a2: T_BNX);
procedure BN_add_(var a1: T_BNX; const a2: T_BNX); overload;
procedure BN_add_(var a1: T_BNX; const a2: T_BNX; var imv: T_BNX); overload;
procedure BN_add_int (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX); overload;
procedure BN_add_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX); overload;
procedure BN_add_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX; var imv: T_BNX); overload;
procedure BN_add_int (var a3: T_BNX; const a1: T_BNX; a2: IntX); overload;
procedure BN_add_int_(var a1: T_BNX; a2: IntX); overload;
procedure BN_inc (var a3: T_BNX; const a1: T_BNX); overload;
procedure BN_inc_(var a: T_BNX); overload;
procedure BN_inc_(var a: T_BNX; var imv: T_BNX); overload;

procedure BN_sub (var a3: T_BNX; const a1, a2: T_BNX);
procedure BN_sub_(var a1: T_BNX; const a2: T_BNX); overload;
procedure BN_sub_(var a1: T_BNX; const a2: T_BNX; var imv: T_BNX); overload;
procedure BN_sub_int (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX); overload;
procedure BN_sub_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX); overload;
procedure BN_sub_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX; var imv: T_BNX); overload;
procedure BN_sub_int (var a3: T_BNX; const a1: T_BNX; a2: IntX); overload;
procedure BN_sub_int_(var a1: T_BNX; a2: IntX); overload;
procedure BN_dec (var a3: T_BNX; const a1: T_BNX); overload;
procedure BN_dec_(var a: T_BNX); overload;
procedure BN_dec_(var a: T_BNX; var imv: T_BNX); overload;

function  BN_mul_div_sign (ASign1, ASign2: ShortInt): ShortInt; {$IFDEF BN_INLINE} inline; {$ENDIF}

procedure BN_mul_base (var a3: T_BNX; const a1, a2: T_BNX);
procedure BN_mul_base_(var a1: T_BNX; const a2: T_BNX); overload;
procedure BN_mul_base_(var a1: T_BNX; const a2: T_BNX; var imv: T_BNX); overload;
procedure BN_mul (var a3: T_BNX; const a1, a2: T_BNX);
procedure BN_mul_(var a1: T_BNX; const a2: T_BNX);
procedure BN_mul_int (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX); overload;
procedure BN_mul_int (var a3: T_BNX; const a1: T_BNX; a2: IntX); overload;
procedure BN_mul_int_nz (var a3: T_BNX; const a1: T_BNX; a2: UIntX; ACount1: IntX; ASign3: ShortInt); overload;
procedure BN_mul_int_nz_(var a1: T_BNX; a2: UIntX; ACount1: IntX; ASign3: ShortInt); overload;
procedure BN_mul_intn (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX; n: IntX); overload;
procedure BN_mul_intn (var a3: T_BNX; const a1: T_BNX; a2: IntX; n: IntX); overload;

procedure BN_div_mod (var adiv, amod: T_BNX; const a1, a2: T_BNX);
procedure BN_div (var adiv: T_BNX; const a1, a2: T_BNX);
procedure BN_mod (var amod: T_BNX; const a1, a2: T_BNX);

function  BN_div_mod_int (var adiv: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX; APack: Boolean = true): UIntX; overload;
function  BN_div_mod_int (var adiv: T_BNX; const a1: T_BNX; a2: IntX; APack: Boolean = true): UIntX; overload;
function  BN_div_mod_intn (var adiv: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX; n: IntX): UIntX; overload;
function  BN_div_mod_intn (var adiv: T_BNX; const a1: T_BNX; a2: IntX; n: IntX): UIntX; overload;
procedure BN_div_int (var adiv: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX); overload;
procedure BN_div_int (var adiv: T_BNX; const a1: T_BNX; a2: IntX); overload;
procedure BN_div_intn (var adiv: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX; n: IntX); overload;
procedure BN_div_intn (var adiv: T_BNX; const a1: T_BNX; a2: IntX; n: IntX); overload;
function  BN_mod_int (const a1: T_BNX; ASign2: ShortInt; a2: UIntX): UIntX; overload;
function  BN_mod_int (const a1: T_BNX; a2: IntX): UIntX; overload;
function  BN_mod_intn (const a1: T_BNX; ASign2: ShortInt; a2: UIntX; n: IntX): UIntX; overload;
function  BN_mod_intn (const a1: T_BNX; a2: IntX; n: IntX): UIntX; overload;

procedure BN_div_mod_2n (var adiv, amod: T_BNX; const a: T_BNX; n: IntX);
procedure BN_div_2n (var adiv: T_BNX; const a: T_BNX; n: IntX); overload;
procedure BN_div_2n_(var a: T_BNX; n: IntX); overload;
procedure BN_mod_2n (var amod: T_BNX; const a: T_BNX; n: IntX); overload;
function  BN_mod_2n (const a: T_BNX; n: IntX): UIntX; overload;
procedure BN_mod_2n_(var a: T_BNX; n: IntX); overload;

procedure BN_exp2 (var P: T_BNX; ASign: ShortInt; AExp: IntX);
procedure BN_exp10 (var P: T_BNX; ASign: ShortInt; AExp: IntX);

procedure BN_pow_base (var P: T_BNX; const ABase: T_BNX; AExp: UIntX); overload;
procedure BN_pow_base (var P: T_BNX; const ABase, AExp: T_BNX); overload;
procedure BN_pow (var P: T_BNX; const ABase: T_BNX; AExp: UIntX); overload;
procedure BN_pow (var P: T_BNX; const ABase, AExp: T_BNX); overload;
procedure BN_pow_int (var P: T_BNX; ASign: ShortInt; ABase: UIntX; AExp: UIntX); overload;
procedure BN_pow_int (var P: T_BNX; ABase: IntX; AExp: UIntX); overload;
procedure BN_sqr_base (var P: T_BNX; const ABase: T_BNX);
procedure BN_sqr_base_(var ABase: T_BNX); overload;
procedure BN_sqr_base_(var ABase: T_BNX; var imv: T_BNX); overload;
procedure BN_sqr (var P: T_BNX; const ABase: T_BNX);
procedure BN_sqr_(var ABase: T_BNX); overload;
procedure BN_sqr_(var ABase: T_BNX; var imv: T_BNX); overload;

function  BN_ppsqr (const s: T_BNX): Boolean;
function  BN_ppp2m (const s: T_BNX; m: UIntX): Boolean;

function  BN_root (var aroot: T_BNX; const a: T_BNX; n: UIntX): Integer;
function  BN_root_rem (var aroot, arem: T_BNX; const a: T_BNX; n: UIntX): Integer;
function  BN_sqrt (var aroot: T_BNX; const a: T_BNX): Integer;
function  BN_sqrt_rem (var aroot, arem: T_BNX; const a: T_BNX): Integer;

function  BN_dsqrt (var aroot: T_BNX; const a: T_BNX): Integer;
function  BN_rnd_dsqrt (var aroot: T_BNX; const a: T_BNX): Integer;
function  BN_dsqrt_rem (var aroot, arem: T_BNX; const a: T_BNX): Integer;
function  BN_rnd_dsqrt_rem (var aroot, arem: T_BNX; const a: T_BNX): Integer;

function  BN_m2n (const a: T_BNX): UIntX;
function  BN_m2n_(const a: T_BNX; var sha: T_BNX): UIntX; overload;
function  BN_m2n_(var a: T_BNX): UIntX; overload;
function  BN_m2n_int (a: UIntX): UIntX;
function  BN_m2n_int_(a: UIntX; var sha: UIntX): UIntX;

procedure BN_gcd (var R: T_BNX; const m, n: T_BNX);
procedure BN_gcd_bin (var R: T_BNX; const m, n: T_BNX);
procedure BN_gcd_mod (var R: T_BNX; const m, n: T_BNX);
procedure BN_gcd_sub (var R: T_BNX; const m, n: T_BNX);

function  BN_coprime (const a, b: T_BNX): Boolean;

procedure BN_lcm (var R: T_BNX; const m, n: T_BNX);
procedure BN_lcm_bin (var R: T_BNX; const m, n: T_BNX);

procedure BN_gcd_bezout (var g, x, y: T_BNX; const a, b: T_BNX);
procedure BN_bezout (var x, y: T_BNX; const a, b: T_BNX);

function  BN_jacobi (const a, n: T_BNX): Integer;

function  BN_inv_mod (var a_inv: T_BNX; const a, b: T_BNX): Boolean;
function  BN_sqrt_mod (var R: T_BNX; const a, p: T_BNX; AVerify: Boolean): Boolean;

procedure BN_pow_mod (var P: T_BNX; const a, e, n: T_BNX); overload;
procedure BN_pow_mod (var P: T_BNX; const a: T_BNX; e: UIntX; const n: T_BNX); overload;
function  BN_pow_mod_mon (var P: T_BNX; const a, e, n: T_BNX): Boolean;

function  BN_inv (var ainv: T_BNX; const A: T_BNX): IntX; overload;
function  BN_inv (var ainv: T_BNX; A: UIntX): IntX; overload;
function  BN_ninv (var ainv: T_BNX; const A: T_BNX): IntX;
function  BN_cinv (var ainv: T_BNX; const A: T_BNX): IntX;

procedure BN_fact (var R: T_BNX; const a: T_BNX); overload;
procedure BN_fact (var R: T_BNX; a: UIntX); overload;

var // [+] 1.2.1
  BN_dup  : T_BN_copy  = BN_copy;
  BN_sdup : T_BN_scopy = BN_scopy;
  BN_adup : T_BN_acopy = BN_acopy;
  BN_ndup : T_BN_ncopy = BN_ncopy;
  BN_cdup : T_BN_ccopy = BN_ccopy;
  BN_hdup : T_BN_hcopy = BN_hcopy;

implementation

uses
  SysConst, SysUtils, mxBNMathUtils, mxBNMathProcs;

{
  V = A
}
function BN_copy (var V: T_BNX; const A: T_BNX): IntX;
begin
  (*
  if @V = @A then begin
    Result := V.Count;
    Exit;
  end;
  // *)

  Result := BN_Alloc (V, A.Sign, A.Count);
  if Result <= 0 then Exit;

  BN_CopyData (V.PLData, A.PLData, V.Count);
end;

// [+] 1.2.1
{
  V = A
  Free(A)
}
function BN_move (var V, A: T_BNX): IntX;
begin
  Result := BN_Alloc (V, A.Sign, A.Count);
  if Result <= 0 then begin
    BN_Free (A);
    Exit;
  end;

  BN_CopyData (V.PLData, A.PLData, V.Count);
  BN_Free (A);
end;

// [+] 1.2.1
{
  A <-> B
}
function BN_swap (var A, B: T_BNX): IntX;
var
  V : T_BNX;
begin
  BN_copy (V, A);
  Result := BN_copy (A, B);
  BN_copy (B, V);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  A <-> B
  imv - external intermediate variable
}
function BN_swap (var A, B: T_BNX; var imv: T_BNX): IntX;
begin
  BN_copy (imv, A);
  Result := BN_copy (A, B);
  BN_copy (B, imv);
end;

{
  |V| = |A|
  V.Sign = ASign
}
function BN_scopy (var V: T_BNX; ASign: ShortInt; const A: T_BNX): IntX;
begin
  Result := BN_Alloc (V, ASign, A.Count);
  if Result <= 0 then Exit;

  BN_CopyData (V.PLData, A.PLData, V.Count);
end;

// [+] 1.2.1
{
  V = |A|
}
function BN_acopy (var V: T_BNX; const A: T_BNX): IntX;
var
  VSign : ShortInt;
begin
  VSign := A.Sign;
  if VSign <> 0 then VSign := 1;

  Result := BN_Alloc (V, VSign, A.Count);
  if Result <= 0 then Exit;

  BN_CopyData (V.PLData, A.PLData, V.Count);
end;

{
  V = -A
}
function BN_ncopy (var V: T_BNX; const A: T_BNX): IntX;
var
  VSign : ShortInt;
begin
  VSign := A.Sign;
  if VSign <> 0 then VSign := -VSign;

  Result := BN_Alloc (V, VSign, A.Count);
  if Result <= 0 then Exit;

  BN_CopyData (V.PLData, A.PLData, V.Count);
end;

{
  V <- A
    ACount 
  copy the first ACount elements
}
function BN_ccopy (var V: T_BNX; const A: T_BNX; ACount: IntX): IntX;
var
  VCount  : IntX;
  VDCount : IntX;
begin
  Result := BN_Alloc (V, A.Sign, ACount);
  if Result <= 0 then Exit;

  VCount := A.Count;
  VDCount := V.Count - VCount;
  if VDCount > 0 then begin
    BN_CopyData (V.PLData, A.PLData, VCount);
    BN_ZeroData (@V.Data[VCount], VDCount);
  end else begin
    BN_CopyData (V.PLData, A.PLData, V.Count);
  end;
end;

{
  V <- A
  ,    #AIndex
  copy, starting from the element #AIndex
}
function BN_hcopy (var V: T_BNX; const A: T_BNX; AIndex: IntX): IntX;
var
  VCount : IntX;
begin
  if (AIndex < 0) or (AIndex >= A.Count) or (A.Sign = 0) then begin
    Result := BN_AssignZero (V);
    Exit;
  end;

  VCount := A.Count - AIndex;
  if VCount <= 0 then begin
    Result := BN_AssignZero (V);
    Exit;
  end;

  Result := BN_Alloc (V, A.Sign, VCount);
  BN_CopyData (V.PLData, @A.Data[AIndex], VCount);
end;

// [+] 1.2.1
{
  V <- A
   ACount ,    #AIndex
  copy ACount elements, starting from the element #AIndex
}
function BN_hcopy (var V: T_BNX; const A: T_BNX; AIndex: IntX; ACount: IntX): IntX;
var
  VCount : IntX;
begin
  if (AIndex < 0) or (AIndex >= A.Count) or (A.Sign = 0) or (ACount <= 0) then begin
    Result := BN_AssignZero (V);
    Exit;
  end;

  VCount := ACount;
  if VCount + AIndex > A.Count then VCount := A.Count - AIndex;

  if VCount <= 0 then begin
    Result := BN_AssignZero (V);
    Exit;
  end;

  Result := BN_Init (V, A.Sign, ACount);
  BN_CopyData (V.PLData, @A.Data[AIndex], VCount);
end;

// [+] 1.2.1
{
  [A, B] <- C
}
procedure BN_split (var A, B: T_BNX; const C: T_BNX; AIndex: IntX);
var
  I      : LoopInt;
  VCount : IntX;
begin
  VCount := C.Count;
  if VCount <= 0 then begin
    BN_AssignZero (A);
    BN_AssignZero (B);
    Exit;
  end;

  if AIndex <= 0 then begin
    BN_AssignZero (A);
    BN_copy (B, C);
    Exit;
  end;

  if AIndex > C.HIndex then begin
    BN_copy (A, C);
    BN_AssignZero (B);
    Exit;
  end;

  VCount := AIndex;
  BN_Alloc (A, C.Sign, VCount);
  for I := 0 to VCount - 1 do begin
    A.Data[I] := C.Data[I];
  end;

  VCount := C.Count - VCount;
  BN_Alloc (B, C.Sign, VCount);
  for I := 0 to VCount - 1 do begin
    B.Data[I] := C.Data[AIndex + I];
  end;
end;

// [+] 1.2.1
{
  [A, B] <- C
}
procedure BN_split (var A: UIntX; var B: T_BNX; const C: T_BNX);
var
  I      : LoopInt;
  VCount : IntX;
begin
  VCount := C.Count;
  if VCount <= 0 then begin
    A := 0;
    BN_AssignZero (B);
    Exit;
  end;

  if VCount = 1 then begin
    A := C.Data[0];
    BN_AssignZero (B);
    Exit;
  end;

  A := C.Data[0];

  VCount := VCount - 1;
  BN_Alloc (B, C.Sign, VCount);
  for I := 0 to VCount - 1 do begin
    B.Data[I] := C.Data[1 + I];
  end;
end;

// [+] 1.2.1
{
  [A, B] <- C
}
procedure BN_split (var A: T_BNX; var B: UIntX; const C: T_BNX);
var
  I      : LoopInt;
  VCount : IntX;
begin
  VCount := C.Count;
  if VCount <= 0 then begin
    BN_AssignZero (A);
    B := 0;
    Exit;
  end;

  if VCount = 1 then begin
    BN_AssignZero (A);
    B := C.Data[0];
    Exit;
  end;

  VCount := VCount - 1;
  BN_Alloc (A, C.Sign, VCount);
  for I := 0 to VCount - 1 do begin
    A.Data[I] := C.Data[I];
  end;

  B := C.Data[VCount];
end;

// [+] 1.2.1
{
  [AL, AH] <- A
}
procedure BN_split_n2 (var AL, AH: T_BNX; const A: T_BNX);
var
  VIndex : IntX;
begin
  VIndex := A.Count shr 1; // Count div 2
  BN_split (AL, AH, A, VIndex);
end;

// [+] 1.2.1
{
  [A1, A2, A3, A4] <- A
}
procedure BN_split_n4 (var A1, A2, A3, A4: T_BNX; const A: T_BNX);
var
  V_AL : T_BNX;
  V_AH : T_BNX;
begin
  BN_split_n2 (V_AL, V_AH, A);
  BN_split_n2 (A1, A2, V_AL);
  BN_split_n2 (A3, A4, V_AH);

  BN_LFree (V_AL);
  BN_LFree (V_AH);
end;

// [+] 1.2.1
{
  V = [A, B]
  V.Sign = 1
}
function BN_merge (var V: T_BNX; const A, B: T_BNX): IntX;
var
  I, P    : LoopInt;
  VCount  : IntX;
  VCountA : IntX;
  VCountB : IntX;
begin
  VCountA := A.Count;
  VCountB := B.Count;

  if VCountA <= 0 then begin
    if VCountB <= 0 then begin
      // (VCountA <= 0) and (VCountB <= 0)
      Result := BN_AssignZero (V);
      Exit;
    end else begin
      // (VCountA <= 0) and (VCountB > 0)
      Result := BN_copy (V, B);
      Exit;
    end;
  end else begin
    if VCountB <= 0 then begin
      // (VCountB <= 0) and (VCountA > 0)
      Result := BN_copy (V, A);
      Exit;
    end;
  end;

  VCount := VCountA + VCountB;
  Result := VCount;

  BN_Alloc (V, VCount);
  P := 0;
  for I := 0 to VCountA - 1 do begin
    V.Data[P] := A.Data[I];
    Inc (P);
  end;
  for I := 0 to VCountB - 1 do begin
    V.Data[P] := B.Data[I];
    Inc (P);
  end;
end;

// [+] 1.2.1
{
  V = [A, B]
  V.Sign = 1
}
function BN_merge (var V: T_BNX; const A: T_BNX; B: UIntX): IntX;
var
  I, P    : LoopInt;
  VCount  : IntX;
  VCountA : IntX;
begin
  VCountA := A.Count;
  if VCountA <= 0 then begin
    Result := BN_AssignInt (V, 1, B);
    Exit;
  end;

  VCount := VCountA + 1;
  Result := VCount;

  BN_Alloc (V, VCount);
  P := 0;
  for I := 0 to VCountA - 1 do begin
    V.Data[P] := A.Data[I];
    Inc (P);
  end;
  V.Data[P] := B;
end;

// [+] 1.2.1
{
  V = [A, B]
  V.Sign = 1
}
function BN_merge (var V: T_BNX; A: UIntX; const B: T_BNX): IntX;
var
  I, P    : LoopInt;
  VCount  : IntX;
  VCountB : IntX;
begin
  VCountB := B.Count;
  if VCountB <= 0 then begin
    Result := BN_AssignInt (V, 1, A);
    Exit;
  end;

  VCount := 1 + VCountB;
  Result := VCount;

  BN_Alloc (V, VCount);
  V.Data[0] := A;
  P := 1;
  for I := 0 to VCountB - 1 do begin
    V.Data[P] := B.Data[I];
    Inc (P);
  end;
end;

// [+] 1.2.1
{
  V = [A1, ..., A0]
  V.Sign = 1
}
function BN_merge (var V: T_BNX; const A: T_BNX_Array): IntX;
var
  I, J, P : LoopInt;
  VACount : IntX;
  VCount  : IntX;
begin
  VCount := 0;
  for I := 0 to HIGH(A) do begin
    VACount := A[I].Count;
    if VACount > 0 then Inc (VCount, VACount);
  end;
  Result := VCount;

  BN_Alloc (V, VCount);
  P := 0;
  for I := 0 to HIGH(A) do begin
    VACount := A[I].Count;
    if VACount <= 0 then Continue;

    for J := 0 to VACount - 1 do begin
      V.Data[P] := A[I].Data[J];
      Inc (P);
    end;
  end;
end;

// [+] 1.2.1
{
  Absolute value
  V = |V|
}
procedure BN_abs (var V: T_BNX);
begin
  if V.Sign = 0 then Exit;
  V.Sign := +1;
end;

// [+] 1.2.1
{
  Absolute value
  V = |A|
}
procedure BN_abs (var V: T_BNX; const A: T_BNX);
begin
  BN_copy (V, A);
  if V.Sign = 0 then Exit;
  V.Sign := +1;
end;

// [+] 1.2.1
{
  Negative absolute value
  V = -|V|
}
procedure BN_nabs (var V: T_BNX);
begin
  if V.Sign = 0 then Exit;
  V.Sign := -1;
end;

// [+] 1.2.1
{
  Negative absolute value
  V = -|A|
}
procedure BN_nabs (var V: T_BNX; const A: T_BNX);
begin
  BN_copy (V, A);
  if V.Sign = 0 then Exit;
  V.Sign := -1;
end;

// [+] 1.2.1
{
  Change sign
  V = -V
}
procedure BN_chs (var V: T_BNX);
begin
  if V.Sign = 0 then Exit;
  V.Sign := -V.Sign;
end;

// [+] 1.2.1
{
  Change sign
  V = -A
}
procedure BN_chs (var V: T_BNX; const A: T_BNX);
begin
  BN_copy (V, A);
  if V.Sign = 0 then Exit;
  V.Sign := -V.Sign;
end;

// [+] 1.2.1
{
  Bitwise not
  V = not A
}
procedure BN_not (var V: T_BNX; const A: T_BNX);
var
  I : LoopInt;
begin
  BN_copy (V, A);

  for I := 0 to A.Count - 1 do begin
    V.Data[I] := not A.Data[I];
  end;
end;

// [+] 1.2.1
{
  Bitwise not
  A = not A
}
procedure BN_not (var A: T_BNX);
var
  I : LoopInt;
begin
  for I := 0 to A.Count - 1 do begin
    A.Data[I] := not A.Data[I];
  end;
end;

// [+] 1.2.1
{
  Bitwise or
  V = A or B
}
procedure BN_or (var V: T_BNX; const A, B: T_BNX); overload;
var
  I : LoopInt;
  C : LoopInt;
begin
  BN_copy (V, A);

  C := A.Count;
  if B.Count < C then C := B.Count;

  for I := 0 to C - 1 do begin
    V.Data[I] := A.Data[I] or B.Data[I];
  end;
end;

// [+] 1.2.1
{
  Bitwise or
  A = A or B
}
procedure BN_or (var A: T_BNX; const B: T_BNX); overload;
var
  I : LoopInt;
  C : LoopInt;
begin
  C := A.Count;
  if B.Count < C then C := B.Count;

  for I := 0 to C - 1 do begin
    A.Data[I] := A.Data[I] or B.Data[I];
  end;
end;

// [+] 1.2.1
{
  Bitwise xor
  V = A xor B
}
procedure BN_xor (var V: T_BNX; const A, B: T_BNX); overload;
var
  I : LoopInt;
  C : LoopInt;
begin
  BN_copy (V, A);

  C := A.Count;
  if B.Count < C then C := B.Count;

  for I := 0 to C - 1 do begin
    V.Data[I] := A.Data[I] xor B.Data[I];
  end;
end;

// [+] 1.2.1
{
  Bitwise xor
  A = A xor B
}
procedure BN_xor (var A: T_BNX; const B: T_BNX); overload;
var
  I : LoopInt;
  C : LoopInt;
begin
  C := A.Count;
  if B.Count < C then C := B.Count;

  for I := 0 to C - 1 do begin
    A.Data[I] := A.Data[I] xor B.Data[I];
  end;
end;

// [+] 1.2.1
{
  Bitwise and
  V = A and B
}
procedure BN_and (var V: T_BNX; const A, B: T_BNX); overload;
var
  I : LoopInt;
  C : LoopInt;
begin
  BN_copy (V, A);

  C := A.Count;
  if B.Count < C then C := B.Count;

  for I := 0 to C - 1 do begin
    V.Data[I] := A.Data[I] and B.Data[I];
  end;
  for I := C to V.Count - 1 do V.Data[I] := 0;
end;

// [+] 1.2.1
{
  Bitwise and
  A = A and B
}
procedure BN_and (var A: T_BNX; const B: T_BNX); overload;
var
  I : LoopInt;
  C : LoopInt;
begin
  C := A.Count;
  if B.Count < C then C := B.Count;

  for I := 0 to C - 1 do begin
    A.Data[I] := A.Data[I] and B.Data[I];
  end;
  for I := C to A.Count - 1 do A.Data[I] := 0;
end;

{
  Result = a div (32|64)
  amod   = a mod (32|64)
  rmod   = (32|64) - amod
}
function BNX_div_mod_x_asm (a: UIntX; var amod: UIntX; var rmod: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | a      |
// rdx | @amod  | rcx
// r8  | @rmod  |
//     | CXBits | r9
asm
       .NOFRAME

        mov     rax, rcx
        mov     rcx, rdx
        xor     rdx, rdx
        mov     r9, CXBits
        div     r9
        mov     [rcx], rdx
        sub     r9, rdx
        mov     [r8], r9
end;
{$ELSE}
// eax | a      |
// edx | @amod  | edi
// ecx | @rmod  |
//     | CXBits | ebx
asm
        push    ebx
        push    edi

        mov     edi, edx
        xor     edx, edx
        mov     ebx, CXBits
        div     ebx
        mov     [edi], edx
        sub     ebx, edx
        mov     [ecx], ebx

        pop     edi
        pop     ebx
end;
{$ENDIF}

// [+] 1.2.1
{
  shift left by ABits bits, 0 < ABits < 32|64, AHIndex >= 1
  V = V shl ABits
}
procedure BN_shld_asm (APHData: PUIntX; AHIndex: IntX; ABits: UIntX);
{$IFDEF CPUX64}
// rcx | APHData | rax
// rdx | AHIndex |
// r8  | ABits   | rcx
asm
       .NOFRAME

        mov     rax, rcx // rax = APHData
        mov     rcx, r8  // rcx = ABits

        mov     r8, [rax]
        mov     r9, [rax - CXBytes]
        shld    r8, r9, cl
        mov     [rax], r8
  @A:
        sub     rax, CXBytes
        dec     rdx
        jz      @B
        mov     r8, r9
        mov     r9, [rax - CXBytes]
        shld    r8, r9, cl
        mov     [rax], r8
        jmp     @A
  @B:
        mov     r8, [rax]
        shl     r8, cl
        mov     [rax], r8
end;
{$ELSE}
// eax | APHData
// edx | AHIndex
// ecx | ABits
asm
        push    esi
        push    edi

        mov     edi, [eax]
        mov     esi, [eax - CXBytes]
        shld    edi, esi, cl
        mov     [eax], edi
  @A:
        sub     eax, CXBytes
        dec     edx
        jz      @B
        mov     edi, esi
        mov     esi, [eax - CXBytes]
        shld    edi, esi, cl
        mov     [eax], edi
        jmp     @A
  @B:
        mov     edi, [eax]
        shl     edi, cl
        mov     [eax], edi

        pop     edi
        pop     esi
end;
{$ENDIF}

// [+] 1.2.1
{
  shift left by 1 bit, AHIndex >= 1
  V = V shl 1
}
procedure BN_shl_1_asm (APHData: PUIntX; AHIndex: IntX);
const
  CSHIVA = 1; // shift value

{$IFDEF CPUX64}
// rcx | APHData | rax
// rdx | AHIndex | rcx
asm
       .NOFRAME

        mov     rax, rcx // rax = APHData
        mov     rcx, rdx // rcx = AHIndex

        mov     r8, [rax]
        mov     r9, [rax - CXBytes]
        shld    r8, r9, CSHIVA
        mov     [rax], r8
  @A:
        sub     rax, CXBytes
        dec     rcx
        jrcxz   @B
        mov     r8, r9
        mov     r9, [rax - CXBytes]
        shld    r8, r9, CSHIVA
        mov     [rax], r8
        jmp     @A
  @B:
        mov     r8, [rax]
        shl     r8, CSHIVA
        mov     [rax], r8
end;
{$ELSE}
// eax | APHData |
// edx | AHIndex | ecx
asm
        push    edi

        mov     ecx, edx // ecx = AHIndex

        mov     edi, [eax]
        mov     edx, [eax - CXBytes]
        shld    edi, edx, CSHIVA
        mov     [eax], edi
  @A:
        sub     eax, CXBytes
        dec     ecx
        jecxz   @B
        mov     edi, edx
        mov     edx, [eax - CXBytes]
        shld    edi, edx, CSHIVA
        mov     [eax], edi
        jmp     @A
  @B:
        mov     edi, [eax]
        shl     edi, CSHIVA
        mov     [eax], edi

        pop     edi
end;
{$ENDIF}

// [+] 1.2.1
{
  shift left by 2 bits, AHIndex >= 1
  V = V shl 2
}
procedure BN_shl_2_asm (APHData: PUIntX; AHIndex: IntX);
const
  CSHIVA = 2; // shift value

{$IFDEF CPUX64}
// rcx | APHData | rax
// rdx | AHIndex | rcx
asm
       .NOFRAME

        mov     rax, rcx // rax = APHData
        mov     rcx, rdx // rcx = AHIndex

        mov     r8, [rax]
        mov     r9, [rax - CXBytes]
        shld    r8, r9, CSHIVA
        mov     [rax], r8
  @A:
        sub     rax, CXBytes
        dec     rcx
        jrcxz   @B
        mov     r8, r9
        mov     r9, [rax - CXBytes]
        shld    r8, r9, CSHIVA
        mov     [rax], r8
        jmp     @A
  @B:
        mov     r8, [rax]
        shl     r8, CSHIVA
        mov     [rax], r8
end;
{$ELSE}
// eax | APHData |
// edx | AHIndex | ecx
asm
        push    edi

        mov     ecx, edx // ecx = AHIndex

        mov     edi, [eax]
        mov     edx, [eax - CXBytes]
        shld    edi, edx, CSHIVA
        mov     [eax], edi
  @A:
        sub     eax, CXBytes
        dec     ecx
        jecxz   @B
        mov     edi, edx
        mov     edx, [eax - CXBytes]
        shld    edi, edx, CSHIVA
        mov     [eax], edi
        jmp     @A
  @B:
        mov     edi, [eax]
        shl     edi, CSHIVA
        mov     [eax], edi

        pop     edi
end;
{$ENDIF}

// [+] 1.2.1
{
  shift left by ABits bits, 0 < ABits < 32|64 (positive)
  V = V shl ABits
}
procedure BN_pshld (var V: T_BNX; ABits: IntX);
begin
  if ABits <= 0 then Exit;
  if V.Count <= 0 then Exit;

  BN_Increase (V);
  BN_shld_asm (V.PHData, V.HIndex, ABits);
  BN_Pack (V);
end;

// [+] 1.2.1
{
  shift left by ABits bits, 0 < |ABits| < 32|64
  V = V shl ABits
}
procedure BN_shld (var V: T_BNX; ABits: IntX);
begin
  if ABits = 0 then Exit;
  if V.Count <= 0 then Exit;

  if ABits > 0 then begin
    BN_pshld (V, ABits);
  end else begin
    BN_pshrd (V, -ABits);
  end;
end;

// [+] 1.2.1
{
  shift right by ABits bits, 0 < ABits < 32|64, AHIndex >= 0
  V = V shr ABits
}
procedure BN_shrd_asm (APLData: PUIntX; AHIndex: IntX; ABits: UIntX);
{$IFDEF CPUX64}
// rcx | APLData | rax
// rdx | AHIndex |
// r8  | ABits   | rcx
asm
       .NOFRAME

        mov     rax, rcx // rax = APLData
        mov     rcx, r8  // rcx = ABits

        test    rdx, rdx
        jz      @B

        mov     r8, [rax]
        mov     r9, [rax + CXBytes]
        shrd    r8, r9, cl
        mov     [rax], r8
  @A:
        add     rax, CXBytes
        dec     rdx
        jz      @B
        mov     r8, r9
        mov     r9, [rax + CXBytes]
        shrd    r8, r9, cl
        mov     [rax], r8
        jmp     @A
  @B:
        mov     r8, [rax]
        shr     r8, cl
        mov     [rax], r8
end;
{$ELSE}
// eax | APLData
// edx | AHIndex
// ecx | ABits
asm
        push    esi
        push    edi

        test    edx, edx
        jz      @B

        mov     edi, [eax]
        mov     esi, [eax + CXBytes]
        shrd    edi, esi, cl
        mov     [eax], edi
  @A:
        add     eax, CXBytes
        dec     edx
        jz      @B
        mov     edi, esi
        mov     esi, [eax + CXBytes]
        shrd    edi, esi, cl
        mov     [eax], edi
        jmp     @A
  @B:
        mov     edi, [eax]
        shr     edi, cl
        mov     [eax], edi

        pop     edi
        pop     esi
end;
{$ENDIF}

// [+] 1.2.1
{
  shift right by 1 bit, AHIndex >= 0
  V = V shr 1
}
procedure BN_shr_1_asm (APLData: PUIntX; AHIndex: IntX);
const
  CSHIVA = 1; // shift value

{$IFDEF CPUX64}
// rcx | APLData | rax
// rdx | AHIndex | rcx
asm
       .NOFRAME

        mov     rax, rcx // rax = APLData
        mov     rcx, rdx // rcx = AHIndex

        jrcxz   @B

        mov     r8, [rax]
        mov     r9, [rax + CXBytes]
        shrd    r8, r9, CSHIVA
        mov     [rax], r8
  @A:
        add     rax, CXBytes
        dec     rcx
        jrcxz   @B
        mov     r8, r9
        mov     r9, [rax + CXBytes]
        shrd    r8, r9, CSHIVA
        mov     [rax], r8
        jmp     @A
  @B:
        mov     r8, [rax]
        shr     r8, CSHIVA
        mov     [rax], r8
end;
{$ELSE}
// eax | APLData |
// edx | AHIndex | ecx
asm
        push    edi

        mov     ecx, edx // ecx = AHIndex
        jecxz   @B

        mov     edi, [eax]
        mov     edx, [eax + CXBytes]
        shrd    edi, edx, CSHIVA
        mov     [eax], edi
  @A:
        add     eax, CXBytes
        dec     ecx
        jecxz   @B
        mov     edi, edx
        mov     edx, [eax + CXBytes]
        shrd    edi, edx, CSHIVA
        mov     [eax], edi
        jmp     @A
  @B:
        mov     edi, [eax]
        shr     edi, CSHIVA
        mov     [eax], edi

        pop     edi
end;
{$ENDIF}

// [+] 1.2.1
{
  shift right by 2 bits, AHIndex >= 0
  V = V shr 2
}
procedure BN_shr_2_asm (APLData: PUIntX; AHIndex: IntX);
const
  CSHIVA = 2; // shift value

{$IFDEF CPUX64}
// rcx | APLData | rax
// rdx | AHIndex | rcx
asm
       .NOFRAME

        mov     rax, rcx // rax = APLData
        mov     rcx, rdx // rcx = AHIndex

        jrcxz   @B

        mov     r8, [rax]
        mov     r9, [rax + CXBytes]
        shrd    r8, r9, CSHIVA
        mov     [rax], r8
  @A:
        add     rax, CXBytes
        dec     rcx
        jrcxz   @B
        mov     r8, r9
        mov     r9, [rax + CXBytes]
        shrd    r8, r9, CSHIVA
        mov     [rax], r8
        jmp     @A
  @B:
        mov     r8, [rax]
        shr     r8, CSHIVA
        mov     [rax], r8
end;
{$ELSE}
// eax | APLData |
// edx | AHIndex | ecx
asm
        push    edi

        mov     ecx, edx // ecx = AHIndex
        jecxz   @B

        mov     edi, [eax]
        mov     edx, [eax + CXBytes]
        shrd    edi, edx, CSHIVA
        mov     [eax], edi
  @A:
        add     eax, CXBytes
        dec     ecx
        jecxz   @B
        mov     edi, edx
        mov     edx, [eax + CXBytes]
        shrd    edi, edx, CSHIVA
        mov     [eax], edi
        jmp     @A
  @B:
        mov     edi, [eax]
        shr     edi, CSHIVA
        mov     [eax], edi

        pop     edi
end;
{$ENDIF}

// [+] 1.2.1
{
  shift right by ABits bits, 0 < ABits < 32|64 (positive)
  V = V shr ABits
}
procedure BN_pshrd (var V: T_BNX; ABits: IntX);
begin
  if ABits <= 0 then Exit;
  if V.Count <= 0 then Exit;

  BN_shrd_asm (V.PLData, V.HIndex, ABits);
  BN_Pack (V);
end;

// [+] 1.2.1
{
  shift right by ABits bits, 0 < |ABits| < 32|64
  V = V shr ABits
}
procedure BN_shrd (var V: T_BNX; ABits: IntX);
begin
  if ABits = 0 then Exit;
  if V.Count <= 0 then Exit;

  if ABits > 0 then begin
    BN_pshrd (V, ABits);
  end else begin
    BN_pshld (V, -ABits);
  end;
end;

// [+] 1.2.1
{
  shift left by 1 bit
  V = A shl 1
}
function BN_shl_1 (var V: T_BNX; const A: T_BNX): IntX; overload;
begin
  Result := BN_copy (V, A);
  if V.Count <= 0 then Exit;

  BN_Increase (V);
  BN_shl_1_asm (V.PHData, V.HIndex);
  Result := BN_Pack (V);
end;

// [+] 1.2.1
{
  shift left by 1 bit
  V = V shl 1
}
function BN_shl_1 (var V: T_BNX): IntX; overload;
begin
  Result := 0;
  if V.Count <= 0 then Exit;

  BN_Increase (V);
  BN_shl_1_asm (V.PHData, V.HIndex);
  Result := BN_Pack (V);
end;

// [+] 1.2.1
{
  shift right by 1 bit
  V = A shr 1
}
function BN_shr_1 (var V: T_BNX; const A: T_BNX): IntX; overload;
begin
  Result := BN_copy (V, A);
  if V.Count <= 0 then Exit;

  BN_shr_1_asm (V.PLData, V.HIndex);
  Result := BN_Pack (V);
end;

// [+] 1.2.1
{
  shift right by 1 bit
  V = V shr 1
}
function BN_shr_1 (var V: T_BNX): IntX; overload;
begin
  Result := 0;
  if V.Count <= 0 then Exit;

  BN_shr_1_asm (V.PLData, V.HIndex);
  Result := BN_Pack (V);
end;

// [+] 1.2.1
{
  shift left by 2 bits
  V = A shl 2
}
function BN_shl_2 (var V: T_BNX; const A: T_BNX): IntX; overload;
begin
  Result := BN_copy (V, A);
  if V.Count <= 0 then Exit;

  BN_Increase (V);
  BN_shl_2_asm (V.PHData, V.HIndex);
  Result := BN_Pack (V);
end;

// [+] 1.2.1
{
  shift left by 2 bits
  V = V shl 2
}
function BN_shl_2 (var V: T_BNX): IntX; overload;
begin
  Result := 0;
  if V.Count <= 0 then Exit;

  BN_Increase (V);
  BN_shl_2_asm (V.PHData, V.HIndex);
  Result := BN_Pack (V);
end;

// [+] 1.2.1
{
  shift right by 2 bits
  V = A shr 2
}
function BN_shr_2 (var V: T_BNX; const A: T_BNX): IntX; overload;
begin
  Result := BN_copy (V, A);
  if V.Count <= 0 then Exit;

  BN_shr_2_asm (V.PLData, V.HIndex);
  Result := BN_Pack (V);
end;

// [+] 1.2.1
{
  shift right by 2 bits
  V = V shr 2
}
function BN_shr_2 (var V: T_BNX): IntX; overload;
begin
  Result := 0;
  if V.Count <= 0 then Exit;

  BN_shr_2_asm (V.PLData, V.HIndex);
  Result := BN_Pack (V);
end;

{
  shift left by ABits bits, ABits > 0 (positive)
  b = a shl ABits
}
procedure BN_pshl (var b: T_BNX; const a: T_BNX; ABits: IntX);
begin
  BN_copy (b, a);
  if ABits <= 0 then Exit;

  BN_pshl (b, ABits);
end;

{
  shift left by ABits bits, ABits > 0 (positive)
  a = a shl ABits
}
procedure BN_pshl (var a: T_BNX; ABits: IntX);
var
  I       : LoopInt;
  P       : IntX;
  C       : IntX;
  CB, CX  : UIntX;
  V, U    : UIntX;
  VCount  : IntX;
  VRCount : IntX; // Result count
begin
  if ABits <= 0 then Exit;
  if a.Count <= 0 then Exit;

  if ABits < CXBits then begin
    BN_pshld (a, ABits);
    Exit;
  end;

  C := BNX_div_mod_x_asm (ABits, CB, CX);

  VCount := BN_DataCount (a);
  if CB = 0 then VRCount := VCount + C else VRCount := VCount + C + 1;
  if VRCount <= VCount then begin
    BN_AssignZero (a);
    Exit;
  end;

  BN_Resize (a, VRCount);

  if CB = 0 then begin
    // VRCount = VCount + C
    if C > 0 then begin
      I := VCount - 1;
      BN_CopyDataDesc (@a.Data[I + C], @a.Data[I], VCount);
      BN_ZeroData (a.PLData, C);
    end;
  end else begin
    // VRCount = VCount + C + 1
    P := VRCount - 1;
    a.Data[P] := 0;

    for I := VCount - 1 downto 0 do begin
      V := a.Data[I] shr CX;
      U := a.Data[P] or V;
      a.Data[P] := U;
      Dec (P);
      U := a.Data[I] shl CB;
      a.Data[P] := U;
    end;

    if C > 0 then begin
      BN_ZeroData (a.PLData, C);
    end;
    BN_Pack (a);
  end;
end;

{
  shift left by ABits bits
  b = a shl ABits
}
procedure BN_shl (var b: T_BNX; const a: T_BNX; ABits: IntX);
begin
  if ABits = 0 then begin
    BN_copy (b, a);
    Exit;
  end;

  if ABits > 0 then begin
    BN_pshl (b, a, ABits);
  end else begin
    BN_pshr (b, a, -ABits);
  end;
end;

{
  shift left by ABits bits
  a = a shl ABits
}
procedure BN_shl (var a: T_BNX; ABits: IntX);
begin
  if ABits = 0 then Exit;

  if ABits > 0 then begin
    BN_pshl (a, ABits);
  end else begin
    BN_pshr (a, -ABits);
  end;
end;

{
  shift right by ABits bits, ABits > 0 (positive)
  b = a shr ABits
}
procedure BN_pshr (var b: T_BNX; const a: T_BNX; ABits: IntX);
begin
  BN_copy (b, a);
  if ABits <= 0 then Exit;

  BN_pshr (b, ABits);
end;

{
  shift right by ABits bits, ABits > 0 (positive)
  a = a shr ABits
}
procedure BN_pshr (var a: T_BNX; ABits: IntX);
var
  I       : LoopInt;
  C       : IntX;
  CB, CX  : UIntX;
  V, U    : UIntX;
  VCount  : IntX;
  VRCount : IntX; // Result count
begin
  if ABits <= 0 then Exit;
  if a.Count <= 0 then Exit;

  if ABits < CXBits then begin
    BN_pshrd (a, ABits);
    Exit;
  end;

  C := BNX_div_mod_x_asm (ABits, CB, CX);

  VCount := BN_DataCount (a);
  VRCount := VCount - C;
  if VRCount <= 0 then begin
    BN_AssignZero (a);
    Exit;
  end;

  if CB = 0 then begin
    if C > 0 then begin
      BN_CopyData (a.PLData, @a.Data[C], VRCount);
    end;

    BN_Resize (a, VRCount);
  end else begin
    V := a.Data[C] shr CB;
    a.Data[0] := V;

    for I := 1 to VRCount - 1 do begin
      V := a.Data[I + C] shl CX;
      U := a.Data[I - 1] or V;
      a.Data[I - 1] := U;
      U := a.Data[I + C] shr CB;
      a.Data[I] := U;
    end;

    BN_Resize (a, VRCount);
    BN_Pack (a);
  end;
end;

{
  shift right by ABits bits
  b = a shr ABits
}
procedure BN_shr (var b: T_BNX; const a: T_BNX; ABits: IntX);
begin
  if ABits = 0 then begin
    BN_copy (b, a);
    Exit;
  end;

  if ABits > 0 then begin
    BN_pshr (b, a, ABits);
  end else begin
    BN_pshl (b, a, -ABits);
  end;
end;

{
  shift right by ABits bits
  a = a shr ABits
}
procedure BN_shr (var a: T_BNX; ABits: IntX);
begin
  if ABits <= 0 then Exit;

  if ABits > 0 then begin
    BN_pshr (a, ABits);
  end else begin
    BN_pshl (a, -ABits);
  end;
end;

{
  shift left by C*(32|64) bits, C > 0 (positive)
  b = a shl C*(32|64)
}
procedure BN_pcshl (var b: T_BNX; const a: T_BNX; C: IntX);
begin
  BN_copy (b, a);
  if C <= 0 then Exit;

  BN_pcshl (b, C);
end;

{
  shift left by C*(32|64) bits, C > 0 (positive)
  a = a shl C*(32|64)
}
procedure BN_pcshl (var a: T_BNX; C: IntX);
var
  I       : IntX;
  VCount  : IntX;
  VRCount : IntX; // Result count
begin
  if C <= 0 then Exit;

  VCount := BN_DataCount (a);
  VRCount := VCount + C;
  if VRCount <= VCount then begin
    BN_AssignZero (a);
    Exit;
  end;

  BN_Resize (a, VRCount);
  I := VCount - 1;
  BN_CopyDataDesc (@a.Data[I + C], @a.Data[I], VCount);
  BN_ZeroData (a.PLData, C);
end;

{
  shift left by C*(32|64) bits
  b = a shl C*(32|64)
}
procedure BN_cshl (var b: T_BNX; const a: T_BNX; C: IntX);
begin
  if C = 0 then begin
    BN_copy (b, a);
    Exit;
  end;

  if C > 0 then begin
    BN_pcshl (b, a, C);
  end else begin
    BN_pcshr (b, a, -C);
  end;
end;

{
  shift left by C*(32|64) bits
  a = a shl C*(32|64)
}
procedure BN_cshl (var a: T_BNX; C: IntX);
begin
  if C = 0 then Exit;

  if C > 0 then begin
    BN_pcshl (a, C);
  end else begin
    BN_pcshr (a, -C);
  end;
end;

{
  shift right by C*(32|64) bits, C > 0 (positive)
  b = a shr C*(32|64)
}
procedure BN_pcshr (var b: T_BNX; const a: T_BNX; C: IntX);
begin
  BN_copy (b, a);
  if C <= 0 then Exit;

  BN_pcshr (b, C);
end;

{
  shift right by C*(32|64) bits, C > 0 (positive)
  a = a shr C*(32|64)
}
procedure BN_pcshr (var a: T_BNX; C: IntX);
var
  VCount  : IntX;
  VRCount : IntX; // Result count
begin
  if C <= 0 then Exit;

  VCount := BN_DataCount (a);
  VRCount := VCount - C;
  if VRCount <= 0 then begin
    BN_AssignZero (a);
    Exit;
  end;

  BN_CopyData (a.PLData, @a.Data[C], VRCount);
  BN_Resize (a, VRCount);
end;

{
  shift right by C*(32|64) bits
  b = a shr C*(32|64)
}
procedure BN_cshr (var b: T_BNX; const a: T_BNX; C: IntX);
begin
  if C = 0 then begin
    BN_copy (b, a);
    Exit;
  end;

  if C > 0 then begin
    BN_pcshr (b, a, C);
  end else begin
    BN_pcshl (b, a, -C);
  end;
end;

{
  shift right by C*(32|64) bits
  a = a shr C*(32|64)
}
procedure BN_cshr (var a: T_BNX; C: IntX);
begin
  if C = 0 then Exit;

  if C > 0 then begin
    BN_pcshr (a, C);
  end else begin
    BN_pcshl (a, -C);
  end;
end;

type
  T_BNX_add_sub_asm_proc = procedure (a1, a2: PUIntX; XCount, DCount: IntX);
  T_BNX_add_sub_int_asm_proc = procedure (a1: PUIntX; a2: UIntX; Count: IntX);

{
  a1 = a1 + a2
  |a1| >= |a2|
  XCount = min (a1.Count, a2.Count) = a2.Count
  DCount = a1.Count - a2.Count
}
procedure BNX_add_asm (a1, a2: PUIntX; XCount, DCount: IntX);
{$IFDEF CPUX64}
// rcx | @a1    | r10
// rdx | @a2    |
// r8  | XCount | rcx
// r9  | DCount | rcx
asm
       .NOFRAME

        mov     r10, rcx
        mov     rcx, r8

        mov     rax, [rdx]
        add     [r10], rax

        dec     rcx
        jrcxz   @b

        (*
        lea     r11, [@a]
        lea     r12, [@b]
        // *)

    @a:
        lea     r10, [r10 + CXBytes]
        lea     rdx, [rdx + CXBytes]
        mov     rax, [rdx]
        adc     [r10], rax
        dec     rcx
        jrcxz   @b
        jmp     @a
        (*
        cmovz   r11, r12
        jmp     r11
        // *)

    @b:
        mov     rcx, r9
        jrcxz   @D
        jnc     @R

    @c:
        lea     r10, [r10 + CXBytes]
        adc     qword ptr [r10], 0
        dec     rcx
        jrcxz   @D
        jnc     @R
        jmp     @c

    @D:
        lea     r10, [r10 + CXBytes]
        adc     qword ptr [r10], 0

    @R:
end;
{$ELSE}
// eax | @a1    | ebx
// edx | @a2    |
// ecx | XCount |
//     | DCount | ecx
asm
        push    ebx

        mov     ebx, eax

        mov     eax, [edx]
        add     [ebx], eax

        dec     ecx
        jecxz   @b

    @a:
        lea     ebx, [ebx + CXBytes]
        lea     edx, [edx + CXBytes]
        mov     eax, [edx]
        adc     [ebx], eax
        dec     ecx
        jecxz   @b
        jmp     @a

    @b:
        mov     ecx, [DCount]
        jecxz   @D
        jnc     @R

    @c:
        lea     ebx, [ebx + CXBytes]
        adc     dword ptr [ebx], 0
        dec     ecx
        jecxz   @D
        jnc     @R
        jmp     @c

    @D:
        lea     ebx, [ebx + CXBytes]
        adc     dword ptr [ebx], 0

    @R:
        pop     ebx
end;
{$ENDIF}

{
  a1 = a1 + a2
  Count = a1.Count
}
procedure BNX_add_int_asm (a1: PUIntX; a2: UIntX; Count: IntX);
{$IFDEF CPUX64}
// rcx | @a1   | r10
// rdx | a2    |
// r8  | Count | rcx
asm
       .NOFRAME

        mov     r10, rcx
        mov     rcx, r8

        add     [r10], rdx

        dec     rcx
        jrcxz   @D
        jnc     @R

    @c:
        lea     r10, [r10 + CXBytes]
        adc     qword ptr [r10], 0
        dec     rcx
        jrcxz   @D
        jnc     @R
        jmp     @c

    @D:
        lea     r10, [r10 + CXBytes]
        adc     qword ptr [r10], 0

    @R:
end;
{$ELSE}
// eax | @a1   | ebx
// edx | a2    |
// ecx | Count |
asm
        push    ebx

        mov     ebx, eax

        add     [ebx], edx

        dec     ecx
        jecxz   @D
        jnc     @R

    @c:
        lea     ebx, [ebx + CXBytes]
        adc     dword ptr [ebx], 0
        dec     ecx
        jecxz   @D
        jnc     @R
        jmp     @c

    @D:
        lea     ebx, [ebx + CXBytes]
        adc     dword ptr [ebx], 0

    @R:
        pop     ebx
end;
{$ENDIF}

{
  a1 = a1 - a2
  |a1| > |a2|
  XCount = min (a1.Count, a2.Count) = a2.Count
  DCount = a1.Count - a2.Count
}
procedure BNX_sub_asm (a1, a2: PUIntX; XCount, DCount: IntX);
{$IFDEF CPUX64}
// rcx | @a1    | r10
// rdx | @a2    |
// r8  | XCount | rcx
// r9  | DCount | rcx
asm
       .NOFRAME

        mov     r10, rcx
        mov     rcx, r8

        mov     rax, [rdx]
        sub     [r10], rax

        dec     ecx
        jrcxz   @b

    @a:
        lea     r10, [r10 + CXBytes]
        lea     rdx, [rdx + CXBytes]
        mov     rax, [rdx]
        sbb     [r10], rax
        dec     rcx
        jrcxz   @b
        jmp     @a

    @b:
        mov     rcx, r9
        jrcxz   @R
        jnc     @R

    @c:
        lea     r10, [r10 + CXBytes]
        sbb     qword ptr [r10], 0
        dec     rcx
        jrcxz   @R
        jc      @c

    @R:
end;
{$ELSE}
// eax | @a1    | ebx
// edx | @a2    |
// ecx | XCount |
//     | DCount | ecx
asm
        push    ebx

        mov     ebx, eax

        mov     eax, [edx]
        sub     [ebx], eax

        dec     ecx
        jecxz   @b

    @a:
        lea     ebx, [ebx + CXBytes]
        lea     edx, [edx + CXBytes]
        mov     eax, [edx]
        sbb     [ebx], eax
        dec     ecx
        jecxz   @b
        jmp     @a

    @b:
        mov     ecx, [DCount]
        jecxz   @R
        jnc     @R

    @c:
        lea     ebx, [ebx + CXBytes]
        sbb     dword ptr [ebx], 0
        dec     ecx
        jecxz   @R
        jc      @c

    @R:
        pop     ebx
end;
{$ENDIF}

{
  a1 = a1 - a2
  |a1| > |a2|
  Count = a1.Count
}
procedure BNX_sub_int_asm (a1: PUIntX; a2: UIntX; Count: IntX);
{$IFDEF CPUX64}
// rcx | @a1   | r10
// rdx | a2    |
// r8  | Count | rcx
asm
       .NOFRAME

        mov     r10, rcx
        mov     rcx, r8

        sub     [r10], rdx

        dec     rcx
        jrcxz   @R
        jnc     @R

    @c:
        lea     r10, [r10 + CXBytes]
        sbb     qword ptr [r10], 0
        dec     rcx
        jrcxz   @R
        jc      @c

    @R:
end;
{$ELSE}
// eax | @a1
// edx | a2
// ecx | Count
asm
        sub     [eax], edx

        dec     ecx
        jecxz   @R
        jnc     @R

    @c:
        lea     eax, [eax + CXBytes]
        sbb     dword ptr [eax], 0
        dec     ecx
        jecxz   @R
        jc      @c

    @R:
end;
{$ENDIF}

{
  a3 = a1 + a2
}
procedure BN_add (var a3: T_BNX; const a1, a2: T_BNX);
var
  P3, P2   : PUIntX;
  VSign1   : ShortInt;
  VSign2   : ShortInt;
  VSign3   : ShortInt;
  VCount1  : IntX;
  VCount2  : IntX;
  VCount3  : IntX;
  VCount   : IntX;
  VXCount  : IntX;
  VDCount  : IntX;
  VCompare : Integer;
  VProc    : T_BNX_add_sub_asm_proc;
begin
  VSign1 := a1.Sign;
  VSign2 := a2.Sign;
  VSign3 := VSign1;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // a1 = 0
    // a3 = a2
    BN_copy (a3, a2);
    Exit;
  end;

  if (VSign2 = 0) or (VCount2 <= 0) then begin
    // a2 = 0
    // a3 = a1
    BN_copy (a3, a1);
    Exit;
  end;

  VCount := VCount1;
  VXCount := VCount1;
  VDCount := 0;

  if VCount1 < VCount2 then begin
    VCount := VCount2;
    // VXCount := VCount1;
    VDCount := VCount2 - VCount1;
  end else
  if VCount2 < VCount1 then begin
    // VCount := VCount1;
    VXCount := VCount2;
    VDCount := VCount1 - VCount2;
  end;

  VCompare := BN_DataCompare (a1, a2);

  if VSign1 = VSign2 then begin
    if VCompare = 0 then begin
      // |a1| = |a2|
      // |a3| = |a1| + |a2| = 2 * |a1|
      // ----
      // (+3) + (+3) = +6
      // (-3) + (-3) = -6
      // a3.sign = a1.sign

      // VSign3 := VSign1;

      BN_copy (a3, a1); // a3 = a1
      BN_shl_1 (a3);    // a3 = 2 * a1
      Exit;
    end;

    // |a3| = |a1| + |a2|
    // ----
    // (+3) + (+2) = +5
    // (-3) + (-2) = -5
    // (+2) + (+3) = +5
    // (-2) + (-3) = -5
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount + 1;

    if VCompare > 0 then begin
      // |a1| > |a2|
      BN_ccopy (a3, a1, VCount3); // a3 = a1
      P2 := a2.PLData;            // a2 = a2
    end else begin
      // |a1| < |a2|
      BN_ccopy (a3, a2, VCount3); // a3 = a2
      P2 := a1.PLData;            // a2 = a1
    end;

    VProc := BNX_add_asm;
  end else begin
    if VCompare = 0 then begin
      // |a1| = |a2|
      // |a3| = |a1| - |a2| = 0
      BN_AssignZero (a3);
      Exit;
    end;

    VCount3 := VCount;

    if VCompare > 0 then begin
      // |a1| > |a2|
      // |a3| = |a1| - |a2|
      // ----
      // (+3) + (-2) = +1
      // (-3) + (+2) = -1
      // a3.sign = a1.sign

      // VSign3 := VSign1;

      BN_ccopy (a3, a1, VCount3); // a3 = a1
      P2 := a2.PLData;            // a2 = a2
    end else begin
      // |a1| < |a2|
      // |a3| = |a2| - |a1|
      // ----
      // (+2) + (-3) = -1
      // (-2) + (+3) = +1
      // a3.sign = a2.sign

      VSign3 := VSign2;

      BN_ccopy (a3, a2, VCount3); // a3 = a2
      P2 := a1.PLData;            // a2 = a1
    end;

    VProc := BNX_sub_asm;
  end;

  a3.Sign := VSign3;
  P3 := a3.PLData;

  // (*
  if @VProc = @BNX_add_asm then begin
    while (P3^ = 0) or (P2^ = 0) do begin
      if VXCount = 1 then Break;
      if P3^ = 0 then P3^ := P2^;
      P3 := PUIntX(UIntX(P3) + CXBytes);
      P2 := PUIntX(UIntX(P2) + CXBytes);
      Dec (VXCount);
    end;
  end else begin
    while P2^ = 0 do begin
      if VXCount = 1 then Break;
      P3 := PUIntX(UIntX(P3) + CXBytes);
      P2 := PUIntX(UIntX(P2) + CXBytes);
      Dec (VXCount);
    end;
  end;
  // *)

  VProc (P3, P2, VXCount, VDCount);
  BN_Pack (a3);
end;

{
  a1 = a1 + a2
}
procedure BN_add_(var a1: T_BNX; const a2: T_BNX);
var
  V : T_BNX;
begin
  BN_copy (V, a1);
  BN_add (a1, V, a2);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  a1 = a1 + a2
  imv - external intermediate variable
}
procedure BN_add_(var a1: T_BNX; const a2: T_BNX; var imv: T_BNX);
begin
  BN_copy (imv, a1);
  BN_add (a1, imv, a2);
end;

{
  a3 = a1 + a2
}
procedure BN_add_int (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX);
var
  VSign1   : ShortInt;
  VSign3   : ShortInt;
  VCount1  : IntX;
  VCount3  : IntX;
  VCompare : Integer;
  VProc    : T_BNX_add_sub_int_asm_proc;
begin
  VSign1 := a1.Sign;
  VSign3 := VSign1;

  VCount1 := BN_DataCount (a1);

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // a1 = 0
    // a3 = a2
    BN_AssignInt (a3, ASign2, a2);
    Exit;
  end;

  if (ASign2 = 0) or (a2 = 0) then begin
    // a2 = 0
    // a3 = a1
    BN_copy (a3, a1);
    Exit;
  end;

  if VSign1 = ASign2 then begin
    // |a3| = |a1| + |a2|
    // ----
    // (+3) + (+2) = +5
    // (-3) + (-2) = -5
    // (+2) + (+3) = +5
    // (-2) + (-3) = -5
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount1 + 1;

    VProc := BNX_add_int_asm;
  end else begin
    VCompare := BN_DataCompare (a1, a2);

    if VCompare = 0 then begin
      // |a1| = |a2|
      // |a3| = |a1| - |a2| = 0
      BN_AssignZero (a3);
      Exit;
    end;

    if VCompare < 0 then begin
      // |a1| < |a2|
      // |a3| = |a2| - |a1|
      // VCount1 = 1
      // ----
      // (+2) + (-3) = -1
      // (-2) + (+3) = +1
      // a3.sign = a2.sign

      // VSign3 := ASign2;

      BN_AssignInt (a3, ASign2, a2 - a1.Data[0]);
      Exit;
    end;

    // |a1| > |a2|
    // |a3| = |a1| - |a2|
    // VCount1 > 1
    // ----
    // (+3) + (-2) = +1
    // (-3) + (+2) = -1
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount1;

    VProc := BNX_sub_int_asm;
  end;

  BN_ccopy (a3, a1, VCount3);
  a3.Sign := VSign3;

  VProc (a3.PLData, a2, VCount1);
  BN_Pack (a3);
end;

{
  a1 = a1 + a2
}
procedure BN_add_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX);
var
  V : T_BNX;
begin
  BN_copy (V, a1);
  BN_add_int (a1, V, ASign2, a2);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  a1 = a1 + a2
  imv - external intermediate variable
}
procedure BN_add_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX;
  var imv: T_BNX);
begin
  BN_copy (imv, a1);
  BN_add_int (a1, imv, ASign2, a2);
end;

{
  a3 = a1 + a2
  a2 - signed integer
}
procedure BN_add_int (var a3: T_BNX; const a1: T_BNX; a2: IntX);
var
  VSign2 : ShortInt;
begin
  VSign2 := 0;
  if a2 > 0 then VSign2 := 1 else
  if a2 < 0 then begin
    VSign2 := -1;
    a2 := -a2;
  end;

  BN_add_int (a3, a1, VSign2, a2);
end;

{
  a1 = a1 + a2
  a2 - signed integer
}
procedure BN_add_int_(var a1: T_BNX; a2: IntX);
var
  V      : T_BNX;
  VSign2 : ShortInt;
begin
  VSign2 := 0;
  if a2 > 0 then VSign2 := 1 else
  if a2 < 0 then begin
    VSign2 := -1;
    a2 := -a2;
  end;

  BN_copy (V, a1);
  BN_add_int (a1, V, VSign2, a2);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  a3 = a1 + 1
}
procedure BN_inc (var a3: T_BNX; const a1: T_BNX);
var
  VSign1   : ShortInt;
  VSign3   : ShortInt;
  VCount1  : IntX;
  VCount3  : IntX;
  VCompare : Integer;
  VProc    : T_BNX_add_sub_int_asm_proc;
begin
  VSign1 := a1.Sign;
  VSign3 := VSign1;

  VCount1 := BN_DataCount (a1);

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // a1 = 0
    // a3 = 1
    BN_AssignPOne (a3);
    Exit;
  end;

  if VSign1 = 1 then begin
    // |a3| = |a1| + 1
    // ----
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount1 + 1;

    VProc := BNX_add_int_asm;
  end else begin
    // VSign1 = -1
    VCompare := BN_DataCompare (a1, 1);

    if VCompare = 0 then begin
      // |a1| = 1
      // |a3| = |a1| - 1 = 0
      BN_AssignZero (a3);
      Exit;
    end;

    // |a1| > 1
    // |a3| = |a1| - 1
    // VCount1 > 1
    // ----
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount1;

    VProc := BNX_sub_int_asm;
  end;

  BN_ccopy (a3, a1, VCount3);
  a3.Sign := VSign3;

  VProc (a3.PLData, 1, VCount1);
  BN_Pack (a3);
end;

// [+] 1.2.1
{
  a = a + 1
}
procedure BN_inc_(var a: T_BNX);
var
  V : T_BNX;
begin
  BN_copy (V, a);
  BN_inc (a, V);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  a = a + 1
  imv - external intermediate variable
}
procedure BN_inc_(var a: T_BNX; var imv: T_BNX);
begin
  BN_copy (imv, a);
  BN_inc (a, imv);
end;

{
  a3 = a1 - a2
}
procedure BN_sub (var a3: T_BNX; const a1, a2: T_BNX);
var
  P3, P2   : PUIntX;
  VSign1   : ShortInt;
  VSign2   : ShortInt;
  VSign3   : ShortInt;
  VCount1  : IntX;
  VCount2  : IntX;
  VCount3  : IntX;
  VCount   : IntX;
  VXCount  : IntX;
  VDCount  : IntX;
  VCompare : Integer;
  VProc    : T_BNX_add_sub_asm_proc;
begin
  VSign1 := a1.Sign;
  VSign2 := a2.Sign;
  VSign3 := VSign1;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // a1 = 0
    // a3 = -a2
    BN_ncopy (a3, a2);
    Exit;
  end;

  if (VSign2 = 0) or (VCount2 <= 0) then begin
    // a2 = 0
    // a3 = a1
    BN_copy (a3, a1);
    Exit;
  end;

  VCount := VCount1;
  VXCount := VCount1;
  VDCount := 0;

  if VCount1 < VCount2 then begin
    VCount := VCount2;
    // VXCount := VCount1;
    VDCount := VCount2 - VCount1;
  end else
  if VCount2 < VCount1 then begin
    // VCount := VCount1;
    VXCount := VCount2;
    VDCount := VCount1 - VCount2;
  end;

  VCompare := BN_DataCompare (a1, a2);

  if VSign1 = VSign2 then begin
    if VCompare = 0 then begin
      // |a1| = |a2|
      // |a3| = |a1| - |a2| = 0
      BN_AssignZero (a3);
      Exit;
    end;

    VCount3 := VCount;

    if VCompare > 0 then begin
      // |a1| > |a2|
      // |a3| = |a1| - |a2|
      // ----
      // (+3) - (+2) = +1
      // (-3) - (-2) = -1
      // a3.sign = a1.sign

      // VSign3 := VSign1;

      BN_ccopy (a3, a1, VCount3); // a3 = a1
      P2 := a2.PLData;            // a2 = a2
    end else begin
      // |a1| < |a2|
      // |a3| = |a2| - |a1|
      // ----
      // (+2) - (+3) = -1
      // (-2) - (-3) = +1
      // a3.sign = -a1.sign

      VSign3 := -VSign1;

      BN_ccopy (a3, a2, VCount3); // a3 = a2
      P2 := a1.PLData;            // a2 = a1
    end;

    VProc := BNX_sub_asm;
  end else begin
    if VCompare = 0 then begin
      // |a1| = |a2|
      // |a3| = |a1| + |a2| = 2 * |a1|
      // ----
      // (+3) - (-3) = +6
      // (-3) - (+3) = -6
      // a3.sign = a1.sign

      // VSign3 := VSign1;

      BN_copy (a3, a1); // a3 = a1
      BN_shl_1 (a3);    // a3 = 2 * a1
      Exit;
    end;

    // |a3| = |a1| + |a2|
    // ----
    // (+3) - (-2) = +5
    // (-3) - (+2) = -1
    // (+2) - (-3) = +5
    // (-2) - (+3) = -1
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount + 1;

    if VCompare > 0 then begin
      // |a1| > |a2|
      BN_ccopy (a3, a1, VCount3); // a3 = a1
      P2 := a2.PLData;            // a2 = a2
    end else begin
      // |a1| < |a2|
      BN_ccopy (a3, a2, VCount3); // a3 = a2
      P2 := a1.PLData;            // a2 = a1
    end;

    VProc := BNX_add_asm;
  end;

  a3.Sign := VSign3;
  P3 := a3.PLData;

  // (*
  if @VProc = @BNX_add_asm then begin
    while (P3^ = 0) or (P2^ = 0) do begin
      if VXCount = 1 then Break;
      if P3^ = 0 then P3^ := P2^;
      P3 := PUIntX(UIntX(P3) + CXBytes);
      P2 := PUIntX(UIntX(P2) + CXBytes);
      Dec (VXCount);
    end;
  end else begin
    while P2^ = 0 do begin
      if VXCount = 1 then Break;
      P3 := PUIntX(UIntX(P3) + CXBytes);
      P2 := PUIntX(UIntX(P2) + CXBytes);
      Dec (VXCount);
    end;
  end;
  // *)

  VProc (P3, P2, VXCount, VDCount);
  BN_Pack (a3);
end;

{
  a1 = a1 - a2
}
procedure BN_sub_(var a1: T_BNX; const a2: T_BNX);
var
  V : T_BNX;
begin
  BN_copy (V, a1);
  BN_sub (a1, V, a2);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  a1 = a1 - a2
  imv - external intermediate variable
}
procedure BN_sub_(var a1: T_BNX; const a2: T_BNX; var imv: T_BNX);
begin
  BN_copy (imv, a1);
  BN_sub (a1, imv, a2);
end;

{
  a3 = a1 - a2
}
procedure BN_sub_int (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX);
var
  VSign1   : ShortInt;
  VSign3   : ShortInt;
  VCount1  : IntX;
  VCount3  : IntX;
  VCompare : Integer;
  VProc    : T_BNX_add_sub_int_asm_proc;
begin
  VSign1 := a1.Sign;
  VSign3 := VSign1;

  VCount1 := BN_DataCount (a1);

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // a1 = 0
    // a3 = -a2
    BN_AssignInt (a3, -ASign2, a2);
    Exit;
  end;

  if (ASign2 = 0) or (a2 = 0) then begin
    // a2 = 0
    // a3 = a1
    BN_copy (a3, a1);
    Exit;
  end;

  if VSign1 = ASign2 then begin
    VCompare := BN_DataCompare (a1, a2);

    if VCompare = 0 then begin
      // |a1| = |a2|
      // |a3| = |a1| - |a2| = 0
      BN_AssignZero (a3);
      Exit;
    end;

    if VCompare < 0 then begin
      // |a1| < |a2|
      // |a3| = |a2| - |a1|
      // VCount1 = 1
      // ----
      // (-2) - (-3) = +1
      // (+2) - (+3) = -1
      // a3.sign = -a1.sign

      // VSign3 := -VSign1;

      BN_AssignInt (a3, -VSign1, a2 - a1.Data[0]);
      Exit;
    end;

    // |a1| > |a2|
    // |a3| = |a1| - |a2|
    // VCount1 > 1
    // ----
    // (+3) - (+2) = +1
    // (-3) - (-2) = -1
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount1;

    VProc := BNX_sub_int_asm;
  end else begin
    // |a3| = |a1| + |a2|
    // ----
    // (+3) - (-2) = +5
    // (-3) - (+2) = -1
    // (+2) - (-3) = +5
    // (-2) - (+3) = -1
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount1 + 1;

    VProc := BNX_add_int_asm;
  end;

  BN_ccopy (a3, a1, VCount3);
  a3.Sign := VSign3;

  VProc (a3.PLData, a2, VCount1);
  BN_Pack (a3);
end;

{
  a1 = a1 - a2
}
procedure BN_sub_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX);
var
  V : T_BNX;
begin
  BN_copy (V, a1);
  BN_sub_int (a1, V, ASign2, a2);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  a1 = a1 - a2
  imv - external intermediate variable
}
procedure BN_sub_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX;
  var imv: T_BNX);
begin
  BN_copy (imv, a1);
  BN_sub_int (a1, imv, ASign2, a2);
end;

{
  a3 = a1 - a2
  a2 - signed integer
}
procedure BN_sub_int (var a3: T_BNX; const a1: T_BNX; a2: IntX);
var
  VSign2 : ShortInt;
begin
  VSign2 := 0;
  if a2 > 0 then VSign2 := 1 else
  if a2 < 0 then begin
    VSign2 := -1;
    a2 := -a2;
  end;

  BN_sub_int (a3, a1, VSign2, a2);
end;

{
  a1 = a1 - a2
  a2 - signed integer
}
procedure BN_sub_int_(var a1: T_BNX; a2: IntX);
var
  V      : T_BNX;
  VSign2 : ShortInt;
begin
  VSign2 := 0;
  if a2 > 0 then VSign2 := 1 else
  if a2 < 0 then begin
    VSign2 := -1;
    a2 := -a2;
  end;

  BN_copy (V, a1);
  BN_sub_int (a1, V, VSign2, a2);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  a3 = a1 - 1
}
procedure BN_dec (var a3: T_BNX; const a1: T_BNX);
var
  VSign1   : ShortInt;
  VSign3   : ShortInt;
  VCount1  : IntX;
  VCount3  : IntX;
  VCompare : Integer;
  VProc    : T_BNX_add_sub_int_asm_proc;
begin
  VSign1 := a1.Sign;
  VSign3 := VSign1;

  VCount1 := BN_DataCount (a1);

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // a1 = 0
    // a3 = -1
    BN_AssignNOne (a3);
    Exit;
  end;

  if VSign1 = 1 then begin
    VCompare := BN_DataCompare (a1, 1);

    if VCompare = 0 then begin
      // |a1| = 1
      // |a3| = |a1| - 1 = 0
      BN_AssignZero (a3);
      Exit;
    end;

    // |a1| > 1
    // |a3| = |a1| - 1
    // VCount1 > 1
    // ----
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount1;

    VProc := BNX_sub_int_asm;
  end else begin
    // |a3| = |a1| + 1
    // ----
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount1 + 1;

    VProc := BNX_add_int_asm;
  end;

  BN_ccopy (a3, a1, VCount3);
  a3.Sign := VSign3;

  VProc (a3.PLData, 1, VCount1);
  BN_Pack (a3);
end;

// [+] 1.2.1
{
  a = a - 1
}
procedure BN_dec_(var a: T_BNX);
var
  V : T_BNX;
begin
  BN_copy (V, a);
  BN_dec (a, V);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  a = a - 1
  imv - external intermediate variable
}
procedure BN_dec_(var a: T_BNX; var imv: T_BNX);
begin
  BN_copy (imv, a);
  BN_dec (a, imv);
end;

{
     ()     ASign1  ASign2
  Return the sign of the product (division) of two numbers with signs ASign1 and ASign2
}
function BN_mul_div_sign (ASign1, ASign2: ShortInt): ShortInt;
begin
  Result := ASign1 * ASign2;
end;

{
  a3 = a1 * a2
}
procedure BNX_mul_int_asm (a1, a2: UIntX; var a3: UIntX; var Carry: UIntX);
{$IFDEF CPUX64}
// rcx | a1
// rdx | a2
// r8  | @a3
// r9  | @Carry
asm
       .NOFRAME

        mov     rax, rcx
        mul     rdx

        add     rax, [r9]
        adc     rdx, 0
        add     rax, [r8]
        adc     rdx, 0
        mov     [r8], rax
        mov     [r9], rdx
end;
{$ELSE}
// eax | a1     |
// edx | a2     |
// ecx | @a3    |
//     | @Carry | ebx
asm
        push    ebx

        mov     ebx, [Carry]
        mul     edx

        add     eax, [ebx]
        adc     edx, 0
        add     eax, [ecx]
        adc     edx, 0
        mov     [ecx], eax
        mov     [ebx], edx

        pop     ebx
end;
{$ENDIF}

{
  a3 = a1 * a2
  Count1 = a1.Count
  Count2 = a2.Count
}
function BNX_mul_base_asm (a1, a2, a3: PUIntX; Count1, Count2: IntX): UIntX;
{$IFDEF CPUX64}
// rcx | @a1    | r10
// rdx | @a2    | r11
// r8  | @a3    | r12
// r9  | Count1 | r13
//     | Count2 | r14
//     | Carry  | r15
asm
        push    r12
        push    r13
        push    r14
        push    r15

        mov     r10, rcx
        mov     r11, rdx
        mov     r12, r8
        mov     r13, r9
        mov     r14, [Count2]

    @loop_I:
        xor     r15, r15 // Carry = 0
    @loop_J:
        mov     rax, [rcx]
        mul     qword ptr [r11] // rax * a2

        add     rax, r15
        adc     rdx, 0
        add     rax, [r8]
        adc     rdx, 0
        mov     [r8], rax
        mov     r15, rdx
    @next_J:
        lea     rcx, [rcx + CXBytes] // Inc (@a1)
        lea     r8, [r8 + CXBytes]   // Inc (@a3)
        dec     r9
        jnz     @loop_J
    @out_J:

        add     [r8], r15
        mov     rax, [r8]

    @next_I:
        mov     rcx, r10             // Restore @a1
        lea     r11, [r11 + CXBytes] // Inc (@a2)
        lea     r12, [r12 + CXBytes] // Inc (@a3)
        mov     r8, r12
        mov     r9, r13              // Restore Loop_J counter
        dec     r14
        jnz     @loop_I
    @out_I:

        pop     r15
        pop     r14
        pop     r13
        pop     r12
end;
{$ELSE}
// eax | @a1    | ebp [esp + 04]
// edx | @a2    | esi
// ecx | @a3    | edi
//     | Count1 | ebx [esp + 08]
//     | Count2 |     [esp + 12]
//     | Carry  |     [esp]
asm
        push    ebx
        push    esi
        push    edi
        push    ebp

        sub     esp, 16 // Alloc Local variables

        mov     esi, edx
        mov     edi, ecx

        mov     ebx, [Count2]
        mov     [esp + 12], ebx // I
        mov     ebx, [Count1]
        mov     [esp + 08], ebx // J

        mov     ebp, eax
        mov     [esp + 04], eax

    @loop_I:
        mov     dword ptr [esp], 0 // Carry = 0
    @loop_J:
        mov     eax, [ebp]
        mul     dword ptr [esi] // eax * a2

        add     eax, [esp]
        adc     edx, 0
        add     eax, [ecx]
        adc     edx, 0
        mov     [ecx], eax
        mov     [esp], edx
    @next_J:
        lea     ebp, [ebp + CXBytes] // Inc (@a1)
        lea     ecx, [ecx + CXBytes] // Inc (@a3)
        dec     ebx
        jnz     @loop_J
    @out_J:

        mov     eax, [ecx]
        add     eax, [esp]
        mov     [ecx], eax

    @next_I:
        mov     ebp, [esp + 04]      // Restore @a1
        lea     esi, [esi + CXBytes] // Inc (@a2)
        lea     edi, [edi + CXBytes] // Inc (@a3)
        mov     ecx, edi
        mov     ebx, [esp + 08]      // Restore Loop_J counter
        dec     dword ptr [esp + 12]
        jnz     @loop_I
    @out_I:

        add     esp, 16 // Free Local variables

        pop     ebp
        pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
  a3 = a1 * a2
  Count1 = a1.Count
}
function BNX_mul_int_base_asm (a1: PUIntX; a2: UIntX; a3: PUIntX; Count1: IntX): UIntX;
{$IFDEF CPUX64}
// rcx | @a1    |
// rdx | a2     | r10
// r8  | @a3    |
// r9  | Count1 |
//     | Carry  | r11
asm
       .NOFRAME

        mov     r10, rdx
        xor     r11, r11

    @loop_J:
        mov     rax, [rcx]
        mul     r10

        add     rax, r11
        adc     rdx, 0
        add     rax, [r8]
        adc     rdx, 0
        mov     [r8], rax
        mov     r11, rdx
    @next_J:
        lea     rcx, [rcx + CXBytes]
        lea     r8, [r8 + CXBytes]
        dec     r9
        jnz     @loop_J
    @out_J:

        add     [r8], r11
        mov     rax, [r8]
end;
{$ELSE}
// eax | @a1    | ebp
// edx | a2     | edi
// ecx | @a3    |
//     | Count1 | ebx
//     | Carry  | esi
asm
        push    ebx
        push    esi
        push    edi
        push    ebp

        mov     ebx, [Count1]

        mov     ebp, eax
        mov     edi, edx
        xor     esi, esi

    @loop_J:
        mov     eax, [ebp]
        mul     edi

        add     eax, esi
        adc     edx, 0
        add     eax, [ecx]
        adc     edx, 0
        mov     [ecx], eax
        mov     esi, edx
    @next_J:
        lea     ebp, [ebp + CXBytes]
        lea     ecx, [ecx + CXBytes]
        dec     ebx
        jnz     @loop_J
    @out_J:

        add     [ecx], esi
        mov     eax, [ecx]

        pop     ebp
        pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
  a3 = a1 * a2
}
procedure BN_mul_base (var a3: T_BNX; const a1, a2: T_BNX);
var
  VUInt   : UIntX;
  VSign3  : ShortInt;
  VCount1 : IntX;
  VCount2 : IntX;
  VCount3 : IntX;
begin
  VSign3 := BN_mul_div_sign (a1.Sign, a2.Sign);

  if (VSign3 = 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VCount1 <= 0) or (VCount2 <= 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  if BN_IsInt (a1, VUInt) then begin
    BN_mul_int_nz (a3, a2, VUInt, VCount2, VSign3);
    Exit;
  end;

  if BN_IsInt (a2, VUInt) then begin
    BN_mul_int_nz (a3, a1, VUInt, VCount1, VSign3);
    Exit;
  end;

  VCount3 := VCount1 + VCount2;
  BN_Init (a3, VSign3, VCount3);

  if BNX_mul_base_asm (a1.PLData, a2.PLData, a3.PLData, VCount1, VCount2) = 0 then
  BN_Pack (a3);
end;

{
  a1 = a1 * a2
}
procedure BN_mul_base_(var a1: T_BNX; const a2: T_BNX);
var
  V : T_BNX;
begin
  BN_mul_base (V, a1, a2);
  BN_copy (a1, V);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  a1 = a1 * a2
  imv - external intermediate variable
}
procedure BN_mul_base_(var a1: T_BNX; const a2: T_BNX; var imv: T_BNX);
begin
  BN_mul_base (imv, a1, a2);
  BN_copy (a1, imv);
end;

{
  a3 = a1 * a2

   
  Karatsuba algorithm
}
procedure BN_mul_karatsuba_(var a3: T_BNX; const a1, a2: T_BNX; Am: IntX);
var
  V_a0   : T_BNX;
  V_a1   : T_BNX;
  V_aa   : T_BNX;
  V_b0   : T_BNX;
  V_b1   : T_BNX;
  V_bb   : T_BNX;
  V_ab00 : T_BNX;
  V_ab11 : T_BNX;
  V_aabb : T_BNX;
begin
  // (a0 + a1 * 2^[m]) * (b0 + b1 * 2^[m]) =
  // = (a0 * b0) + (a0 * b1 + a1 * b0) * 2^[m] + (a1 * b1) * 2^[2m]
  // | ab00 = (a0 * b0)
  // | ab11 = (a1 * b1)
  // | aabb = (a0 + a1) * (b0 + b1) = (a0 * b0) + (a1 * b1) + (a0 * b1 + a1 * b0)
  // | (a0 * b1 + a1 * b0) = (a0 + a1) * (b0 + b1) - ((a0 * b0) + (a1 * b1)) =
  // | = aabb - (ab00 + ab11)
  // = ab00 + (aabb - (ab00 + ab11)) *  2^[m] + ab11 * 2^[2m]

  BN_split (V_a0, V_a1, a1, Am);
  BN_split (V_b0, V_b1, a2, Am);

  BN_add (V_aa, V_a0, V_a1);     // V_aa = V_a0 + V_a1
  BN_add (V_bb, V_b0, V_b1);     // V_bb = V_b0 + V_b1

  BN_mul (V_ab00, V_a0, V_b0);   // V_ab00 = V_a0 * V_b0
  BN_mul (V_ab11, V_a1, V_b1);   // V_ab11 = V_a1 * V_b1
  BN_mul (V_aabb, V_aa, V_bb);   // V_aabb = (V_a0 + V_a1) * (V_b0 + V_b1)

  BN_add (V_aa, V_ab00, V_ab11); // V_aa = V_ab00 + V_ab11
  BN_sub (V_bb, V_aabb, V_aa);   // V_bb = V_aabb - (V_ab00 + V_ab11)
  BN_pcshl (V_bb, Am);           // V_bb = (V_aabb - (V_ab00 + V_ab11)) * 2^[m]
  BN_pcshl (V_ab11, Am shl 1);   // V_ab11 = V_ab11 * 2^[2m]
  BN_add (V_aa, V_ab00, V_bb);   // V_aa = V_ab00 + (V_aabb - (V_ab00 + V_ab11)) * 2^[m]
  BN_add (a3, V_aa, V_ab11);     // a3 = V_ab00 + (V_aabb - (V_ab00 + V_ab11)) * 2^[m] + V_ab11 * 2^[2m]

  BN_LFree (V_a0);
  BN_LFree (V_a1);
  BN_LFree (V_aa);
  BN_LFree (V_b0);
  BN_LFree (V_b1);
  BN_LFree (V_bb);
  BN_LFree (V_ab00);
  BN_LFree (V_ab11);
  BN_LFree (V_aabb);
end;

{
  a3 = a1 * a2
}
procedure BN_mul (var a3: T_BNX; const a1, a2: T_BNX);
var
  VUInt   : UIntX;
  VSign3  : ShortInt;
  VCount1 : IntX;
  VCount2 : IntX;
  VCount3 : IntX;
  VmCount : IntX;
  Vm      : IntX;
begin
  VSign3 := BN_mul_div_sign (a1.Sign, a2.Sign);

  if (VSign3 = 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VCount1 <= 0) or (VCount2 <= 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  {$IFDEF KARATSUBA_V2}
  // VmCount = max (VCount1, VCount2)
  VmCount := VCount1;
  if VCount2 > VmCount then VmCount := VCount2;

  if VmCount > Karatsuba_mul_TCount then begin
    Vm := VmCount shr 1; // VmCount div 2
    BN_mul_karatsuba_(a3, a1, a2, Vm);
    a3.Sign := VSign3;

    Exit;
  end;
  {$ELSE}
  if (VCount1 > Karatsuba_mul_TCount) and (VCount2 > Karatsuba_mul_TCount) then begin
    // VmCount = max (VCount1, VCount2)
    VmCount := VCount1;
    if VCount2 > VmCount then VmCount := VCount2;

    Vm := VmCount shr 1; // VmCount div 2
    BN_mul_karatsuba_(a3, a1, a2, Vm);
    a3.Sign := VSign3;

    Exit;
  end;
  {$ENDIF}

  if BN_IsInt (a1, VUInt) then begin
    BN_mul_int_nz (a3, a2, VUInt, VCount2, VSign3);
    Exit;
  end;

  if BN_IsInt (a2, VUInt) then begin
    BN_mul_int_nz (a3, a1, VUInt, VCount1, VSign3);
    Exit;
  end;

  VCount3 := VCount1 + VCount2;
  BN_Init (a3, VSign3, VCount3);

  if BNX_mul_base_asm (a1.PLData, a2.PLData, a3.PLData, VCount1, VCount2) = 0 then
  BN_Pack (a3);
end;

{
  a1 = a1 * a2
}
procedure BN_mul_(var a1: T_BNX; const a2: T_BNX);
var
  VUInt   : UIntX;
  VSign3  : ShortInt;
  VCount1 : IntX;
  VCount2 : IntX;
  VCount3 : IntX;
  VmCount : IntX;
  Vm      : IntX;
  U       : T_BNX;
begin
  VSign3 := BN_mul_div_sign (a1.Sign, a2.Sign);

  if (VSign3 = 0) then begin
    BN_AssignZero (a1);
    Exit;
  end;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VCount1 <= 0) or (VCount2 <= 0) then begin
    BN_AssignZero (a1);
    Exit;
  end;

  {$IFDEF KARATSUBA_V2}
  // VmCount = max (VCount1, VCount2)
  VmCount := VCount1;
  if VCount2 > VmCount then VmCount := VCount2;

  if VmCount > Karatsuba_mul_TCount then begin
    Vm := VmCount shr 1; // VmCount div 2
    BN_mul_karatsuba_(U, a1, a2, Vm);
    U.Sign := VSign3;

    BN_copy (a1, U);
    BN_LFree (U);

    Exit;
  end;
  {$ELSE}
  if (VCount1 > Karatsuba_mul_TCount) and (VCount2 > Karatsuba_mul_TCount) then begin
    // VmCount = max (VCount1, VCount2)
    VmCount := VCount1;
    if VCount2 > VmCount then VmCount := VCount2;

    Vm := VmCount shr 1; // VmCount div 2
    BN_mul_karatsuba_(U, a1, a2, Vm);
    U.Sign := VSign3;

    BN_copy (a1, U);
    BN_LFree (U);

    Exit;
  end;
  {$ENDIF}

  if BN_IsInt (a1, VUInt) then begin
    BN_mul_int_nz (a1, a2, VUInt, VCount2, VSign3);
    Exit;
  end;

  if BN_IsInt (a2, VUInt) then begin
    BN_mul_int_nz_(a1, VUInt, VCount1, VSign3);
    Exit;
  end;

  VCount3 := VCount1 + VCount2;
  BN_Init (U, VSign3, VCount3);

  if BNX_mul_base_asm (a1.PLData, a2.PLData, U.PLData, VCount1, VCount2) = 0 then
  BN_Pack (U);
  BN_copy (a1, U);
  BN_LFree (U);
end;

{
  a3 = a1 * a2
}
procedure BN_mul_int (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX);
var
  VSign3  : ShortInt;
  VCount1 : IntX;
  VCount3 : IntX;
begin
  if (a2 = 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VSign3 := BN_mul_div_sign (a1.Sign, ASign2);

  if (VSign3 = 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount1 := BN_DataCount (a1);

  if (VCount1 <= 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount3 := VCount1 + 1;
  BN_Init (a3, VSign3, VCount3);

  if BNX_mul_int_base_asm (a1.PLData, a2, a3.PLData, VCount1) = 0 then
  BN_Pack (a3);
end;

{
  a3 = a1 * a2
  a2 - signed integer
}
procedure BN_mul_int (var a3: T_BNX; const a1: T_BNX; a2: IntX);
var
  VSign2  : ShortInt;
  VSign3  : ShortInt;
  VCount1 : IntX;
  VCount3 : IntX;
begin
  if (a2 = 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VSign2 := 1;
  if a2 < 0 then begin
    VSign2 := -1;
    a2 := -a2;
  end;

  VSign3 := BN_mul_div_sign (a1.Sign, VSign2);

  if (VSign3 = 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount1 := BN_DataCount (a1);

  if (VCount1 <= 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount3 := VCount1 + 1;
  BN_Init (a3, VSign3, VCount3);

  if BNX_mul_int_base_asm (a1.PLData, a2, a3.PLData, VCount1) = 0 then
  BN_Pack (a3);
end;

{
  a3 = a1 * a2
  a1 <> 0
  a2 <> 0
  a1.Count = ACount1
  a3.Sign  = ASign3
}
procedure BN_mul_int_nz (var a3: T_BNX; const a1: T_BNX; a2: UIntX;
  ACount1: IntX; ASign3: ShortInt);
var
  VCount3 : IntX;
begin
  VCount3 := ACount1 + 1;
  BN_Init (a3, ASign3, VCount3);

  if BNX_mul_int_base_asm (a1.PLData, a2, a3.PLData, ACount1) = 0 then
  BN_Pack (a3);
end;

{
  a1(a3) = a1 * a2
  a1 <> 0
  a2 <> 0
  a1.Count = ACount1
  a3.Sign  = ASign3
}
procedure BN_mul_int_nz_(var a1: T_BNX; a2: UIntX;
  ACount1: IntX; ASign3: ShortInt);
var
  V : T_BNX;
begin
  BN_mul_int_nz (V, a1, a2, ACount1, ASign3);
  BN_copy (a1, V);
  BN_LFree (V);
end;

{
  a3 = a1 * a2'
  a2'[n] = a2,   (other elements are zeros)
}
procedure BN_mul_intn (var a3: T_BNX; const a1: T_BNX;
  ASign2: ShortInt; a2: UIntX; n: IntX);
var
  VSign3  : ShortInt;
  VCount1 : IntX;
  VCount3 : IntX;
begin
  if (a2 = 0) or (n < 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VSign3 := BN_mul_div_sign (a1.Sign, ASign2);

  if (VSign3 = 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount1 := BN_DataCount (a1);

  if (VCount1 <= 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount3 := VCount1 + n + 1;
  BN_Init (a3, VSign3, VCount3);

  if BNX_mul_int_base_asm (a1.PLData, a2, @a3.Data[n], VCount1) = 0 then
  BN_Pack (a3);
end;

{
  a3 = a1 * a2'
  a2'[n] = a2,   (other elements are zeros)
  a2 - signed integer
}
procedure BN_mul_intn (var a3: T_BNX; const a1: T_BNX; a2: IntX; n: IntX);
var
  VSign2 : ShortInt;
begin
  VSign2 := 0;
  if a2 > 0 then VSign2 := 1 else
  if a2 < 0 then begin
    VSign2 := -1;
    a2 := -a2;
  end;

  BN_mul_intn (a3, a1, VSign2, a2, n);
end;

{
  c = a div b
  Result = c
}
function BNX_div_asm (aL, aH, b: UIntX; c: PUIntX): UIntX; overload;
{$IFDEF CPUX64}
// rcx | aL | rax  | adiv
// rdx | aH | amod
// r8  | b  |
// r9  | @c |
asm
       .NOFRAME

        mov     rax, rcx
        div     r8
        mov     [r9], rax
end;
{$ELSE}
// eax | aL | adiv
// edx | aH | amod
// ecx | b  |
//     | @c | ecx
asm
        div     ecx
        mov     ecx, [c]
        mov     [ecx], eax
end;
{$ENDIF}

// [+] 1.2.1
{
  adiv = a div b
}
procedure BNX_div_asm (a: PUIntX; b: UIntX; adiv: PUIntX; Count: IntX); overload;
{$IFDEF CPUX64}
// rcx | @a    |
// rdx | b     | r10
// r8  | @adiv |
// r9  | Count |
asm
       .NOFRAME

        mov     r10, rdx
        xor     rdx, rdx

  @loop:
        mov     rax, [rcx]
        div     r10
        mov     [r8], rax
        dec     r9
        jz      @R
        lea     rcx, [rcx - CXBytes]
        lea     r8, [r8 - CXBytes]
        jmp     @loop

  @R:
end;
{$ELSE}
// eax | @a    | esi
// edx | b     | edi
// ecx | @adiv |
//     | Count | ebx
asm
        push    ebx
        push    esi
        push    edi

        mov     esi, eax
        mov     edi, edx
        mov     ebx, [Count]
        xor     edx, edx

  @loop:
        mov     eax, [esi]
        div     edi
        mov     [ecx], eax
        dec     ebx
        jz      @R
        lea     esi, [esi - CXBytes]
        lea     ecx, [ecx - CXBytes]
        jmp     @loop

  @R:
        pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
  adiv = a div ADivisor
}
function BNX_div_intn_(var adiv: T_BNX; const a: T_BNX;
  ADivisor: UIntX; ADIndex: IntX): UIntX;
var
  VCount  : IntX;
  VRCount : IntX;
  VHIndex : IntX;
  VHiData : UIntX;
  VLoData : UIntX;
begin
  Result := 0;

  VCount := BN_DataCount (a);
  VHIndex := VCount - 1;
  VHiData := a.Data[VHIndex];

  (*
  if ADivisor = 1 then begin
    VRCount := VCount - ADIndex;
    VCount := BN_Init (adiv, a.Sign, VRCount);
    if VCount > 0 then begin
      Result := VHiData;
      adiv.PHData^ := VHiData;
    end;
    Exit;
  end;
  // *)

  if VHiData = ADivisor then begin
    VRCount := VCount - ADIndex;
    VCount := BN_Init (adiv, a.Sign, VRCount);
    if VCount > 0 then begin
      Result := 1;
      adiv.PHData^ := 1;
    end;
    Exit;
  end;

  if VHiData > ADivisor then begin
    VRCount := VCount - ADIndex;
    VCount := BN_Init (adiv, a.Sign, VRCount);
    if VCount > 0 then begin
      Result := BNX_div_asm (VHiData, 0, ADivisor, adiv.PHData);
    end;
    Exit;
  end;

  // VHiData < ADivisor
  VRCount := VHIndex - ADIndex; // VCount - ADIndex - 1;
  VCount := BN_Init (adiv, a.Sign, VRCount);
  if VCount > 0 then begin
    VLoData := a.Data[VHIndex - 1];
    Result := BNX_div_asm (VLoData, VHiData, ADivisor, adiv.PHData);
  end;
end;

{
  adiv = a div ADivisor
}
procedure BNX_div_intnn_(var adiv: T_BNX; const a: T_BNX;
  ADivisor: UIntX; ADIndex: IntX);
var
  VCount  : IntX;
  VRCount : IntX;
  VHIndex : IntX;
  VHiData : UIntX;
begin
  VCount := BN_DataCount (a);
  VHIndex := VCount - 1;
  VHiData := a.Data[VHIndex];

  if VHiData > ADivisor then begin
    VRCount := VCount - ADIndex;
    VCount := BN_Init (adiv, a.Sign, VRCount);
    if VCount > 0 then begin
      BNX_div_asm (a.PHData, ADivisor, adiv.PHData, VCount);
    end;
    Exit;
  end;

  // VHiData < ADivisor
  VRCount := VHIndex - ADIndex; // VCount - ADIndex - 1;
  VCount := BN_Init (adiv, a.Sign, VRCount);
  if VCount > 0 then begin
    BNX_div_asm (a.PHData, ADivisor, adiv.PHData, VCount);
  end;
end;

// [+] 1.2.1
{
  Fast recursive division (Burnikel-Ziegler)
}
procedure RecursiveDivision_(var Q, R: T_BNX; const A, B: T_BNX; n: UIntX);
var
  V_B1 : T_BNX;
  V_B2 : T_BNX;

  // [A1,A2,A3] / [V_B1,V_B2]
  procedure DivThreeHalvesByTwo_(var Q, R: T_BNX; const A1, A2, A3: T_BNX; n: UIntX);
  var
    V_A : T_BNX;
    V_C : T_BNX;
    V_D : T_BNX;
    V_U : T_BNX;
  begin
    // if A1 = B1 ...

    BN_merge (V_A, A2, A1);
    RecursiveDivision_(Q, V_C, V_A, V_B1, n);

    BN_mul (V_D, Q, V_B2);
    BN_merge (V_U, A3, V_C);
    BN_sub (R, V_U, V_D);

    if BN_LT_Zero (R) then begin   // Q is too large by at least one
      // BN_merge (V_B, B2, B1);
      BN_add_(R, B, V_U);
      BN_dec_(Q, V_U);
      if BN_LT_Zero (R) then begin // Q is still too large
        BN_add_(R, B, V_U);      // now R is correct
        BN_dec_(Q, V_U);
      end;
    end;

    BN_LFree (V_A);
    BN_LFree (V_C);
    BN_LFree (V_D);
    BN_LFree (V_U);
  end;

var
  n2   : UIntX;
  V_A1 : T_BNX;
  V_A2 : T_BNX;
  V_A3 : T_BNX;
  V_A4 : T_BNX;
  V_Q1 : T_BNX;
  V_Q2 : T_BNX;
  V_R  : T_BNX;
  V_R1 : T_BNX;
  V_R2 : T_BNX;
begin
  if Odd (n) or (n <= BZ_DIV_LIMIT) then begin
    BN_div_mod (Q, R, A, B); // base of the recursion
    Exit;
  end;

  n2 := n shr 1; // n div 2

  // Let [A1,A2,A3,A4] = A
  BN_split_n4 (V_A4, V_A3, V_A2, V_A1, A);
  // Let [B1,B2] = B
  BN_split_n2 (V_B2, V_B1, B);

  // [Q1,R] = ...
  DivThreeHalvesByTwo_(V_Q1, V_R, V_A1, V_A2, V_A3, n2);
  BN_Resize (V_Q1, n2);
  BN_Resize (V_R, n);

  // Let [R1,R2] = R
  BN_split_n2 (V_R2, V_R1, V_R);

  // [Q2,R] = ...
  DivThreeHalvesByTwo_(V_Q2, R, V_R1, V_R2, V_A4, n2);
  BN_Resize (V_Q2, n2);
  BN_Resize (R, n);

  // Return the quotient Q = [Q1,Q2] and the remainder R
  BN_merge (Q, V_Q2, V_Q1);

  BN_LFree (V_A1);
  BN_LFree (V_A2);
  BN_LFree (V_A3);
  BN_LFree (V_A4);
  BN_LFree (V_B1);
  BN_LFree (V_B2);
  BN_LFree (V_Q1);
  BN_LFree (V_Q2);
  BN_LFree (V_R);
  BN_LFree (V_R1);
  BN_LFree (V_R2);
end;

// [+] 1.2.1
{
  Fast recursive division (Burnikel-Ziegler)

  adiv = A div B
  amod = A mod B

  A > 0, B > 0
  |A| > |B|
}
procedure BN_div_mod_bz_(var adiv, amod: T_BNX; const A, B: T_BNX);
var
  I     : LoopInt;
  V_A   : T_BNX;
  V_Au  : T_BNX;
  V_B   : T_BNX;
  V_Z   : T_BNX;
  V_R   : T_BNX;
  V_Q   : T_BNX_Array;
  r, s  : UIntX;
  m, j  : UIntX;
  n, n2 : UIntX;
  sigma : UIntX;
  t, t2 : UIntX;
begin
  // r := BN_DataCount (A);
  s := BN_DataCount (B);

  // Set m = min {2^k | 2^k * DIV_LIMIT > s}
  m := 1;
  if s >= BZ_DIV_LIMIT then begin
    m := s div BZ_DIV_LIMIT;
    m := UIntX(1) shl BN_Bits (m);
  end;

  // Set j = ceil(s/m) and n = j*m
  j := cdiv (s, m);
  n := j * m;
  n2 := n shl 1; // n*2

  // Set sigma = max {tau | 2^tau * B < base^n}
  sigma := n * CXBits - BN_Bits (B);

  // Set B = B * 2^sigma to normalize B
  BN_acopy (V_B, B); // B = |B|
  BN_pshl (V_B, sigma);
  BN_Resize (V_B, n);

  // Set A = A * 2^sigma to shift A by the same amount as B
  BN_acopy (V_A, A); // A = |A|
  BN_pshl (V_A, sigma);
  r := BN_DataCount (V_A);

  // Set t = min {u>=2 | A < base^(u*n) / 2}
  t := (r div n) + 1;
  if Odd (t) then Inc (t);
  t2 := t - 2;

  BN_Resize (V_A, t * n);
  SetLength (V_Q, t - 1);

  // Split A conceptually into t blocks [A(t-1),...,A(0)] = A
  // Note that the leftmost bit of A(t-1) is 0 by our choice of t

  // Set Z(t-2) = [A(t-1),A(t-2)]
  m := n * t2;
  BN_hcopy (V_Z, V_A, m, n2);

  for I := t2 downto 0 do begin
    // Using algorithm D(2n/1n) compute Q(i), R(i) such that Z(i) = B*Q(i) + R(i)
    RecursiveDivision_(V_Q[I], V_R, V_Z, V_B, n);
    if I = 0 then Break;

    // Set Z(i-1) = [R(i),A(i-1)]
    Dec (m, n);
    BN_hcopy (V_Au, V_A, m, n);
    BN_merge (V_Z, V_Au, V_R);
  end;

  BN_merge (adiv, V_Q);
  BN_Pack (adiv);

  BN_pshr (amod, V_R, sigma);
  BN_Pack (amod);

  BN_LFree (V_A);
  BN_LFree (V_Au);
  BN_LFree (V_B);
  BN_LFree (V_Z);
  BN_LFree (V_R);
  BN_LFree (V_Q);
end;

// [+] 1.2.1
{
  Fast recursive division (Burnikel-Ziegler)

  amod = A mod B

  A > 0, B > 0
  |A| > |B|
}
procedure BN_mod_bz_(var amod: T_BNX; const A, B: T_BNX);
var
  I     : LoopInt;
  V_A   : T_BNX;
  V_Au  : T_BNX;
  V_B   : T_BNX;
  V_Z   : T_BNX;
  V_R   : T_BNX;
  V_Q   : T_BNX;
  r, s  : UIntX;
  m, j  : UIntX;
  n, n2 : UIntX;
  sigma : UIntX;
  t, t2 : UIntX;
begin
  // r := BN_DataCount (A);
  s := BN_DataCount (B);

  // Set m = min {2^k | 2^k * DIV_LIMIT > s}
  m := 1;
  if s >= BZ_DIV_LIMIT then begin
    m := s div BZ_DIV_LIMIT;
    m := UIntX(1) shl BN_Bits (m);
  end;

  // Set j = ceil(s/m) and n = j*m
  j := cdiv (s, m);
  n := j * m;
  n2 := n shl 1; // n*2

  // Set sigma = max {tau | 2^tau * B < base^n}
  sigma := n * CXBits - BN_Bits (B);

  // Set B = B * 2^sigma to normalize B
  BN_acopy (V_B, B); // B = |B|
  BN_pshl (V_B, sigma);
  BN_Resize (V_B, n);

  // Set A = A * 2^sigma to shift A by the same amount as B
  BN_acopy (V_A, A); // A = |A|
  BN_pshl (V_A, sigma);
  r := BN_DataCount (V_A);

  // Set t = min {u>=2 | A < base^(u*n) / 2}
  t := (r div n) + 1;
  if Odd (t) then Inc (t);
  t2 := t - 2;

  BN_Resize (V_A, t * n);

  // Split A conceptually into t blocks [A(t-1),...,A(0)] = A
  // Note that the leftmost bit of A(t-1) is 0 by our choice of t

  // Set Z(t-2) = [A(t-1),A(t-2)]
  m := n * t2;
  BN_hcopy (V_Z, V_A, m, n2);

  for I := t2 downto 0 do begin
    // Using algorithm D(2n/1n) compute Q(i), R(i) such that Z(i) = B*Q(i) + R(i)
    RecursiveDivision_(V_Q, V_R, V_Z, V_B, n);
    if I = 0 then Break;

    // Set Z(i-1) = [R(i),A(i-1)]
    Dec (m, n);
    BN_hcopy (V_Au, V_A, m, n);
    BN_merge (V_Z, V_Au, V_R);
  end;

  BN_pshr (amod, V_R, sigma);
  BN_Pack (amod);

  BN_LFree (V_A);
  BN_LFree (V_Au);
  BN_LFree (V_B);
  BN_LFree (V_Z);
  BN_LFree (V_R);
  BN_LFree (V_Q);
end;

{
  adiv = a1 div a2
  amod = a1 mod a2, amod > 0
}
procedure BN_div_mod (var adiv, amod: T_BNX; const a1, a2: T_BNX);
var
  V_u      : T_BNX;
  V_xdiv   : T_BNX;
  V_nn     : array [0..3] of T_BNX;
  I_U      : Integer;
  I_d0     : Integer;
  I_d1     : Integer;
  I_m0     : Integer;
  I_m1     : Integer;

  VDIndex  : IntX;
  VDivisor : UIntX;

  procedure div_mod_(const A_a1, A_a2: T_BNX);
  var
    VDiv     : UIntX;
    VCompare : Integer;
  begin
    //   
    VDiv := BNX_div_intn_(V_nn[I_d0], A_a1, VDivisor, VDIndex);
    BN_mul_intn (V_u, A_a2, V_nn[I_d0].Sign, VDiv, V_nn[I_d0].HIndex); // u = adiv * a2
    BN_sub (V_nn[I_m0], A_a1, V_u); // amod = a1 - adiv * a2

    while true do begin
      VCompare := BN_DataCompare (V_nn[I_m0], A_a2);

      if VCompare = 0 then begin
        //  
        //     
        if V_nn[I_m0].Sign < 0 then begin
          BN_dec (adiv, V_nn[I_d0]); // adiv = adiv - 1
        end else begin
          BN_inc (adiv, V_nn[I_d0]); // adiv = adiv + 1
        end;
        BN_AssignZero (amod);
        Exit;
      end;

      if VCompare < 0 then begin
        //  
        //     
        if V_nn[I_m0].Sign < 0 then begin
          BN_dec (adiv, V_nn[I_d0]);       // adiv = adiv - 1
          BN_add (amod, V_nn[I_m0], A_a2); // amod = amod + a2
        end else begin
          BN_copy (adiv, V_nn[I_d0]);
          BN_copy (amod, V_nn[I_m0]);
        end;
        Exit;
      end;

      //   
      VDiv := BNX_div_intn_(V_xdiv, V_nn[I_m0], VDivisor, VDIndex);
      BN_mul_intn (V_u, A_a2, V_xdiv.Sign, VDiv, V_xdiv.HIndex); // u = adiv * a2
      BN_sub (V_nn[I_m1], V_nn[I_m0], V_u); // amod = amod - adiv * a2
      BN_add (V_nn[I_d1], V_nn[I_d0], V_xdiv);

      I_U := I_d0;
      I_d0 := I_d1;
      I_d1 := I_U;

      I_U := I_m0;
      I_m0 := I_m1;
      I_m1 := I_U;
    end;
  end;

var
  V_a1     : T_BNX;
  V_a2     : T_BNX;
  V_u1     : T_BNX;
  V_u2     : T_BNX;
  VShift   : IntX;
  VSign1   : ShortInt;
  VSign2   : ShortInt;
  VSign    : ShortInt;
  VCount1  : IntX;
  VCount2  : IntX;
  VData    : UIntX;
  VCompare : Integer;
begin
  VSign1 := a1.Sign;
  VSign2 := a2.Sign;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VSign2 = 0) or (VCount2 <= 0) then begin
    // |a2| = 0
    raise EDivByZero.Create (SDivByZero);
  end;

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // |a1| = 0
    // adiv = 0
    // amod = 0
    BN_AssignZero (adiv);
    BN_AssignZero (amod);
    Exit;
  end;

  VSign := BN_mul_div_sign (VSign1, VSign2);
  VCompare := BN_DataCompare (a1, a2);

  if VCompare = 0 then begin
    // |a1| = |a2|
    // adiv = 1
    // amod = 0
    BN_AssignOne (adiv, VSign);
    BN_AssignZero (amod);
    Exit;
  end;

  if VCompare < 0 then begin
    // |a1| < |a2|
    // adiv = 0
    // amod = |a1|
    BN_AssignZero (adiv);
    BN_acopy (amod, a1); // amod = |a1|
    Exit;
  end;

  // |a1| > |a2|

  try
    if VCount2 > BZ_DIV_LIMIT then begin
      BN_acopy (V_a1, a1); // a1 = |a1|
      BN_acopy (V_a2, a2); // a2 = |a2|
      BN_div_mod_bz_(adiv, amod, V_a1, V_a2);

      adiv.Sign := VSign;

      // amod.Sign := VSign1;
      if VSign1 < 0 then begin
        if BN_IsNotZero (amod) then begin
          BN_sub (V_u, V_a2, amod);
          BN_copy (amod, V_u);
          if VSign < 0 then begin
            BN_dec_(adiv, V_u);
          end else begin
            BN_inc_(adiv, V_u);
          end;
        end;
      end;

      Exit;
    end;

    I_d0 := 0; // V_adiv = V_nn[I_d0]
    I_d1 := 1; // V_adiu = V_nn[I_d1]
    I_m0 := 2; // V_amod = V_nn[I_m0]
    I_m1 := 3; // V_amou = V_nn[I_m1]

    VDIndex := VCount2 - 1;
    VDivisor := a2.Data[VDIndex];

    if VCount2 = 1 then begin
      VData := BN_div_mod_int (adiv, a1, 1, VDivisor);
      if VData = 0 then begin
        BN_AssignZero (amod);
      end else begin
        BN_AssignInt (amod, 1, VData);
      end;
      adiv.Sign := VSign;
      Exit;
    end;

    VShift := CXBits - BN_Bits (VDivisor);
    if VShift > 0 then begin
      //  VDivisor
      BN_acopy (V_a1, a1); // a1 = |a1|
      BN_acopy (V_a2, a2); // a2 = |a2|
      BN_pshl (V_u1, V_a1, VShift);
      BN_pshl (V_u2, V_a2, VShift);
      VDivisor := V_u2.Data[VDIndex];

      div_mod_(V_u1, V_u2);
      BN_pshr (amod, VShift);
    end else begin
      BN_acopy (V_a1, a1); // a1 = |a1|
      BN_acopy (V_a2, a2); // a2 = |a2|

      div_mod_(V_a1, V_a2);
    end;

    adiv.Sign := VSign;

    // amod.Sign := VSign1;
    if VSign1 < 0 then begin
      if BN_IsNotZero (amod) then begin
        BN_sub (V_u, V_a2, amod);
        BN_copy (amod, V_u);
        if VSign < 0 then begin
          BN_dec_(adiv, V_u);
        end else begin
          BN_inc_(adiv, V_u);
        end;
      end;
    end;
  finally
    BN_LFree (V_a1);
    BN_LFree (V_a2);
    BN_LFree (V_u1);
    BN_LFree (V_u2);
    BN_LFree (V_u);
    BN_LFree (V_xdiv);
    BN_LFree (V_nn[0]);
    BN_LFree (V_nn[1]);
    BN_LFree (V_nn[2]);
    BN_LFree (V_nn[3]);
  end;
end;

{
  amod = a1 mod a2, amod > 0
}
procedure BN_mod (var amod: T_BNX; const a1, a2: T_BNX);
var
  V_u      : T_BNX;
  V_xdiv   : T_BNX;
  V_nn     : array [0..1] of T_BNX;
  I_U      : Integer;
  I_m0     : Integer;
  I_m1     : Integer;

  VDIndex  : IntX;
  VDivisor : UIntX;

  procedure mod_(const A_a1, A_a2: T_BNX);
  var
    VDiv     : UIntX;
    VCompare : Integer;
  begin
    //   
    VDiv := BNX_div_intn_(V_xdiv, A_a1, VDivisor, VDIndex);
    BN_mul_intn (V_u, A_a2, V_xdiv.Sign, VDiv, V_xdiv.HIndex); // u = adiv * a2
    BN_sub (V_nn[I_m0], A_a1, V_u); // amod = a1 - adiv * a2

    while true do begin
      VCompare := BN_DataCompare (V_nn[I_m0], A_a2);

      if VCompare = 0 then begin
        //  
        //     
        BN_AssignZero (amod);
        Exit;
      end;

      if VCompare < 0 then begin
        if V_nn[I_m0].Sign < 0 then begin
          BN_add (amod, V_nn[I_m0], A_a2); // amod = amod + a2
        end else begin
          BN_copy (amod, V_nn[I_m0]);
        end;
        Exit;
      end;

      //   
      VDiv := BNX_div_intn_(V_xdiv, V_nn[I_m0], VDivisor, VDIndex);
      BN_mul_intn (V_u, A_a2, V_xdiv.Sign, VDiv, V_xdiv.HIndex); // V_u = adiv * a2
      BN_sub (V_nn[I_m1], V_nn[I_m0], V_u); // amod = amod - adiv * a2

      I_U := I_m0;
      I_m0 := I_m1;
      I_m1 := I_U;
    end;
  end;

var
  V_a1     : T_BNX;
  V_a2     : T_BNX;
  V_u1     : T_BNX;
  V_u2     : T_BNX;
  VShift   : IntX;
  VSign1   : ShortInt;
  VSign2   : ShortInt;
  VCount1  : IntX;
  VCount2  : IntX;
  VData    : UIntX;
  VCompare : Integer;
begin
  VSign1 := a1.Sign;
  VSign2 := a2.Sign;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VSign2 = 0) or (VCount2 <= 0) then begin
    // |a2| = 0
    raise EDivByZero.Create (SDivByZero);
  end;

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // |a1| = 0
    // adiv = 0
    // amod = 0
    BN_AssignZero (amod);
    Exit;
  end;

  VCompare := BN_DataCompare (a1, a2);

  if VCompare = 0 then begin
    // |a1| = |a2|
    // adiv = 1
    // amod = 0
    BN_AssignZero (amod);
    Exit;
  end;

  if VCompare < 0 then begin
    // |a1| < |a2|
    // adiv = 0
    // amod = |a1|
    BN_acopy (amod, a1); // amod = |a1|
    Exit;
  end;

  // |a1| > |a2|

  try
    if VCount2 > BZ_DIV_LIMIT then begin
      BN_acopy (V_a1, a1); // a1 = |a1|
      BN_acopy (V_a2, a2); // a2 = |a2|
      BN_mod_bz_(amod, V_a1, V_a2);

      // amod.Sign := VSign1;
      if VSign1 < 0 then begin
        if BN_IsNotZero (amod) then begin
          BN_sub (V_u, V_a2, amod);
          BN_copy (amod, V_u);
        end;
      end;

      Exit;
    end;

    I_m0 := 0; // V_amod = V_nn[I_m0]
    I_m1 := 1; // V_amou = V_nn[I_m1]

    VDIndex := VCount2 - 1;
    VDivisor := a2.Data[VDIndex];

    if VCount2 = 1 then begin
      VData := BN_div_mod_int (V_a1, a1, 1, VDivisor);
      if VData = 0 then begin
        BN_AssignZero (amod);
      end else begin
        BN_AssignInt (amod, 1, VData);
      end;
      BN_LFree (V_a1);
      Exit;
    end;

    VShift := CXBits - BN_Bits (VDivisor);
    if VShift > 0 then begin
      //  VDivisor
      BN_acopy (V_a1, a1); // a1 = |a1|
      BN_acopy (V_a2, a2); // a2 = |a2|
      BN_pshl (V_u1, V_a1, VShift);
      BN_pshl (V_u2, V_a2, VShift);
      VDivisor := V_u2.Data[VDIndex];

      mod_(V_u1, V_u2);
      BN_pshr (amod, VShift);
    end else begin
      BN_acopy (V_a1, a1); // a1 = |a1|
      BN_acopy (V_a2, a2); // a2 = |a2|

      mod_(V_a1, V_a2);
    end;

    // amod.Sign := VSign1;
    if VSign1 < 0 then begin
      if BN_IsNotZero (amod) then begin
        BN_sub (V_u, V_a2, amod);
        BN_copy (amod, V_u);
      end;
    end;
  finally
    BN_LFree (V_a1);
    BN_LFree (V_a2);
    BN_LFree (V_u1);
    BN_LFree (V_u2);
    BN_LFree (V_u);
    BN_LFree (V_xdiv);
    BN_LFree (V_nn[0]);
    BN_LFree (V_nn[1]);
  end;
end;

{
  adiv = a1 div a2

  Result -  
  Result - number of iterations
}
procedure BN_div (var adiv: T_BNX; const a1, a2: T_BNX);
var
  V_mod : T_BNX;
begin
  BN_div_mod (adiv, V_mod, a1, a2);
  BN_LFree (V_mod);
end;

{
  a1 div a2
  a1 mod a2
}
function BNX_div_mod_int_asm (a1: PUIntX; a2: UIntX; Count: IntX;
  adiv: PUIntX): UIntX;
{$IFDEF CPUX64}
// rcx | @a1   |
// rdx | a2    | r10
// r8  | Count |
// r9  | @adiv |
asm
       .NOFRAME

        mov     r10, rdx
        xor     rdx, rdx
  @loop:
        mov     rax, [rcx]
        div     r10
        mov     [r9], rax
        lea     rcx, [rcx - CXBytes]
        lea     r9, [r9 - CXBytes]
  @next:
        dec     r8
        jnz     @loop

        mov     rax, rdx
end;
{$ELSE}
// eax | @a1   | esi
// edx | a2    | edi
// ecx | Count |
//     | @adiv | ebx
asm
        push    ebx
        push    esi
        push    edi

        mov     esi, eax
        mov     ebx, [adiv]
        mov     edi, edx
        xor     edx, edx
  @loop:
        mov     eax, [esi]
        div     edi
        mov     [ebx], eax
        lea     esi, [esi - CXBytes]
        lea     ebx, [ebx - CXBytes]
  @next:
        dec     ecx
        jnz     @loop

        mov     eax, edx

        pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
  adiv = a1 div a2
  amod = a1 mod a2, amod > 0
  Result = amod
}
function BN_div_mod_int (var adiv: T_BNX; const a1: T_BNX;
  ASign2: ShortInt; a2: UIntX; APack: Boolean = true): UIntX;
var
  VSign  : ShortInt;
  VCount : IntX;
begin
  if a2 = 0 then begin
    raise EDivByZero.Create (SDivByZero);
  end;

  Result := 0;
  VSign := BN_mul_div_sign (a1.Sign, ASign2);

  if a2 = 1 then begin
    BN_scopy (adiv, VSign, a1);
    if APack then BN_Pack (adiv);
    Exit;
  end;

  VCount := BN_Init (adiv, VSign, a1.Count);
  if VCount <= 0 then Exit;

  Result := BNX_div_mod_int_asm (a1.PHData, a2, VCount, adiv.PHData);
  if APack then BN_Pack (adiv);
end;

{
  adiv = a1 div a2
  amod = a1 mod a2, amod > 0
  a2 - signed integer
  Result = amod
}
function BN_div_mod_int (var adiv: T_BNX; const a1: T_BNX; a2: IntX;
  APack: Boolean = true): UIntX;
var
  VSign2 : ShortInt;
begin
  if a2 = 0 then begin
    raise EDivByZero.Create (SDivByZero);
  end;

  if a2 > 0 then VSign2 := 1
  else begin
    VSign2 := -1;
    a2 := -a2;
  end;

  Result := BN_div_mod_int (adiv, a1, VSign2, a2, APack);
end;

{
  adiv = a1 div a2'
  amod = a1 mod a2', amod > 0
  a2'[n] = a2,   (other elements are zeros)
  Result = amod
}
function BN_div_mod_intn (var adiv: T_BNX; const a1: T_BNX;
  ASign2: ShortInt; a2: UIntX; n: IntX): UIntX;
var
  VSign  : ShortInt;
  VCount : IntX;
begin
  if a2 = 0 then begin
    raise EDivByZero.Create (SDivByZero);
  end;

  Result := 0;
  VSign := BN_mul_div_sign (a1.Sign, ASign2);
  VCount := BN_Init (adiv, VSign, a1.Count - n);
  if VCount <= 0 then Exit;

  Result := BNX_div_mod_int_asm (a1.PHData, a2, VCount, adiv.PHData);
  BN_Pack (adiv);
end;

{
  adiv = a1 div a2'
  amod = a1 mod a2', amod > 0
  a2'[n] = a2,   (other elements are zeros)
  a2 - signed integer
  Result = amod
}
function BN_div_mod_intn (var adiv: T_BNX; const a1: T_BNX; a2: IntX;
  n: IntX): UIntX;
var
  VSign2 : ShortInt;
begin
  if a2 = 0 then begin
    raise EDivByZero.Create (SDivByZero);
  end;

  if a2 > 0 then VSign2 := 1
  else begin
    VSign2 := -1;
    a2 := -a2;
  end;

  Result := BN_div_mod_intn (adiv, a1, VSign2, a2, n);
end;

{
  adiv = a1 div a2
}
procedure BN_div_int (var adiv: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX);
begin
  BN_div_mod_int (adiv, a1, ASign2, a2);
end;

{
  adiv = a1 div a2
  a2 - signed integer
}
procedure BN_div_int (var adiv: T_BNX; const a1: T_BNX; a2: IntX);
begin
  BN_div_mod_int (adiv, a1, a2);
end;

{
  adiv = a1 div a2'
  a2'[n] = a2,   (other elements are zeros)
}
procedure BN_div_intn (var adiv: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX;
  n: IntX);
begin
  BN_div_mod_intn (adiv, a1, ASign2, a2, n);
end;

{
  adiv = a1 div a2'
  a2'[n] = a2,   (other elements are zeros)
  a2 - signed integer
}
procedure BN_div_intn (var adiv: T_BNX; const a1: T_BNX; a2: IntX;
  n: IntX);
begin
  BN_div_mod_intn (adiv, a1, a2, n);
end;

// [+] 1.2.1
{
  amod = a1 mod a2, amod > 0
  Result = amod
}
function BN_mod_int (const a1: T_BNX; ASign2: ShortInt; a2: UIntX): UIntX;
var
  V : T_BNX;
begin
  Result := BN_div_mod_int (V, a1, ASign2, a2);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  amod = a1 mod a2, amod > 0
  a2 - signed integer
  Result = amod
}
function BN_mod_int (const a1: T_BNX; a2: IntX): UIntX;
var
  V : T_BNX;
begin
  Result := BN_div_mod_int (V, a1, a2);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  amod = a1 mod a2', amod > 0
  a2'[n] = a2,   (other elements are zeros)
  Result = amod
}
function BN_mod_intn (const a1: T_BNX; ASign2: ShortInt; a2: UIntX; n: IntX): UIntX;
var
  V : T_BNX;
begin
  Result := BN_div_mod_intn (V, a1, ASign2, a2, n);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  amod = a1 mod a2', amod > 0
  a2'[n] = a2,   (other elements are zeros)
  a2 - signed integer
  Result = amod
}
function BN_mod_intn (const a1: T_BNX; a2: IntX; n: IntX): UIntX;
var
  V : T_BNX;
begin
  Result := BN_div_mod_intn (V, a1, a2, n);
  BN_LFree (V);
end;

// [+] 1.2.1
{
  adiv = a div 2^n
  amod = a mod 2^n
}
procedure BN_div_mod_2n (var adiv, amod: T_BNX; const a: T_BNX; n: IntX);
var
  V_div  : IntX;
  V_mod  : IntX;
  V_mask : UIntX;
begin
  if BN_IsZero (a) then begin
    BN_AssignZero (adiv);
    BN_AssignZero (amod);
    Exit;
  end;

  if n <= 0 then begin
    BN_copy (adiv, a);
    BN_AssignZero (amod);
    Exit;
  end;

  // div

  BN_pshr (adiv, a, n);
  if BN_IsZero (adiv) then begin
    BN_copy (amod, a);
    Exit;
  end;

  // mod

  V_div := cdiv_mod (n, CXBits, V_mod);
  if V_div > a.Count then begin
    BN_copy (amod, a);
    Exit;
  end;
  BN_ccopy (amod, a, V_div);
  if V_mod = 0 then Exit;

  V_mask := CXmask shr (CXBits - V_mod);
  amod.PHData^ := amod.PHData^ and V_mask;
end;

// [+] 1.2.1
{
  adiv = a div 2^n
}
procedure BN_div_2n (var adiv: T_BNX; const a: T_BNX; n: IntX);
begin
  if BN_IsZero (a) then begin
    BN_AssignZero (adiv);
    Exit;
  end;

  if n <= 0 then begin
    BN_copy (adiv, a);
    Exit;
  end;

  // div

  BN_pshr (adiv, a, n);
end;

// [+] 1.2.1
{
  a = a div 2^n
}
procedure BN_div_2n_(var a: T_BNX; n: IntX);
begin
  if BN_IsZero (a) then begin
    BN_AssignZero (a);
    Exit;
  end;

  if n <= 0 then begin
    // BN_copy (a, a);
    Exit;
  end;

  // div

  BN_pshr (a, n);
end;

// [+] 1.2.1
{
  amod = a mod 2^n
}
procedure BN_mod_2n (var amod: T_BNX; const a: T_BNX; n: IntX);
var
  V_div  : IntX;
  V_mod  : IntX;
  V_mask : UIntX;
begin
  if BN_IsZero (a) then begin
    BN_AssignZero (amod);
    Exit;
  end;

  if n <= 0 then begin
    BN_AssignZero (amod);
    Exit;
  end;

  // mod

  V_div := cdiv_mod (n, CXBits, V_mod);
  if V_div > a.Count then begin
    BN_copy (amod, a);
    Exit;
  end;
  BN_ccopy (amod, a, V_div);
  if V_mod = 0 then Exit;

  V_mask := CXmask shr (CXBits - V_mod);
  amod.PHData^ := amod.PHData^ and V_mask;
end;

// [+] 1.2.1
{
  Result = (a mod 2^n).LO
}
function BN_mod_2n (const a: T_BNX; n: IntX): UIntX;
var
  V_div  : IntX;
  V_mod  : IntX;
  V_mask : UIntX;
begin
  Result := 0;

  if BN_IsZero (a) then Exit;
  if n <= 0 then Exit;

  // mod

  V_div := cdiv_mod (n, CXBits, V_mod);
  if V_div > 1 then begin
    Result := a.Data[0];
    Exit;
  end;

  V_mask := CXmask shr (CXBits - V_mod);
  Result := a.Data[0] and V_mask;
end;

// [+] 1.2.1
{
  a = a mod 2^n
}
procedure BN_mod_2n_(var a: T_BNX; n: IntX);
var
  V_div  : IntX;
  V_mod  : IntX;
  V_mask : UIntX;
begin
  if BN_IsZero (a) then begin
    BN_AssignZero (a);
    Exit;
  end;

  if n <= 0 then begin
    BN_AssignZero (a);
    Exit;
  end;

  // mod

  V_div := cdiv_mod (n, CXBits, V_mod);
  if V_div > a.Count then begin
    // BN_copy (a, a);
    Exit;
  end;
  if V_div <> a.Count then begin
    BN_Resize (a, V_div);
  end;
  if V_mod = 0 then Exit;

  V_mask := CXmask shr (CXBits - V_mod);
  a.PHData^ := a.PHData^ and V_mask;
end;

{
  P = 2^AExp
}
procedure BN_exp2 (var P: T_BNX; ASign: ShortInt; AExp: IntX);
begin
  if AExp < 0 then begin
    BN_AssignZero (P); // P = 0
    Exit;
  end;

  if AExp = 1 then begin
    BN_AssignInt (P, ASign, 2); // P = 2
    Exit;
  end;

  BN_AssignPOne (P); // P = 1
  if AExp = 0 then Exit;

  BN_pshl (P, AExp);
  P.Sign := ASign;
end;

{
  P = 10^AExp
    = 5^AExp * 2^AExp
}
procedure BN_exp10 (var P: T_BNX; ASign: ShortInt; AExp: IntX);
var
  VBase : T_BNX;
  VExp  : IntX;
begin
  if AExp < 0 then begin
    BN_AssignZero (P); // P = 0
    Exit;
  end;

  if AExp = 1 then begin
    BN_AssignInt (P, ASign, 10); // P = 10
    Exit;
  end;

  BN_AssignPOne (P); // P = 1
  if AExp = 0 then Exit;

  VExp := AExp;
  BN_AssignInt (VBase, ASign, 5);
  while AExp <> 0 do begin
    while not Odd (AExp) do begin
      AExp := AExp shr 1;
      BN_sqr_(VBase);  // VBase = VBase * VBase
    end;
    Dec (AExp);
    BN_mul_(P, VBase); // P = P * VBase
  end;
  BN_pshl (P, VExp);   // P = P * 2^Exp

  BN_LFree (VBase);
end;

{
  P = ABase^AExp
  Using BN_mul_base
}
procedure BN_pow_base (var P: T_BNX; const ABase: T_BNX; AExp: UIntX);
var
  V     : T_BNX;
  VBase : T_BNX;
begin
  if AExp = 1 then begin
    BN_copy (P, ABase); // P = ABase
    Exit;
  end;

  BN_AssignPOne (P); // P = 1
  if AExp = 0 then Exit;

  BN_copy (VBase, ABase);
  while AExp <> 0 do begin
    while not Odd (AExp) do begin
      AExp := AExp shr 1;
      BN_mul_base (V, VBase, VBase);
      BN_copy (VBase, V); // VBase = VBase * VBase
    end;
    Dec (AExp);
    BN_copy (V, P);
    BN_mul_base (P, V, VBase); // P = P * VBase
  end;

  BN_LFree (V);
  BN_LFree (VBase);
end;

// [+] 1.2.1
{
  P = ABase^AExp
  Using BN_mul_base
}
procedure BN_pow_base (var P: T_BNX; const ABase, AExp: T_BNX);
var
  V     : T_BNX;
  VBase : T_BNX;
  VExp  : T_BNX;
begin
  if AExp.Sign < 0 then begin
    BN_AssignZero (P); // P = 0
    Exit;
  end;

  if BN_IsOne (AExp) then begin
    BN_copy (P, ABase); // P = ABase
    Exit;
  end;

  BN_AssignPOne (P); // P = 1
  if BN_IsZero (AExp) then Exit;

  BN_copy (VBase, ABase);
  BN_copy (VExp, AExp);
  while BN_IsNotZero (VExp) do begin
    while BN_IsEven (VExp) do begin
      BN_shr_1 (VExp);
      BN_mul_base (V, VBase, VBase);
      BN_copy (VBase, V); // VBase = VBase * VBase
    end;
    BN_dec_(VExp, V);
    BN_copy (V, P);
    BN_mul_base (P, V, VBase); // P = P * VBase
  end;

  BN_LFree (V);
  BN_LFree (VBase);
  BN_LFree (VExp);
end;

{
  P = ABase^AExp
}
procedure BN_pow (var P: T_BNX; const ABase: T_BNX; AExp: UIntX);
var
  VBase : T_BNX;
begin
  if AExp = 1 then begin
    BN_copy (P, ABase); // P = ABase
    Exit;
  end;

  BN_AssignPOne (P); // P = 1
  if AExp = 0 then Exit;

  BN_copy (VBase, ABase);
  while AExp <> 0 do begin
    while not Odd (AExp) do begin
      AExp := AExp shr 1;
      BN_sqr_(VBase);  // VBase = VBase * VBase
    end;
    Dec (AExp);
    BN_mul_(P, VBase); // P = P * VBase
  end;

  BN_LFree (VBase);
end;

// [+] 1.2.1
{
  P = ABase^AExp
}
procedure BN_pow (var P: T_BNX; const ABase, AExp: T_BNX);
var
  VBase : T_BNX;
  VExp  : T_BNX;
begin
  if AExp.Sign < 0 then begin
    BN_AssignZero (P); // P = 0
    Exit;
  end;

  if BN_IsOne (AExp) then begin
    BN_copy (P, ABase); // P = ABase
    Exit;
  end;

  BN_AssignPOne (P); // P = 1
  if BN_IsZero (AExp) then Exit;

  BN_copy (VBase, ABase);
  BN_copy (VExp, AExp);
  while BN_IsNotZero (VExp) do begin
    while BN_IsEven (VExp) do begin
      BN_shr_1 (VExp);
      BN_sqr_(VBase);  // VBase = VBase * VBase
    end;
    BN_dec_(VExp);
    BN_mul_(P, VBase); // P = P * VBase
  end;

  BN_LFree (VBase);
  BN_LFree (VExp);
end;

{
  P = ABase^AExp
}
procedure BN_pow_int (var P: T_BNX; ASign: ShortInt; ABase: UIntX; AExp: UIntX);
var
  VBase : T_BNX;
begin
  if AExp = 1 then begin
    BN_AssignInt (P, ASign, ABase); // P = ABase
    Exit;
  end;

  BN_AssignPOne (P); // P = 1
  if AExp = 0 then Exit;

  BN_AssignInt (VBase, ASign, ABase);
  while AExp <> 0 do begin
    while not Odd (AExp) do begin
      AExp := AExp shr 1;
      BN_sqr_(VBase);  // VBase = VBase * VBase
    end;
    Dec (AExp);
    BN_mul_(P, VBase); // P = P * VBase
  end;

  BN_LFree (VBase);
end;

{
  P = ABase^AExp
  ABase - signed integer
}
procedure BN_pow_int (var P: T_BNX; ABase: IntX; AExp: UIntX);
var
  VBase : T_BNX;
begin
  if AExp = 1 then begin
    BN_AssignInt (P, ABase); // P = ABase
    Exit;
  end;

  BN_AssignPOne (P); // P = 1
  if AExp = 0 then Exit;

  BN_AssignInt (VBase, ABase);
  while AExp <> 0 do begin
    while not Odd (AExp) do begin
      AExp := AExp shr 1;
      BN_sqr_(VBase);  // VBase = VBase * VBase
    end;
    Dec (AExp);
    BN_mul_(P, VBase); // P = P * VBase
  end;

  BN_LFree (VBase);
end;

{
  P = ABase^2
}
procedure BN_sqr_base (var P: T_BNX; const ABase: T_BNX);
var
  V : T_BNX;
begin
  BN_mul_base (V, ABase, ABase); // V = ABase^2
  BN_copy (P, V);                // P = ABase^2
  BN_LFree (V);
end;

{
  ABase = ABase^2
}
procedure BN_sqr_base_(var ABase: T_BNX);
var
  V : T_BNX;
begin
  BN_mul_base (V, ABase, ABase); // V = ABase^2
  BN_copy (ABase, V);            // ABase = ABase^2
  BN_LFree (V);
end;

// [+] 1.2.1
{
  ABase = ABase^2
  imv - external intermediate variable
}
procedure BN_sqr_base_(var ABase: T_BNX; var imv: T_BNX);
begin
  BN_mul_base (imv, ABase, ABase); // V = ABase^2
  BN_copy (ABase, imv);            // ABase = ABase^2
end;

{
  P = ABase^2

   
  Karatsuba algorithm
}
procedure BN_sqr_karatsuba_(var P: T_BNX; const ABase: T_BNX; Am: IntX);
var
  V_a0   : T_BNX;
  V_a1   : T_BNX;
  V_aa   : T_BNX;
  V_aa00 : T_BNX;
  V_aa11 : T_BNX;
  V_aaaa : T_BNX;
begin
  // (a0 + a1 * 2^[m]) * (a0 + a1 * 2^[m]) =
  // = (a0 * a0) + (2 * a0 * a1) * 2^[m] + (a1 * a1) * 2^[2m]
  // | aa00 = (a0 * a0)
  // | aa11 = (a1 * a1)
  // | aaaa = (a0 + a1) * (a0 + a1) = (a0 * a0) + (a1 * a1) + (2 * a0 * a1)
  // | (2 * a0 * a1) = (a0 + a1) * (a0 + a1) - ((a0 * a0) + (a1 * a1))
  // | = aaaa - (aa00 + aa11)
  // = aa00 + (aaaa - (aa00 + aa11)) *  2^[m] + aa11 * 2^[2m]

  BN_split (V_a0, V_a1, ABase, Am);

  BN_add (V_aa, V_a0, V_a1);     // V_aa = V_a0 + V_a1

  BN_sqr (V_aa00, V_a0);         // V_aa00 = a0 * a0
  BN_sqr (V_aa11, V_a1);         // V_aa11 = a1 * a1
  BN_sqr (V_aaaa, V_aa);         // V_aaaa = (a0 + a1) * (a0 + a1)

  BN_add (V_a0, V_aa00, V_aa11); // V_a0 = V_aa00 + V_aa11
  BN_sub (V_aa, V_aaaa, V_a0);   // V_aa = V_aaaa - (V_aa00 + V_aa11)
  BN_pcshl (V_aa, Am);           // V_aa = (V_aaaa - (V_aa00 + V_aa11)) * 2^[m]
  BN_pcshl (V_aa11, Am shl 1);   // V_aa11 = V_aa11 * 2^[2m]
  BN_add (V_a0, V_aa00, V_aa);   // V_a0 = V_aa00 + (V_aaaa - (V_aa00 + V_aa11)) * 2^[m]
  BN_add (P, V_a0, V_aa11);      // P = V_aa00 + (V_aaaa - (V_aa00 + V_aa11)) * 2^[m] + V_aa11 * 2^[2m]

  BN_LFree (V_a0);
  BN_LFree (V_a1);
  BN_LFree (V_aa);
  BN_LFree (V_aa00);
  BN_LFree (V_aa11);
  BN_LFree (V_aaaa);
end;

{
  P = ABase^2
}
procedure BN_sqr (var P: T_BNX; const ABase: T_BNX);
var
  VSignP  : ShortInt;
  VCountB : IntX;
  Vm      : IntX;
begin
  if ABase.Sign = 0 then VSignP := 0 else VSignP := 1;
  VCountB := BN_DataCount (ABase);

  if (VSignP = 0) or (VCountB <= 0) then begin
    BN_AssignZero (P);
    Exit;
  end;

  if VCountB > Karatsuba_sqr_TCount then begin
    Vm := VCountB shr 1; // VCountB div 2

    BN_sqr_karatsuba_(P, ABase, Vm);
    P.Sign := VSignP;

    Exit;
  end;

  BN_sqr_base (P, ABase);
end;

{
  ABase = ABase^2
}
procedure BN_sqr_(var ABase: T_BNX);
var
  VSignP  : ShortInt;
  VCountB : IntX;
  Vm      : IntX;
  U       : T_BNX;
begin
  if ABase.Sign = 0 then VSignP := 0 else VSignP := 1;
  VCountB := BN_DataCount (ABase);

  if (VSignP = 0) or (VCountB <= 0) then begin
    BN_AssignZero (ABase);
    Exit;
  end;

  if VCountB > Karatsuba_sqr_TCount then begin
    Vm := VCountB shr 1; // VCountB div 2

    BN_sqr_karatsuba_(U, ABase, Vm);
    U.Sign := VSignP;
    BN_copy (ABase, U);
    BN_LFree (U);

    Exit;
  end;

  BN_sqr_base_(ABase);
end;

// [+] 1.2.1
{
  ABase = ABase^2
  imv - external intermediate variable
}
procedure BN_sqr_(var ABase: T_BNX; var imv: T_BNX);
var
  VSignP  : ShortInt;
  VCountB : IntX;
  Vm      : IntX;
begin
  if ABase.Sign = 0 then VSignP := 0 else VSignP := 1;
  VCountB := BN_DataCount (ABase);

  if (VSignP = 0) or (VCountB <= 0) then begin
    BN_AssignZero (ABase);
    Exit;
  end;

  if VCountB > Karatsuba_sqr_TCount then begin
    Vm := VCountB shr 1; // VCountB div 2

    BN_sqr_karatsuba_(imv, ABase, Vm);
    imv.Sign := VSignP;
    BN_copy (ABase, imv);

    Exit;
  end;

  BN_sqr_base_(ABase);
end;

// [+] 1.2.1
{
  returns TRUE if <s> is possible perfect square (s=u^2)
}
function BN_ppsqr (const s: T_BNX): Boolean;
var
  B : Byte;
  L : Byte;
  H : Byte;
begin
  Result := true;
  if BN_IsZero (s) then Exit;

  B := s.Data[0];
  L := B and $0F;

  if Odd (L) then begin
    // $_1, $_9
    if L = 1 then Exit;
    if L = 9 then Exit;
  end else begin
    // $00, $10, $40, $90
    // $04, $24, $44, $64, $84, $A4, $C4, $E4
    if B = $00 then Exit;
    if B = $10 then Exit;
    if B = $40 then Exit;
    if B = $90 then Exit;
    if L = 4 then begin
      H := (B and $F0) shr 4;
      if (H and 1) = 0 then Exit; // H - even
    end;
  end;

  Result := false;
end;

// [+] 1.2.1
{
  returns TRUE if <s> is possible perfect square (s=u^2)
}
function BN_ppsqr_(const s: T_BNX): Boolean;
var
  L : Byte;
begin
  Result := true;
  if BN_IsZero (s) then Exit;

  L := s.Data[0] and $0F;

  if Odd (L) then begin
    // $_1, $_9
    if L = 1 then Exit;
    if L = 9 then Exit;
  end else begin
    // $_0, $_4
    if L = 0 then Exit;
    if L = 4 then Exit;
  end;

  Result := false;
end;

// [+] 1.2.1
{
  returns TRUE if <s> is possible perfect even power (s=u^2m)
  m >= 1
}
function BN_ppp2m (const s: T_BNX; m: UIntX): Boolean;
var
  L : Byte;
begin
  if m = 1 then begin
    Result := BN_ppsqr (s);
    Exit;
  end;

  Result := true;
  if BN_IsZero (s) then Exit;

  L := s.Data[0] and $0F;

  if Odd (L) then begin
    if Odd (m) then begin
      // $_1, $_9
      if L = 1 then Exit;
      if L = 9 then Exit;
    end else begin
      // $_1
      if L = 1 then Exit;
    end;
  end else begin
    // $_0
    if L = 0 then Exit;
  end;

  Result := false;
end;

{
     
  Internal procedure

   n- 
   

  n-th Root
  Newton's method

  a > 1, n > 1
  aroot = a^(1/n)

  Result -  
  Result - number of iterations
}
function BN_root_(var aroot: T_BNX; const a: T_BNX; n: UIntX): Integer;
var
  V2E      : IntX;
  V2P      : UIntX;
  B_aroot  : Boolean;
  Vnm      : UIntX;
  vroot    : T_BNX;
  VX, VXS  : T_BNX;
  VXN, VNX : T_BNX;
  VCompare : Integer;
begin
  Result := 0;

  try
    Vnm := n - 1;

    //   
    // assign the first approximation

    // a^(1/n) = ((2^e)*(1+eps))^(1/n)
    //         = ((2^e)^(1/n))*((1+eps)^(1/n))
    //         = (2^(e/n))*m
    //         < (2^((e div n)+1))*2
    // a^(1/n) < 2^((e div n)+2)
    // x0+     = (e div n)+2
    // a^(1/n) > 2^(e div n)
    // x0-     = (e div n)
    // x0 = (x0+ + x0-) div 2
    //    = (e div n)+1

    V2E := BN_bsr (a);
    V2P := 1 + (UIntX(V2E) div n);
    BN_AssignBit (aroot, V2P);

    B_aroot := true;
    // aroot -  

    while true do begin
      //   
      Inc (Result);

      // x1 = ([n - 1] * x0 + a / x0^[n - 1]) / n
      if B_aroot then begin
        // aroot -  
        // vroot -  
        BN_pow (VXN, aroot, Vnm);        // VXN = x0^[n - 1]
        BN_div (VX, a, VXN);             // VX = a / x0^[n - 1]
        BN_mul_int (VNX, aroot, 1, Vnm); // VNX = [n - 1] * x0
        BN_add (VXS, VNX, VX);           // VXS = [n - 1] * x0 + a / x0^[n - 1]
        BN_div_int (vroot, VXS, 1, n);   // aroot = ([n - 1] * x0 + a / x0^[n - 1]) / n

        B_aroot := false;
        // vroot -  
        // aroot -  
        VCompare := BN_DataCompare (vroot, aroot);
      end else begin
        // vroot -  
        // aroot -  
        BN_pow (VXN, vroot, Vnm);        // VXN = x0^[n - 1]
        BN_div (VX, a, VXN);             // VX = a / x0^[n - 1]
        BN_mul_int (VNX, vroot, 1, Vnm); // VNX = [n - 1] * x0
        BN_add (VXS, VNX, VX);           // VXS = [n - 1] * x0 + a / x0^[n - 1]
        BN_div_int (aroot, VXS, 1, n);   // aroot = ([n - 1] * x0 + a / x0^[n - 1]) / n

        B_aroot := true;
        // aroot -  
        // vroot -  
        VCompare := BN_DataCompare (aroot, vroot);
      end;

      if VCompare = 0 then Exit;
      //  
      //    

      if VCompare > 0 then begin
        //   
        if Result > 1 then begin
          //  
          if B_aroot then begin
            // aroot -  
            // vroot -  
            BN_copy (aroot, vroot); // aroot = vroot
       // end else begin
            // vroot -  
            // aroot -  
            // BN_copy (vroot, aroot); // aroot = vroot
          end;
          Exit;
        end;
      end;
    end;
  finally
    BN_LFree (vroot);
    BN_LFree (VX);
    BN_LFree (VXS);
    BN_LFree (VXN);
    BN_LFree (VNX);
  end;
end;

{
   n- 
   

  n-th Root
  Newton's method

  a > 0
  aroot = a^(1/n)

  Result -  
  Result - number of iterations
}
function BN_root (var aroot: T_BNX; const a: T_BNX; n: UIntX): Integer;
begin
  Result := 0;

  if BN_LE_Zero (a) or (n = 0) then begin
    // a <= 0
    // aroot = 0
    BN_AssignZero (aroot);
    Exit;
  end;

  if BN_IsOne (a) then begin
    // a = 1
    // aroot = 1
    BN_AssignPOne (aroot);
    Exit;
  end;

  if n = 1 then begin
    // aroot = a
    BN_acopy (aroot, a);
    Exit;
  end;

  Result := BN_root_(aroot, a, n);
end;

{
   n- 
   

  n-th Root
  Newton's method

  a > 0
  aroot = a^(1/n)
  arem = a - aroot^n

  Result -  
  Result - number of iterations
}
function BN_root_rem (var aroot, arem: T_BNX; const a: T_BNX; n: UIntX): Integer;
var
  V : T_BNX;
begin
  Result := 0;

  if BN_LE_Zero (a) or (n = 0) then begin
    // a <= 0
    // aroot = 0
    // arem = 0
    BN_AssignZero (aroot);
    BN_AssignZero (arem);
    Exit;
  end;

  if BN_IsOne (a) then begin
    // a = 1
    // aroot = 1
    // arem = 0
    BN_AssignPOne (aroot);
    BN_AssignZero (arem);
    Exit;
  end;

  if n = 1 then begin
    // aroot = a
    // arem = 0
    BN_acopy (aroot, a);
    BN_AssignZero (arem);
    Exit;
  end;

  Result := BN_root_(aroot, a, n);
  BN_pow (V, aroot, n); // V = aroot^n
  BN_sub (arem, a, V);  // arem = a - aroot^n

  BN_LFree (V);
end;

{
     
  Internal procedure

   
   
    

  Square Root
  Newton's method

  a > 1
  aroot = a^(1/2)
  vroot = aroot

  Result -  
  Result - number of iterations
}
function BN_sqrt_(var aroot, vroot: T_BNX; const a: T_BNX): Integer;
var
  V2E      : IntX;
  V2P      : UIntX;
  B_aroot  : Boolean;
  VX       : T_BNX;
  VCompare : Integer;
begin
  Result := 0;

  try
    //   
    // assign the first approximation

    // a^(1/n) = ((2^e)*(1+eps))^(1/n)
    //         = ((2^e)^(1/n))*((1+eps)^(1/n))
    //         = (2^(e/n))*m
    //         < (2^((e div n)+1))*2
    // a^(1/n) < 2^((e div n)+2)
    // x0+     = (e div n)+2
    // a^(1/n) > 2^(e div n)
    // x0-     = (e div n)
    // x0 = (x0+ + x0-) div 2
    //    = (e div n)+1

    V2E := BN_bsr (a);
    V2P := 1 + (UIntX(V2E) shr 1); // ... div 2
    BN_AssignBit (aroot, V2P);

    B_aroot := true;
    // aroot -  

    while true do begin
      //   
      Inc (Result);

      // x1 = (x0 + a / x0) / 2
      if B_aroot then begin
        // aroot -  
        // vroot -  
        BN_div (VX, a, aroot);     // VX = a / x0
        BN_add (vroot, aroot, VX); // aroot = x0 + a / x0
        BN_shr_1 (vroot);          // aroot = (x0 + a / x0) / 2

        B_aroot := false;
        // vroot -  
        // aroot -  
        VCompare := BN_DataCompare (vroot, aroot);
      end else begin
        // vroot -  
        // aroot -  
        BN_div (VX, a, vroot);     // VX = a / x0
        BN_add (aroot, vroot, VX); // aroot = x0 + a / x0
        BN_shr_1 (aroot);          // aroot = (x0 + a / x0) / 2

        B_aroot := true;
        // aroot -  
        // vroot -  
        VCompare := BN_DataCompare (aroot, vroot);
      end;

      if VCompare = 0 then Exit;
      //  
      //    

      if VCompare > 0 then begin
        //   
        if Result > 1 then begin
          //  
          if B_aroot then begin
            // aroot -  
            // vroot -  
            BN_copy (aroot, vroot); // aroot = vroot
       // end else begin
            // vroot -  
            // aroot -  
            // BN_copy (vroot, aroot); // aroot = vroot
          end;
          Exit;
        end;
      end;
    end;
  finally
    BN_LFree (VX);
  end;
end;

{
   
   
    

  Square Root
  Newton's method

  a > 0
  aroot = a^(1/2)

  Result -  
  Result - number of iterations
}
function BN_sqrt (var aroot: T_BNX; const a: T_BNX): Integer;
var
  vroot : T_BNX;
begin
  Result := 0;

  if BN_LE_Zero (a) then begin
    // a <= 0
    // aroot = 0
    BN_AssignZero (aroot);
    Exit;
  end;

  if BN_IsOne (a) then begin
    // a = 1
    // aroot = 1
    BN_AssignPOne (aroot);
    Exit;
  end;

  Result := BN_sqrt_(aroot, vroot, a);
  BN_LFree (vroot);
end;

{
   
   
    

  Square Root
  Newton's method

  a > 0
  aroot = a^(1/2)
  arem = a - aroot^2

  Result -  
  Result - number of iterations
}
function BN_sqrt_rem (var aroot, arem: T_BNX; const a: T_BNX): Integer;
var
  V     : T_BNX;
  vroot : T_BNX;
begin
  Result := 0;

  if BN_LE_Zero (a) then begin
    // a <= 0
    // aroot = 0
    // arem = 0
    BN_AssignZero (aroot);
    BN_AssignZero (arem);
    Exit;
  end;

  if BN_IsOne (a) then begin
    // a = 1
    // aroot = 1
    // arem = 0
    BN_AssignPOne (aroot);
    BN_AssignZero (arem);
    Exit;
  end;

  Result := BN_sqrt_(aroot, vroot, a);
  BN_mul (V, aroot, vroot); // V = aroot^2
  BN_sub (arem, a, V);      // arem = a - aroot^2

  BN_LFree (V);
  BN_LFree (vroot);
end;

{
   
  Square Root

  The modified Dijkstras square root algorithm
  Prefer truncation to rounding

  |a| = aroot^2 + arem
  aroot = |a|^(1/2)

  Result -  
  Result - number of iterations
}
function BN_dsqrt (var aroot: T_BNX; const a: T_BNX): Integer;
var
  B_mask : IntX;
  V_mask : T_BNX;
  V_rema : T_BNX;
  V_roma : T_BNX;
begin
  Result := 0;

  BN_AssignZero (aroot);
  if BN_IsZero (a) then Exit;

  B_mask := BN_bsr (a);
  if Odd (B_mask) then Dec (B_mask);
  BN_AssignBit (V_mask, B_mask);
  // mask = 1 << (BITS-2)

  BN_copy (V_rema, a);

  while BN_NE_Zero (V_mask) do begin
    Inc (Result);
    BN_add (V_roma, aroot, V_mask);
    if BN_DataCompare (V_roma, V_rema) <= 0 then begin
      BN_sub_(V_rema, V_roma);
      BN_shl_1 (V_roma, V_mask);
      BN_add_(aroot, V_roma);
    end;
    BN_shr_1 (aroot);
    BN_shr_2 (V_mask);
  end;

  BN_LFree (V_mask);
  BN_LFree (V_rema);
  BN_LFree (V_roma);
end;

{
   
  Square Root

  The modified Dijkstras square root algorithm
  Prefer rounding to truncation

  |a| = aroot^2 + arem
  aroot = |a|^(1/2)

  Result -  
  Result - number of iterations
}
function BN_rnd_dsqrt (var aroot: T_BNX; const a: T_BNX): Integer;
var
  B_mask : IntX;
  V_mask : T_BNX;
  V_rema : T_BNX;
  V_roma : T_BNX;
begin
  Result := 0;

  BN_AssignZero (aroot);
  if BN_IsZero (a) then Exit;

  B_mask := BN_bsr (a);
  if Odd (B_mask) then Dec (B_mask);
  BN_AssignBit (V_mask, B_mask);
  // mask = 1 << (BITS-2)

  BN_copy (V_rema, a);

  while BN_NE_Zero (V_mask) do begin
    Inc (Result);
    BN_add (V_roma, aroot, V_mask);
    if BN_DataCompare (V_roma, V_rema) <= 0 then begin
      BN_sub_(V_rema, V_roma);
      BN_shl_1 (V_roma, V_mask);
      BN_add_(aroot, V_roma);
    end;
    BN_shr_1 (aroot);
    BN_shr_2 (V_mask);
  end;

  // Prefer rounding to truncation
  if BN_DataCompare (V_rema, aroot) > 0 then begin
    Inc (Result);
    BN_inc_(aroot); // root = root+1
  end;

  BN_LFree (V_mask);
  BN_LFree (V_rema);
  BN_LFree (V_roma);
end;

{
   
  Square Root

  The modified Dijkstras square root algorithm
  Prefer truncation to rounding

  |a| = aroot^2 + arem
  aroot = |a|^(1/2)
  arem = |a| - aroot^2

  Result -  
  Result - number of iterations
}
function BN_dsqrt_rem (var aroot, arem: T_BNX; const a: T_BNX): Integer;
var
  B_mask : IntX;
  V_mask : T_BNX;
  V_roma : T_BNX;
begin
  Result := 0;

  BN_AssignZero (aroot);
  BN_copy (arem, a);

  if BN_IsZero (a) then Exit;

  B_mask := BN_bsr (a);
  if Odd (B_mask) then Dec (B_mask);
  BN_AssignBit (V_mask, B_mask);
  // mask = 1 << (BITS-2)

  while BN_NE_Zero (V_mask) do begin
    Inc (Result);
    BN_add (V_roma, aroot, V_mask);
    if BN_DataCompare (V_roma, arem) <= 0 then begin
      BN_sub_(arem, V_roma);
      BN_shl_1 (V_roma, V_mask);
      BN_add_(aroot, V_roma);
    end;
    BN_shr_1 (aroot);
    BN_shr_2 (V_mask);
  end;

  BN_LFree (V_mask);
  BN_LFree (V_roma);
end;

{
   
  Square Root

  The modified Dijkstras square root algorithm
  Prefer rounding to truncation

  |a| = aroot^2 + arem
  aroot = |a|^(1/2)
  arem = |a| - aroot^2

  Result -  
  Result - number of iterations
}
function BN_rnd_dsqrt_rem (var aroot, arem: T_BNX; const a: T_BNX): Integer;
var
  B_mask : IntX;
  V_mask : T_BNX;
  V_rema : T_BNX;
  V_roma : T_BNX;
begin
  Result := 0;

  BN_AssignZero (aroot);
  BN_copy (arem, a);

  if BN_IsZero (a) then Exit;

  B_mask := BN_bsr (a);
  if Odd (B_mask) then Dec (B_mask);
  BN_AssignBit (V_mask, B_mask);
  // mask = 1 << (BITS-2)

  while BN_NE_Zero (V_mask) do begin
    Inc (Result);
    BN_add (V_roma, aroot, V_mask);
    if BN_DataCompare (V_roma, arem) <= 0 then begin
      BN_sub_(arem, V_roma);
      BN_shl_1 (V_roma, V_mask);
      BN_add_(aroot, V_roma);
    end;
    BN_shr_1 (aroot);
    BN_shr_2 (V_mask);
  end;

  // Prefer rounding to truncation
  if BN_DataCompare (arem, aroot) > 0 then begin
    Inc (Result);
    // a = root^2 + rem
    //   = root'^2 + rem'
    //   = (root+1)^2 + rem'
    //   = root^2 + (2*root + 1 + rem')
    // rem'  = rem - 2*root - 1
    //       = (rem-1) - 2*root
    // root' = root+1
    BN_dec (V_rema, arem);          // rema = rem-1
    BN_shl_1 (V_roma, aroot);       // roma = 2*root
    BN_sub (arem, V_rema, V_roma);  // rem = (rem-1) - 2*root
    BN_inc_(aroot);                 // root = root+1
  end;

  BN_LFree (V_mask);
  BN_LFree (V_rema);
  BN_LFree (V_roma);
end;

// [+] 1.2.1
{
  a = a' * 2^n
  Result = n

  n =    
  n = number of low order zero bits
}
function BN_m2n (const a: T_BNX): UIntX;
{$IFDEF CPUX64}
// rcx | @a |
asm
       .NOFRAME

        xor     rax, rax        // Result = 0

        mov     rdx, [rcx].T_BNX.Count
        cmp     rdx, 0          // rdx = a.Count
        jle     @R

        mov     r8, [rcx].T_BNX.PLData
  @A:                           // for I := 0 to HIndex ...
        mov     r9, [r8]        // r9 = a.Data[I]
        test    r9, r9
        jz      @C
  @B:
//      xor     r10, r10
        bsf     r10, r9
        add     rax, r10
        jmp     @R
  @C:
        add     rax, CXBits
  @D:
        add     r8, CXBytes
        dec     rdx
        jnz     @A

  @R:
end;
{$ELSE}
// eax | @a | ecx
asm
        push    ebx
        push    esi
        push    edi

        mov     ecx, eax        // ecx = @a
        xor     eax, eax        // Result = 0

        mov     edx, [ecx].T_BNX.Count
        cmp     edx, 0          // edx = a.Count
        jle     @R

        mov     esi, [ecx].T_BNX.PLData
  @A:                           // for I := 0 to HIndex ...
        mov     edi, [esi]      // edi = a.Data[I]
        test    edi, edi
        jz      @C
  @B:
//      xor     ebx, ebx
        bsf     ebx, edi
        add     eax, ebx
        jmp     @R
  @C:
        add     eax, CXBits
  @D:
        add     esi, CXBytes
        dec     edx
        jnz     @A

  @R:
        pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

// Pure pascal
function BN_m2n_PP (const a: T_BNX): UIntX;
var
  I : LoopInt;
  V : UIntX;
  O : Boolean;
begin
  Result := 0;

  for I := 0 to a.Count - 1 do begin
    V := a.Data[I];
    if V = 0 then begin
      Inc (Result, CXBits);
    end else begin
      O := Odd (V);
      if O then Break;
      repeat
        Inc (Result);
        V := V shr 1;
        O := Odd (V);
      until O;
      if O then Break;
    end;
  end;
end;

{
  a = a' * 2^n
  Result = n
  sha = a shr n = a'

  n =    
  n = number of low order zero bits
}
function BN_m2n_(const a: T_BNX; var sha: T_BNX): UIntX;
begin
  Result := BN_m2n (a);
  if Result = 0 then begin
    BN_copy (sha, a);
    Exit;
  end;

  BN_pshr (sha, a, Result);
end;

// [+] 1.2.1
{
  a = a' * 2^n
  Result = n
  a = a shr n = a'

  n =    
  n = number of low order zero bits
}
function BN_m2n_(var a: T_BNX): UIntX;
begin
  Result := BN_m2n (a);
  if Result = 0 then Exit;

  BN_pshr (a, Result);
end;

// [+] 1.2.1
{
  a = a' * 2^n
  Result = n

  n =    
  n = number of low order zero bits
}
function BN_m2n_int (a: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | a |
asm
       .NOFRAME

        mov     rax, CXBits
        bsf     rax, rcx        // Result = n
end;
{$ELSE}
// eax | a |
asm
        mov     ecx, CXBits
        bsf     ecx, eax        // ecx = n
        mov     eax, ecx        // Result = n
end;
{$ENDIF}

// Pure pascal
function BN_m2n_int_PP (a: UIntX): UIntX;
begin
  Result := 0;
  if a = 0 then Exit;

  while not Odd (a) do begin
    Inc (Result);
    a := a shr 1;
  end;
end;

// [+] 1.2.1
{
  a = a' * 2^n
  Result = n
  sha = a shr n = a'

  n =    
  n = number of low order zero bits
}
function BN_m2n_int_(a: UIntX; var sha: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | a    |
// rdx | @sha |
asm
       .NOFRAME

        mov     rax, CXBits
        bsf     rax, rcx        // Result = n
        mov     r8, rcx         // r8 = a
        mov     rcx, rax        // rcx = n
        shr     r8, cl          // r8 = a shr n
        mov     [rdx], r8       // sha = a shr n
end;
{$ELSE}
// eax | a    | ebx
// edx | @sha |
asm
        push    ebx

        mov     ebx, eax        // ebx = a
        mov     ecx, CXBits
        bsf     ecx, eax        // ecx = n
        mov     eax, ecx        // Result = n
        shr     ebx, cl         // ebx = a shr n
        mov     [edx], ebx      // sha = a shr n

        pop     ebx
end;
{$ENDIF}

// Pure pascal
function BN_m2n_int_PP_(a: UIntX; var sha: UIntX): UIntX;
begin
  Result := 0;
  sha := a;
  if sha = 0 then Exit;

  while not Odd (sha) do begin
    Inc (Result);
    sha := sha shr 1;
  end;
end;

// [+] 1.2.1
{
   
  Argument test

   TRUE,   
  Returns TRUE if the result is defined

  gcd( m, m) = m
  gcd( 0, n) = n
  gcd( 1, n) = 1
  gcd( m, 0) = m
  gcd( m, 1) = 1

  gcd(-m, n) = gcd(m, n)
}
function BN_gcd_arg (var R: T_BNX; const m, n: T_BNX;
  var ACompare: Integer): Boolean;
var
  VUInt : UIntX;
begin
  Result := true;
  ACompare := BN_DataCompare (m, n);

  if ACompare = 0 then begin    // m = n
    BN_acopy (R, m);            // gcd (m, m) = |m|
    Exit;
  end;

  if BN_IsInt (m, VUInt) then begin
    if VUInt = 0 then begin     // m = 0
      BN_acopy (R, n);          // gcd (0, n) = |n|
      Exit;
    end;
    if VUInt = 1 then begin     // m = 1
      BN_AssignPOne (R);        // gcd (1, n) = 1
      Exit;
    end;
  end;

  if BN_IsInt (n, VUInt) then begin
    if VUInt = 0 then begin     // n = 0
      BN_acopy (R, m);          // gcd (m, 0) = |m|
      Exit;
    end;
    if VUInt = 1 then begin     // n = 1
      BN_AssignPOne (R);        // gcd (m, 1) = 1
      Exit;
    end;
  end;

  Result := false;
end;

{
    
  Greatest common divisor
  Highest common factor

   
  Euclidean algorithm

  (m,n)
  (m,n)
  gcd(m,n)
  hcf(m,n)
}
procedure BN_gcd (var R: T_BNX; const m, n: T_BNX);
var
  V_nn     : array [0..2] of T_BNX;
  I_m      : Integer;
  I_n      : Integer;
  I_r      : Integer;
  I_U      : Integer;
  VCompare : Integer;
begin
  if BN_gcd_arg (R, m, n, VCompare) then Exit;

  try
    I_m := 0;
    I_n := 1;
    I_r := 2;

    if VCompare > 0 then begin
      // m > n
      BN_acopy (V_nn[I_m], m);
      BN_acopy (V_nn[I_n], n);
    end else begin
      // m < n
      BN_acopy (V_nn[I_m], n);
      BN_acopy (V_nn[I_n], m);
    end;

    while true do begin
      BN_mod (V_nn[I_r], V_nn[I_m], V_nn[I_n]);
      // r = m mod n

      if BN_IsZero (V_nn[I_r]) then begin
        BN_acopy (R, V_nn[I_n]);
        Exit;
      end;

      I_U := I_m;
      I_m := I_n; // m = n
      I_n := I_r; // n = r
      I_r := I_U; // r = m
    end;
  finally
    BN_LFree (V_nn[0]);
    BN_LFree (V_nn[1]);
    BN_LFree (V_nn[2]);
  end;
end;

// [+] 1.2.1
{
    
  Greatest common divisor
  Highest common factor

   
  Binary algorithm

  (m,n)
  (m,n)
  gcd(m,n)
  hcf(m,n)
}
procedure BN_gcd_bin (var R: T_BNX; const m, n: T_BNX);
var
  V_m      : T_BNX;
  V_n      : T_BNX;
  V_u      : T_BNX;
  VCompare : Integer;
begin
  if BN_gcd_arg (R, m, n, VCompare) then Exit;

  BN_acopy (V_m, m);
  BN_acopy (V_n, n);
  BN_AssignPOne (V_u);

  while BN_IsNotZero (V_m) and BN_IsNotZero (V_n) do begin
    while BN_IsEven (V_m) and BN_IsEven (V_n) do begin
      BN_shr_1 (V_m);
      BN_shr_1 (V_n);
      BN_shl_1 (V_u);
      // if BN_IsZero (V_m) or BN_IsZero (V_n) then Break;
    end;

    BN_m2n_(V_m);
    BN_m2n_(V_n);
    (*
    while BN_IsEven (V_m) do BN_shr_1 (V_m);
    while BN_IsEven (V_n) do BN_shr_1 (V_n);
    // *)

    if BN_DataCompare (V_m, V_n) >= 0 then begin
      BN_sub_(V_m, V_n);
    end else begin
      BN_sub_(V_n, V_m);
    end;
  end;

  if V_n.Count > 0 then begin
    BN_mul (R, V_n, V_u);
  end else begin
    BN_mul (R, V_m, V_u);
  end;
  BN_abs (R);

  BN_LFree (V_m);
  BN_LFree (V_n);
  BN_LFree (V_u);
end;

// [+] 1.2.1
{
    
  Greatest common divisor
  Highest common factor

   
  Euclidean algorithm

  (m,n)
  (m,n)
  gcd(m,n)
  hcf(m,n)
}
procedure BN_gcd_mod (var R: T_BNX; const m, n: T_BNX);
var
  V_m      : T_BNX;
  V_n      : T_BNX;
  V_u      : T_BNX;
  VCompare : Integer;
begin
  if BN_gcd_arg (R, m, n, VCompare) then Exit;

  BN_acopy (V_m, m);
  BN_acopy (V_n, n);

  while BN_IsNotZero (V_m) and BN_IsNotZero (V_n) do begin
    if BN_DataCompare (V_m, V_n) > 0 then begin
      BN_mod (V_u, V_m, V_n);
      BN_copy (V_m, V_u);
    end else begin
      BN_mod (V_u, V_n, V_m);
      BN_copy (V_n, V_u);
    end;
  end;

  BN_add (R, V_m, V_n);

  BN_LFree (V_m);
  BN_LFree (V_n);
  BN_LFree (V_u);
end;

// [+] 1.2.1
{
    
  Greatest common divisor
  Highest common factor

   
  Euclidean algorithm

  (m,n)
  (m,n)
  gcd(m,n)
  hcf(m,n)

  !!!  
  !!! Very slow
}
procedure BN_gcd_sub (var R: T_BNX; const m, n: T_BNX);
var
  V_m      : T_BNX;
  V_n      : T_BNX;
  VCompare : Integer;
begin
  if BN_gcd_arg (R, m, n, VCompare) then Exit;

  BN_acopy (V_m, m);
  BN_acopy (V_n, n);

  while VCompare <> 0 do begin
    if VCompare > 0 then begin
      BN_sub_(V_m, V_n);
    end else begin
      BN_sub_(V_n, V_m);
    end;
    VCompare := BN_DataCompare (V_m, V_n);
  end;

  BN_acopy (R, V_m);

  BN_LFree (V_m);
  BN_LFree (V_n);
end;

// [+] 1.2.1
{
   TRUE,  a  b  
  Returns TRUE, if a and b are coprime numbers
}
function BN_coprime (const a, b: T_BNX): Boolean;
var
  V : T_BNX;
begin
  BN_gcd (V, a, b);
  Result := BN_IsOne (V);
  BN_LFree (V);
end;

{
    
  Least common multiple

  (m,n)
  [m,n]
  lcm(m,n)

  lcm(m,n) = |m*n|/gcd(m,n)
}
procedure BN_lcm (var R: T_BNX; const m, n: T_BNX);
var
  V_m : T_BNX;
  V_d : T_BNX;
begin
  BN_mul (V_m, m, n);
  BN_abs (V_m); // V_m = |m*n|
  BN_gcd (V_d, m, n);
  BN_div (R, V_m, V_d);

  BN_LFree (V_m);
  BN_LFree (V_d);
end;

// [+] 1.2.1
{
    
  Least common multiple

  (m,n)
  [m,n]
  lcm(m,n)

  lcm(m,n) = |m*n|/gcd(m,n)
}
procedure BN_lcm_bin (var R: T_BNX; const m, n: T_BNX);
var
  V_m : T_BNX;
  V_d : T_BNX;
begin
  BN_mul (V_m, m, n);
  BN_abs (V_m); // V_m = |m*n|
  BN_gcd_bin (V_d, m, n);
  BN_div (R, V_m, V_d);

  BN_LFree (V_m);
  BN_LFree (V_d);
end;

// [+] 1.2.1
{
  Bezout's identity
  gcd(a,b) = a*x + b*y

  input  : a, b
  output : g = gcd(a,b), x, y
}
procedure BN_gcd_bezout (var g, x, y: T_BNX; const a, b: T_BNX);
var
  V_nn     : array [0..8] of T_BNX;
  I_U      : Integer;
  I_a      : Integer;
  I_b      : Integer;
  I_r      : Integer;
  I_x2     : Integer;
  I_x1     : Integer;
  I_x0     : Integer;
  I_y2     : Integer;
  I_y1     : Integer;
  I_y0     : Integer;

  V_q      : T_BNX;
  V_u      : T_BNX;

  VCompare : Integer;
begin
  VCompare := BN_DataCompare (a, b);

  if VCompare = 0 then begin // a = b
    BN_acopy (g, a);         // gcd (a, b) = |a|
    BN_AssignPOne (x);       // x = 1
    BN_AssignZero (y);       // y = 0
    Exit;
  end;

  if BN_IsZero (a) then begin // a = 0
    BN_acopy (g, b);          // gcd (0, b) = |b|
    BN_AssignZero (x);        // x = 0
    BN_AssignPOne (y);        // y = 1
    Exit;
  end;

  if BN_IsZero (b) then begin // b = 0
    BN_acopy (g, a);          // gcd (a, 0) = |a|
    BN_AssignPOne (x);        // x = 1
    BN_AssignZero (y);        // y = 0
    Exit;
  end;

  I_a  := 0;
  I_b  := 1;
  I_r  := 2;
  I_x2 := 3;
  I_x1 := 4;
  I_x0 := 5;
  I_y2 := 6;
  I_y1 := 7;
  I_y0 := 8;

  if VCompare > 0 then begin
    // a > b
    BN_acopy (V_nn[I_a], a);
    BN_acopy (V_nn[I_b], b);
  end else begin
    // a < b
    BN_acopy (V_nn[I_a], b);
    BN_acopy (V_nn[I_b], a);
  end;

  BN_AssignPOne (V_nn[I_x2]); // x2 = 1
  BN_AssignZero (V_nn[I_x1]); // x1 = 0
  BN_AssignZero (V_nn[I_y2]); // y2 = 0
  BN_AssignPOne (V_nn[I_y1]); // y1 = 1

  while BN_IsNotZero (V_nn[I_b]) do begin
    BN_div_mod (V_q, V_nn[I_r], V_nn[I_a], V_nn[I_b]);
    // q = a div b
    // r = a mod b

    BN_mul (V_u, V_q, V_nn[I_x1]);
    BN_sub (V_nn[I_x0], V_nn[I_x2], V_u);
    // u  = q*x1
    // x0 = x2 - q*x1

    BN_mul (V_u, V_q, V_nn[I_y1]);
    BN_sub (V_nn[I_y0], V_nn[I_y2], V_u);
    // u  = q*y1
    // y0 = y2 - q*y1

    I_U  := I_a;
    I_a  := I_b; // a = b
    I_b  := I_r; // b = r
    I_r  := I_U; // r = a

    I_U  := I_x2;
    I_x2 := I_x1; // x2 = x1
    I_x1 := I_x0; // x1 = x0
    I_x0 := I_U;  // x0 = x2

    I_U  := I_y2;
    I_y2 := I_y1; // y2 = y1
    I_y1 := I_y0; // y1 = y0
    I_y0 := I_U;  // y0 = y2
  end;

  BN_copy (g, V_nn[I_a]);    // g = a
  if VCompare > 0 then begin
    BN_copy (x, V_nn[I_x2]); // x = x2
    BN_copy (y, V_nn[I_y2]); // y = y2
  end else begin
    BN_copy (x, V_nn[I_y2]); // x = y2
    BN_copy (y, V_nn[I_x2]); // y = x2
  end;

  BN_LFree (V_nn[0]);
  BN_LFree (V_nn[1]);
  BN_LFree (V_nn[2]);
  BN_LFree (V_nn[3]);
  BN_LFree (V_nn[4]);
  BN_LFree (V_nn[5]);
  BN_LFree (V_nn[6]);
  BN_LFree (V_nn[7]);
  BN_LFree (V_nn[8]);
  BN_LFree (V_q);
  BN_LFree (V_u);
end;

// [+] 1.2.1
{
  Bezout's identity
  gcd(a,b) = a*x + b*y

  input  : a, b
  output : x, y
}
procedure BN_bezout (var x, y: T_BNX; const a, b: T_BNX);
var
  V_nn     : array [0..8] of T_BNX;
  I_U      : Integer;
  I_a      : Integer;
  I_b      : Integer;
  I_r      : Integer;
  I_x2     : Integer;
  I_x1     : Integer;
  I_x0     : Integer;
  I_y2     : Integer;
  I_y1     : Integer;
  I_y0     : Integer;

  V_q      : T_BNX;
  V_u      : T_BNX;

  VCompare : Integer;
begin
  VCompare := BN_DataCompare (a, b);

  if VCompare = 0 then begin // a = b
    BN_AssignPOne (x);       // x = 1
    BN_AssignZero (y);       // y = 0
    Exit;
  end;

  if BN_IsZero (a) then begin // a = 0
    BN_AssignZero (x);        // x = 0
    BN_AssignPOne (y);        // y = 1
    Exit;
  end;

  if BN_IsZero (b) then begin // b = 0
    BN_AssignPOne (x);        // x = 1
    BN_AssignZero (y);        // y = 0
    Exit;
  end;

  I_a  := 0;
  I_b  := 1;
  I_r  := 2;
  I_x2 := 3;
  I_x1 := 4;
  I_x0 := 5;
  I_y2 := 6;
  I_y1 := 7;
  I_y0 := 8;

  if VCompare > 0 then begin
    // a > b
    BN_acopy (V_nn[I_a], a);
    BN_acopy (V_nn[I_b], b);
  end else begin
    // a < b
    BN_acopy (V_nn[I_a], b);
    BN_acopy (V_nn[I_b], a);
  end;

  BN_AssignPOne (V_nn[I_x2]); // x2 = 1
  BN_AssignZero (V_nn[I_x1]); // x1 = 0
  BN_AssignZero (V_nn[I_y2]); // y2 = 0
  BN_AssignPOne (V_nn[I_y1]); // y1 = 1

  while BN_IsNotZero (V_nn[I_b]) do begin
    BN_div_mod (V_q, V_nn[I_r], V_nn[I_a], V_nn[I_b]);
    // q = a div b
    // r = a mod b

    BN_mul (V_u, V_q, V_nn[I_x1]);
    BN_sub (V_nn[I_x0], V_nn[I_x2], V_u);
    // u  = q*x1
    // x0 = x2 - q*x1

    BN_mul (V_u, V_q, V_nn[I_y1]);
    BN_sub (V_nn[I_y0], V_nn[I_y2], V_u);
    // u  = q*y1
    // y0 = y2 - q*y1

    I_U  := I_a;
    I_a  := I_b; // a = b
    I_b  := I_r; // b = r
    I_r  := I_U; // r = a

    I_U  := I_x2;
    I_x2 := I_x1; // x2 = x1
    I_x1 := I_x0; // x1 = x0
    I_x0 := I_U;  // x0 = x2

    I_U  := I_y2;
    I_y2 := I_y1; // y2 = y1
    I_y1 := I_y0; // y1 = y0
    I_y0 := I_U;  // y0 = y2
  end;

  if VCompare > 0 then begin
    // a > b
    BN_copy (x, V_nn[I_x2]); // x = x2
    BN_copy (y, V_nn[I_y2]); // y = y2
  end else begin
    // a < b
    BN_copy (x, V_nn[I_y2]); // x = y2
    BN_copy (y, V_nn[I_x2]); // y = x2
  end;

  BN_LFree (V_nn[0]);
  BN_LFree (V_nn[1]);
  BN_LFree (V_nn[2]);
  BN_LFree (V_nn[3]);
  BN_LFree (V_nn[4]);
  BN_LFree (V_nn[5]);
  BN_LFree (V_nn[6]);
  BN_LFree (V_nn[7]);
  BN_LFree (V_nn[8]);
  BN_LFree (V_q);
  BN_LFree (V_u);
end;

// [+] 1.2.1
{
  Jacobi symbol
  J(a,n) = -1, 0, +1
           -2 (error)
}
function BN_jacobi (const a, n: T_BNX): Integer;
var
  V_e   : UIntX;
  V_a   : T_BNX;
  V_n   : T_BNX;
  V_u   : T_BNX;

  n_mod : UIntX;
  a_mod : UIntX;
begin
  Result := -2; // error

  if BN_IsEven (n) then Exit;  // n is even
  if BN_LT_Zero (n) then Exit; // n < 0

  if BN_IsOne (a) then begin   // a = 1
    Result := 1;               // J(1,n) = 1
    Exit;
  end;

  if BN_IsOne (n) then begin   // n = 1
    Result := 1;               // J(a,1) = 1
    Exit;
  end;

  BN_gcd (V_u, a, n);
  if not BN_IsOne (V_u) then begin // a and n are not coprime numbers
    Result := 0;                   // J(a,n) = 0
    BN_LFree (V_u);
    Exit;
  end;

  Result := 1;

  BN_copy (V_a, a);
  BN_copy (V_n, n);

  if BN_LT_Zero (V_a) then begin // a < 0
    BN_chs (V_a);
    n_mod := BN_mod_2n (V_n, 2); // n mod 4 = n mod 2^2
    if n_mod = 3 then Result := -Result;
  end;

  while true do begin
    // a = (2^e)*a'
    V_e := BN_m2n_(V_a);

    if Odd (V_e) then begin
      n_mod := BN_mod_2n (V_n, 3); // n mod 8 = n mod 2^3
      if (n_mod = 3) or (n_mod = 5) then Result := -Result;
    end;

    n_mod := BN_mod_2n (V_n, 2); // n mod 4 = n mod 2^2
    a_mod := BN_mod_2n (V_a, 2); // a mod 4 = a mod 2^2
    if (n_mod = 3) and (a_mod = 3) then Result := -Result;

    BN_copy (V_u, V_a);
    BN_mod (V_a, V_n, V_u);
    if BN_IsZero (V_a) then Break;

    BN_copy (V_n, V_u);
  end;

  BN_LFree (V_a);
  BN_LFree (V_n);
  BN_LFree (V_u);
end;

// [+] 1.2.1
{
  (a_inv*a) mod b = 1 = gcd(a,b)

  input  : a, b
  output : a_inv

  Returns TRUE if gcd(a,b)=1
}
function BN_inv_mod (var a_inv: T_BNX; const a, b: T_BNX): Boolean;
var
  V_nn     : array [0..8] of T_BNX;
  I_U      : Integer;
  I_a      : Integer;
  I_b      : Integer;
  I_r      : Integer;
  I_x2     : Integer;
  I_x1     : Integer;
  I_x0     : Integer;
  I_y2     : Integer;
  I_y1     : Integer;
  I_y0     : Integer;

  V_q      : T_BNX;
  V_u      : T_BNX;

  VCompare : Integer;
begin
  Result := false;
  BN_AssignZero (a_inv); // a_inv = 0

  VCompare := BN_DataCompare (a, b);

  if VCompare = 0 then Exit;  // gcd (a, b) = a
  if BN_IsZero (a) then Exit; // gcd (0, b) = b
  if BN_IsZero (b) then Exit; // gcd (a, 0) = a

  I_a  := 0;
  I_b  := 1;
  I_r  := 2;
  I_x2 := 3;
  I_x1 := 4;
  I_x0 := 5;
  I_y2 := 6;
  I_y1 := 7;
  I_y0 := 8;

  if VCompare > 0 then begin
    // a > b
    BN_acopy (V_nn[I_a], a);
    BN_acopy (V_nn[I_b], b);
  end else begin
    // a < b
    BN_acopy (V_nn[I_a], b);
    BN_acopy (V_nn[I_b], a);
  end;

  BN_AssignPOne (V_nn[I_x2]); // x2 = 1
  BN_AssignZero (V_nn[I_x1]); // x1 = 0
  BN_AssignZero (V_nn[I_y2]); // y2 = 0
  BN_AssignPOne (V_nn[I_y1]); // y1 = 1

  while BN_IsNotZero (V_nn[I_b]) do begin
    BN_div_mod (V_q, V_nn[I_r], V_nn[I_a], V_nn[I_b]);
    // q = a div b
    // r = a mod b

    BN_mul (V_u, V_q, V_nn[I_x1]);
    BN_sub (V_nn[I_x0], V_nn[I_x2], V_u);
    // u  = q*x1
    // x0 = x2 - q*x1

    BN_mul (V_u, V_q, V_nn[I_y1]);
    BN_sub (V_nn[I_y0], V_nn[I_y2], V_u);
    // u  = q*y1
    // y0 = y2 - q*y1

    I_U  := I_a;
    I_a  := I_b; // a = b
    I_b  := I_r; // b = r
    I_r  := I_U; // r = a

    I_U  := I_x2;
    I_x2 := I_x1; // x2 = x1
    I_x1 := I_x0; // x1 = x0
    I_x0 := I_U;  // x0 = x2

    I_U  := I_y2;
    I_y2 := I_y1; // y2 = y1
    I_y1 := I_y0; // y1 = y0
    I_y0 := I_U;  // y0 = y2
  end;

  if VCompare > 0 then begin
    // a > b
    BN_copy (a_inv, V_nn[I_x2]); // a_inv = x2
  end else begin
    // a < b
    BN_copy (a_inv, V_nn[I_y2]); // a_inv = y2
  end;

  if BN_LT_Zero (a_inv) then begin
    // if a_inv < 0 then a_inv = a_inv + |b|
    BN_acopy (V_u, b);
    BN_add_(a_inv, V_u);
  end;

  Result := BN_IsOne (V_nn[I_a]); // Result = gcd(a,b)=1

  BN_LFree (V_nn[0]);
  BN_LFree (V_nn[1]);
  BN_LFree (V_nn[2]);
  BN_LFree (V_nn[3]);
  BN_LFree (V_nn[4]);
  BN_LFree (V_nn[5]);
  BN_LFree (V_nn[6]);
  BN_LFree (V_nn[7]);
  BN_LFree (V_nn[8]);
  BN_LFree (V_q);
  BN_LFree (V_u);
end;

// [+] 1.2.1
{
  (R^2) mod p = a

  input  : a, p(prime number)
  output : R
}
function BN_sqrt_mod (var R: T_BNX; const a, p: T_BNX;
  AVerify: Boolean): Boolean;
var
  I       : LoopInt;
  V_L     : Integer;
  V_s     : UIntX;
  V_u     : T_BNX;
  V_b     : T_BNX;
  V_pmo   : T_BNX;
  V_t     : T_BNX;
  V_a_inv : T_BNX;
  V_c     : T_BNX;
  V_e     : T_BNX;
  V_d     : T_BNX;
begin
  Result := false;

  V_L := BN_jacobi (a, p);
  if V_L < 0 then begin // L = -1, -2 (error)
    BN_AssignZero (R);
    Exit;
  end;

  BN_AssignInt (V_b, 1, 3); // b = 3
  V_L := BN_jacobi (V_b, p);
  while V_L <> -1 do begin
    BN_inc (V_u, V_b);
    BN_copy (V_b, V_u);
    V_L := BN_jacobi (V_b, p);
  end;

  BN_dec (V_pmo, p); // pmo = p-1
  // p-1 = (2^s)*t
  V_s := BN_m2n_(V_pmo, V_t);

  // (a_inv*a) mod p = 1
  BN_inv_mod (V_a_inv, a, p);

  // c = (b^t) mod p
  BN_pow_mod_mon (V_c, V_b, V_t, p);

  BN_inc (V_u, V_t); // u = t+1
  BN_shr_1 (V_u);    // u = (t+1)/2
  // r = (a^u) mod p
  // r = (a^((t+1)/2) mod p
  BN_pow_mod_mon (R, a, V_u, p);

  if V_s >= 2 then begin
    for I := (V_s - 2) downto 1 do begin
      BN_AssignBit (V_e, I); // e = 2^i
      BN_sqr (V_b, R);
      BN_mul (V_u, V_b, V_a_inv);
      // d = ((R^2)*(a_inv))^(2^i) mod p
      BN_pow_mod_mon (V_d, V_u, V_e, p);

      if BN_Compare (V_d, V_pmo) = 0 then begin // d = p-1
        // r = (r*c) mod p
        BN_mul (V_u, R, V_c);
        BN_mod (R, V_u, p);
      end;

      // c = (c*c) mod p
      BN_sqr (V_u, V_c);
      BN_mod (V_c, V_u, p);
    end;

    // I = 0

    BN_sqr (V_b, R);
    BN_mul (V_u, V_b, V_a_inv);
    // d = ((R^2)*(a_inv)) mod p
    BN_mod (V_d, V_u, p);

    if BN_Compare (V_d, V_pmo) = 0 then begin // d = p-1
      // r = (r*c) mod p
      BN_mul (V_u, R, V_c);
      BN_mod (R, V_u, p);
    end;
  end;

  if AVerify then begin
    BN_sqr (V_u, R);
    BN_mod (V_b, V_u, p); // b = (R^2) mod p
    Result := BN_Compare (a, V_b) = 0;
  end else begin
    Result := true;
  end;

  BN_LFree (V_u);
  BN_LFree (V_b);
  BN_LFree (V_pmo);
  BN_LFree (V_t);
  BN_LFree (V_a_inv);
  BN_LFree (V_c);
  BN_LFree (V_e);
  BN_LFree (V_d);
end;

// [+] 1.2.1
{
  P = (a^e) mod n

   
  Argument test

   TRUE,   
  Returns TRUE if the result is defined
}
function BN_pow_mod_arg_(var P: T_BNX; const a, e, n: T_BNX): Boolean; overload;
// {$IFDEF BN_INLINE} inline; {$ENDIF}
begin
  Result := true;

  if BN_IsZero (e) then begin // e = 0
    BN_AssignPOne (P);        // P = 1
    Exit;
  end;

  if BN_IsZero (a) then begin // a = 0
    BN_AssignZero (P);        // P = 0
    Exit;
  end;

  Result := false;
end;

// [+] 1.2.1
{
  P = (a^e) mod n
  e - unsigned integer

   
  Argument test

   TRUE,   
  Returns TRUE if the result is defined
}
function BN_pow_mod_arg_(var P: T_BNX; const a: T_BNX; e: UIntX; const n: T_BNX): Boolean; overload;
// {$IFDEF BN_INLINE} inline; {$ENDIF}
begin
  Result := true;

  if e = 0 then begin  // e = 0
    BN_AssignPOne (P); // P = 1
    Exit;
  end;

  if BN_IsZero (a) then begin // a = 0
    BN_AssignZero (P);        // P = 0
    Exit;
  end;

  Result := false;
end;

// [+] 1.2.1
{
  P = (a^e) mod n

        
    
  Right-to-left binary method
}
procedure BN_pow_mod (var P: T_BNX; const a, e, n: T_BNX);
var
  v_c : T_BNX;
  v_e : T_BNX;
  v_u : T_BNX;
begin
  if BN_pow_mod_arg_(P, a, e, n) then Exit;

  BN_AssignPOne (P); // P = 1
  BN_copy (v_c, a);  // c = a
  BN_copy (v_e, e);  // v = e

  while true do begin
    if BN_IsOdd (v_e) then begin
      BN_mul (v_u, P, v_c);
      BN_mod (P, v_u, n); // s = s*c mod n
    end;

    if BN_shr_1 (v_e) = 0 then Break;
    // e = e div 2 = v shr 1

    BN_sqr (v_u, v_c);
    BN_mod (v_c, v_u, n); // c = c^2 mod n
  end;

  BN_LFree (v_c);
  BN_LFree (v_e);
  BN_LFree (v_u);
end;

// [+] 1.2.1
{
  P = (a^e) mod n
  e - unsigned integer

        
    
  Right-to-left binary method
}
procedure BN_pow_mod (var P: T_BNX; const a: T_BNX; e: UIntX; const n: T_BNX);
var
  v_c : T_BNX;
  v_u : T_BNX;
begin
  if BN_pow_mod_arg_(P, a, e, n) then Exit;

  BN_AssignPOne (P); // P = 1
  BN_copy (v_c, a);  // c = a

  while true do begin
    if Odd (e) then begin
      BN_mul (v_u, P, v_c);
      BN_mod (P, v_u, n); // s = s*c mod n
    end;

    e := e shr 1; // v = v div 2 = v shr 1
    if e = 0 then Break;

    BN_sqr (v_u, v_c);
    BN_mod (v_c, v_u, n); // c = c^2 mod n
  end;

  BN_LFree (v_c);
  BN_LFree (v_u);
end;

// [+] 1.2.1
{
  Returns Montgomery R for given n
  R = 2^K, n < R
}
function BN_mon_K_R_(var K: IntX; var R: T_BNX; const n: T_BNX): Boolean;
// {$IFDEF BN_INLINE} inline; {$ENDIF}
begin
  if BN_LE_Zero (n) then begin
    Result := false;   // n <= 0
    K := 0;            // K = 0
    BN_AssignZero (R); // R = 0
    Exit;
  end;

  K := BN_bsr (n) + 1;
  BN_AssignBit (R, K);

  Result := true;
end;

// [+] 1.2.1
{
  Returns Montgomery R, RR, NN for given n
  R = 2^K, n < R
  gcd (R, n) = 1
  R*RR - N*NN = 1
}
function BN_mon_K_R_N_(var K: IntX; var R, RR, NN: T_BNX; const n: T_BNX): Boolean;
// {$IFDEF BN_INLINE} inline; {$ENDIF}
var
  V_gcd : T_BNX;
begin
  Result := BN_mon_K_R_(K, R, n);
  if not Result then begin
    BN_AssignZero (RR); // RR = 0
    BN_AssignZero (NN); // NN = 0
    Exit;
  end;

  // R*RR - (-NN)*N = 1
  BN_gcd_bezout (V_gcd, RR, NN, R, n);
  Result := BN_IsPOne (V_gcd);
  BN_LFree (V_gcd);

  if not Result then begin
    BN_AssignZero (RR); // RR = 0
    BN_AssignZero (NN); // NN = 0
    Exit;
  end;

  BN_chs (NN);
end;

// [+] 1.2.1
{
  P = (a^e) mod n

   
  Montgomery modular exponentiation
}
function BN_pow_mod_mon (var P: T_BNX; const a, e, n: T_BNX): Boolean;
var
  V_K  : IntX;
  V_NN : T_BNX;
  V_t  : T_BNX;

  // Montgomery reduction
  procedure mon_reduce_(var A_p: T_BNX; const A_t: T_BNX); overload;
  begin
    // u = (t + ((t * NN) mod R) * n) div R
    BN_mul (A_p, A_t, V_NN);
    BN_mod_2n_(A_p, V_K);
    BN_mul_(A_p, n);
    BN_add_(A_p, A_t);
    BN_div_2n_(A_p, V_K);

    if A_p.Sign < 0 then begin
      // if u < 0 then u = u + n
      BN_add_(A_p, n);
    end else begin
      // if u >= n then u = u - n
      if BN_Compare (A_p, n) >= 0 then BN_sub_(A_p, n);
    end;
  end;

  // Montgomery product
  procedure mon_pro_(var A_p: T_BNX; const A_a, A_b: T_BNX); overload;
  begin
    // t = (a') * (b')
    BN_mul (V_t, A_a, A_b);

    // u = (t + ((t * NN) mod R) * n) div R
    BN_mul (A_p, V_t, V_NN);
    BN_mod_2n_(A_p, V_K);
    BN_mul_(A_p, n);
    BN_add_(A_p, V_t);
    BN_div_2n_(A_p, V_K);

    if A_p.Sign < 0 then begin
      // if u < 0 then u = u + n
      BN_add_(A_p, n);
    end else begin
      // if u >= n then u = u - n
      if BN_Compare (A_p, n) >= 0 then BN_sub_(A_p, n);
    end;
  end;

  // Montgomery square
  procedure mon_sqr_(var A_p: T_BNX; const A_a: T_BNX); overload;
  begin
    // t = (a') * (a')
    BN_sqr (V_t, A_a);

    // u = (t + ((t * NN) mod R) * n) div R
    BN_mul (A_p, V_t, V_NN);
    BN_mod_2n_(A_p, V_K);
    BN_mul_(A_p, n);
    BN_add_(A_p, V_t);
    BN_div_2n_(A_p, V_K);

    if A_p.Sign < 0 then begin
      // if u < 0 then u = u + n
      BN_add_(A_p, n);
    end else begin
      // if u >= n then u = u - n
      if BN_Compare (A_p, n) >= 0 then BN_sub_(A_p, n);
    end;
  end;

var
  I    : LoopInt;
  C    : LoopInt;
  H    : IntX;
  B    : IntX;
  D    : UIntX;

  V_R  : T_BNX;
  V_RR : T_BNX;
  V_a  : T_BNX;
  V_x  : T_BNX;
  V_u  : T_BNX;
begin
  if BN_pow_mod_arg_(P, a, e, n) then begin
    Result := true;
    Exit;
  end;

  try
    Result := BN_mon_K_R_N_(V_K, V_R, V_RR, V_NN, n);
    if not Result then Exit;

    // (a') = (a * R) mod n
    BN_pshl (V_u, a, V_K);
    BN_mod (V_a, V_u, n);

    // (x') = (1 * R) mod n
    BN_mod (V_x, V_R, n);

    // Find and process the first non-zero high order element
    H := -1;
    for C := e.HIndex downto 0 do begin
      D := e.Data[C];
      if D = 0 then Continue;

      B := bsr (D);
      for I := B downto 0 do begin
        mon_sqr_(V_x, V_x); // mon_pro_(V_x, V_x, V_x);
        if bit_s_test (D, I) then mon_pro_(V_x, V_x, V_a);
      end;

      H := C - 1;
      Break;
    end;

    // Process the remaining elements
    for C := H downto 0 do begin
      D := e.Data[C];
      for I := CXBits - 1 downto 0 do begin
        mon_sqr_(V_x, V_x); // mon_pro_(V_x, V_x, V_x);
        if bit_s_test (D, I) then mon_pro_(V_x, V_x, V_a);
      end;
    end;

    mon_reduce_(P, V_x);
  finally
    BN_LFree (V_R);
    BN_LFree (V_RR);
    BN_LFree (V_NN);
    BN_LFree (V_a);
    BN_LFree (V_x);
    BN_LFree (V_u);
    BN_LFree (V_t);
  end;
end;

// [+] 1.2.1
{
  ainv = (2^C) * (1/A)
  Return C
}
function BN_inv (var ainv: T_BNX; const A: T_BNX): IntX;
var
  n : IntX;
  U : T_BNX;
begin
  Result := 0;

  n := BN_bsr (A);
  if n < 0 then begin
    BN_AssignZero (ainv);
    Exit;
  end;

  n := 2 * n + 1;
  Result := n;

  BN_AssignBit (U, n);
  BN_div (ainv, U, A);

  BN_LFree (U);
end;

// [+] 1.2.1
{
  ainv = (2^n) * (1/A)
  Return n
}
function BN_inv (var ainv: T_BNX; A: UIntX): IntX;
var
  n : IntX;
  U : T_BNX;
begin
  Result := 0;

  n := bsr (A);
  if n < 0 then begin
    BN_AssignZero (ainv);
    Exit;
  end;

  n := 2 * n + 1;
  Result := n;

  BN_AssignBit (U, n);
  BN_div_int (ainv, U, 1, A);

  BN_LFree (U);
end;

// [+] 1.2.1
{
  ainv = (2^C) div A
  Return C

  Newton's method
  x(n+1) = x(n) - (f/f')
}
function BN_ninv (var ainv: T_BNX; const A: T_BNX): IntX;
var
  U     : T_BNX;
  V     : T_BNX;
  VData : UIntX;
  V_2B  : T_BNX;
  V_nn  : array [0..1] of T_BNX;
  n0, n : IntX;
  I_U   : Integer;
  I_0   : Integer;
  I_1   : Integer;
begin
  Result := 0;

  if BN_IsInt (A, VData) then begin
    Result := BN_inv (ainv, VData);
    Exit;
  end;

  n0 := BN_bsr (A);
  if n0 < 0 then begin
    BN_AssignZero (ainv);
    Exit;
  end;

  I_0 := 0;
  I_1 := 1;

  n := 2 * n0 + 1;
  Result := n;

  BN_AssignBit (U, n);
  BN_AssignBit (V_2B, n + 1);

  VData := A.PHData^;
  if VData = CXmax then begin
    // 2^(2 * n0 + 1) div 2^(n0 + 1)
    // = 2^(2 * n0 + 1 - n0 - 1)
    // = 2^(n0)
    BN_AssignBit (V_nn[I_0], n0);
  end else begin
    BN_div_intn (V_nn[I_0], U, 1, VData + 1, A.HIndex);
  end;

  // B/a    = x
  // f(x)   = a/B - 1/x = 0
  // f'(x)  = 1/x^2
  // f/f'   = x^2*a/B - x = (x^2*a - B*x)/B = -(B*x - x^2*a)/B
  // x(n+1) = x(n) - (f/f')
  //        = x(n) + (B*x(n) - x(n)^2*a)/B
  //        = (2*B*x(n) - x(n)^2*a)/B
  //        = x(n)*(2*B - x(n)*a)/B

  while true do begin
    BN_mul (U, A, V_nn[I_0]);
    BN_sub (V, V_2B, U);
    BN_mul (U, V, V_nn[I_0]);
    BN_shr (V_nn[I_1], U, n);

    if BN_Compare (V_nn[I_1], V_nn[I_0]) = 0 then begin
      BN_copy (ainv, V_nn[I_0]);
      Break;
    end;

    I_U := I_0;
    I_0 := I_1;
    I_1 := I_U;
  end;

  BN_LFree (U);
  BN_LFree (V);
  BN_LFree (V_2B);
  BN_LFree (V_nn[0]);
  BN_LFree (V_nn[1]);
end;

// [+] 1.2.1
{
  ainv = (2^C) div A
  Return C

  x(n+1) = x(n) - (f/f') - ((f/f')^2)*(f''/(2*f'))

  ,  BN_ninv
  Slower than BN_ninv
}
function BN_cinv (var ainv: T_BNX; const A: T_BNX): IntX;
var
  U     : T_BNX;
  P     : T_BNX;
  V     : T_BNX;
  VData : UIntX;
  V_B   : T_BNX;
  V_B2  : T_BNX;
  V_2B  : T_BNX;
  V_nn  : array [0..1] of T_BNX;
  n0, n : IntX;
  n2    : IntX;
  I_U   : Integer;
  I_0   : Integer;
  I_1   : Integer;
begin
  Result := 0;

  if BN_IsInt (A, VData) then begin
    Result := BN_inv (ainv, VData);
    Exit;
  end;

  n0 := BN_bsr (A);
  if n0 < 0 then begin
    BN_AssignZero (ainv);
    Exit;
  end;

  I_0 := 0;
  I_1 := 1;

  n := 2 * n0 + 1;
  Result := n;

  n2 := n * 2;
  BN_AssignBit (V_B, n);
  BN_AssignBit (V_B2, n2);    // B^2
  BN_AssignBit (V_2B, n + 1); // 2*B

  VData := A.PHData^;
  if VData = CXmax then begin
    // 2^(2 * n0 + 1) div 2^(n0 + 1)
    // = 2^(2 * n0 + 1 - n0 - 1)
    // = 2^(n0)
    BN_AssignBit (V_nn[I_0], n0);
  end else begin
    BN_div_intn (V_nn[I_0], V_B, 1, VData + 1, A.HIndex);
  end;

  // B/a    = x
  // f(x)   = a/B - 1/x = 0
  // f'(x)  = 1/x^2
  // f''(x) = -2/x^3
  // x(n+1) = x(n) - (f/f') - ((f/f')^2)*(f''/(2*f'))
  //        = x(n)*(B^2 + (B - x(n)*a)*(2*B - x(n)*a))/(B^2)

  while true do begin
    BN_mul (U, A, V_nn[I_0]);
    BN_sub (P, V_B, U);
    BN_sub (V, V_2B, U);
    BN_mul (U, P, V);
    BN_add (V, U, V_B2);
    BN_mul (U, V, V_nn[I_0]);
    BN_shr (V_nn[I_1], U, n2);

    if BN_Compare (V_nn[I_1], V_nn[I_0]) = 0 then begin
      BN_copy (ainv, V_nn[I_0]);
      Break;
    end;

    I_U := I_0;
    I_0 := I_1;
    I_1 := I_U;
  end;

  BN_LFree (U);
  BN_LFree (P);
  BN_LFree (V);
  BN_LFree (V_B);
  BN_LFree (V_B2);
  BN_LFree (V_2B);
  BN_LFree (V_nn[0]);
  BN_LFree (V_nn[1]);
end;

{
  
  Factorial

  R = a * (a-1) * (a-2) * ... * 3 * 2 * 1
}
procedure BN_fact (var R: T_BNX; const a: T_BNX);
var
  V_nn  : array [0..1] of T_BNX;
  I_U   : Integer;
  I_0   : Integer;
  I_1   : Integer;
  VUInt : UIntX;
  V_sha : T_BNX;
  U, Va : T_BNX;
  n, nn : UIntX;
begin
  if a.Sign < 0 then begin
    raise ERangeError.Create ('a! a<0');
  end;

  if BN_IsInt (a, VUInt) then begin
    BN_fact (R, VUInt);
    Exit;
  end;

  I_0 := 0;
  I_1 := 1;

  nn := BN_m2n_(a, V_sha);
  if nn = 0 then begin
    BN_copy (V_nn[I_0], a);
  end else begin
    BN_copy (V_nn[I_0], V_sha);
  end;
  BN_copy (Va, a);

  while true do begin
    BN_dec_(Va, U);
    n := BN_m2n_(Va, V_sha);
    if n = 0 then begin
      BN_mul (V_nn[I_1], V_nn[I_0], Va);
    end else begin
      nn := nn + n;
      BN_mul (V_nn[I_1], V_nn[I_0], V_sha);
    end;
    if BN_IsOne (Va) then Break;

    I_U := I_0;
    I_0 := I_1;
    I_1 := I_U;
  end;

  BN_pshl (R, V_nn[I_1], nn);

  BN_LFree (V_nn[0]);
  BN_LFree (V_nn[1]);
  BN_LFree (V_sha);
  BN_LFree (U);
  BN_LFree (Va);
end;

{
  
  Factorial

  R = a * (a-1) * (a-2) * ... * 3 * 2 * 1
}
procedure BN_fact (var R: T_BNX; a: UIntX);
var
  V_nn  : array [0..1] of T_BNX;
  I_U   : Integer;
  I_0   : Integer;
  I_1   : Integer;
  V_sha : UIntX;
  n, nn : UIntX;
begin
  if a <= 1 then begin
    BN_AssignPOne (R); // 0! = 1 // 1! = 1
    Exit;
  end;

  I_0 := 0;
  I_1 := 1;

  nn := BN_m2n_int_(a, V_sha);
  if nn = 0 then begin
    BN_AssignInt (V_nn[I_0], 1, a);
  end else begin
    BN_AssignInt (V_nn[I_0], 1, V_sha);
  end;

  while true do begin
    Dec (a);
    n := BN_m2n_int_(a, V_sha);
    if n = 0 then begin
      BN_mul_int (V_nn[I_1], V_nn[I_0], 1, a);
    end else begin
      nn := nn + n;
      BN_mul_int (V_nn[I_1], V_nn[I_0], 1, V_sha);
    end;
    if a = 1 then Break;

    I_U := I_0;
    I_0 := I_1;
    I_1 := I_U;
  end;

  BN_pshl (R, V_nn[I_1], nn);

  BN_LFree (V_nn[0]);
  BN_LFree (V_nn[1]);
end;

end.

